/*********************************/
/*                               */
/* Rotation Data Module (RDM)    */
/*                               */
/*********************************/

/*-Title: Rotation Data Module (RDM)
-end*/

/*-Intro:
The Rotation Data Module (RDM) defines a set of parameters for use in the 
initial stages of processing sets of protein crystallographic X-ray 
diffraction monochromatic rotation images. The RDM is based on a general
set of keyword handling and data storage routines, the  Keyword Data
Module (KDM). This handles all the parameters except for the crystal
symmetry which is handled by a separate set of routines. 
-end*/


/* Define Fortran binding names */

#if    LINKTYP == 1

#define rdmf_setup rdmf_setup_
#define rdmf_parsesym rdmf_parsesym_
#define rdmf_set_sym rdmf_set_sym_
#define rdmf_reset_symm rdmf_reset_symm_
#define rdmf_symm_chreset rdmf_symm_chreset_
#define rdmf_symm_changed rdmf_symm_changed_
#define rdmf_get_symm rdmf_get_symm_
#define rdmf_get_spgrp rdmf_get_spgrp_
#define rdmf_get_nsym rdmf_get_nsym_
#define rdmf_get_sylat rdmf_get_sylat_
#define rdmf_symm_str rdmf_symm_str_
#define rdmf_symm_list rdmf_symm_list_
#define rdmf_symm_write rdmf_symm_write_
#define rdmf_sysabs rdmf_sysabs_
#define rdmf_sysahk rdmf_sysahk_ 
#define rdmf_asuput rdmf_asuput_
#define rdmf_asuget rdmf_asuget_
#define rdmf_centr rdmf_centr_
#define rdmf_getrotations rdmf_getrotations_
#define rdmf_numimg rdmf_numimg_
#define rdmf_rot_start rdmf_rot_start_
#define rdmf_getimgpars rdmf_getimgpars_
#define rdmf_getdetpars rdmf_getdetpars_
#define rdmf_getcellwave rdmf_getcellwave_
#define rdmf_getcentre rdmf_getcentre_
#define rdmf_getdistor rdmf_getdistor_
#define rdmf_getmmrast rdmf_getmmrast_
#define rdmf_write rdmf_write_
#define rdmf_mosdet rdmf_mosdet_
#define rdmf_bmatrx rdmf_bmatrx_
#define rdmf_reccel rdmf_reccel_
#define rdmf_rtm rdmf_rtm_
#define rdmf_rtod rdmf_rtod_
#define rdmf_dtor rdmf_dtor_
#define rdmf_corr rdmf_corr_
#define rdmf_ideal rdmf_ideal_
#define rdmf_mmtoras rdmf_mmtoras_
#define rdmf_rastomm rdmf_rastomm_

#endif

#if  LINKTYP == 2

#define rdmf_setup RDMF_SETUP
#define rdmf_parsesym RDMF_PARSESYM
#define rdmf_set_sym RDMF_SET_SYM
#define rdmf_reset_symm RDMF_RESET_SYMM
#define rdmf_symm_chreset RDMF_SYMM_CHRESET
#define rdmf_symm_changed RDMF_SYMM_CHANGED
#define rdmf_get_symm RDMF_GET_SYMM
#define rdmf_get_spgrp RDMF_GET_SPGRP
#define rdmf_get_nsym RDMF_GET_NSYM
#define rdmf_get_sylat RDMF_GET_SYLAT
#define rdmf_symm_str RDMF_SYMM_STR
#define rdmf_symm_list RDMF_SYMM_LIST
#define rdmf_symm_write RDMF_SYMM_WRITE
#define rdmf_sysabs RDMF_SYSABS
#define rdmf_sysahk RDMF_SYSAHK 
#define rdmf_asuput RDMF_ASUPUT
#define rdmf_asuget RDMF_ASUGET
#define rdmf_centr RDMF_CENTR
#define rdmf_getrotations RDMF_GETROTATIONS
#define rdmf_numimg RDMF_NUMIMG
#define rdmf_rot_start RDMF_ROT_START
#define rdmf_getimgpars RDMF_GETIMGPARS
#define rdmf_getdetpars RDMF_GETDETPARS
#define rdmf_getcellwave RDMF_GETCELLWAVE
#define rdmf_getcentre RDMF_GETCENTRE
#define rdmf_getdistor RDMF_GETDISTOR
#define rdmf_getmmrast RDMF_GETMMRAST
#define rdmf_write RDMF_WRITE
#define rdmf_mosdet RDMF_MOSDET
#define rdmf_bmatrx RDMF_BMATRX
#define rdmf_reccel RDMF_RECCEL
#define rdmf_rtm RDMF_RTM
#define rdmf_rtod RDMF_RTOD
#define rdmf_dtor RDMF_DTOR
#define rdmf_corr RDMF_CORR
#define rdmf_ideal RDMF_IDEAL
#define rdmf_mmtoras RDMF_MMTORAS
#define rdmf_rastomm RDMF_RASTOMM

#endif

void rdmf_setup();
void rdmf_parsesym();
void rdmf_set_sym();
void rdmf_reset_symm();
void rdmf_symm_chreset();
void rdmf_symm_changed();
void rdmf_get_symm();
void rdmf_get_spgrp();
void rdmf_get_nsym();
void rdmf_get_sylat();
void rdmf_symm_str();
void rdmf_symm_list();
void rdmf_symm_write();
void rdmf_sysabs();
void rdmf_sysahk(); 
void rdmf_asuput();
void rdmf_asuget();
void rdmf_centr();
void rdmf_getrotations();
void rdmf_numimg();
void rdmf_rot_start();
void rdmf_getimgpars();
void rdmf_getdetpars();
void rdmf_getcellwave();
void rdmf_getcentre();
void rdmf_getdistor();
void rdmf_getmmrast();
void rdmf_write();
void rdmf_mosdet();
void rdmf_bmatrx();
void rdmf_reccel();
void rdmf_rtm();
void rdmf_rtod();
void rdmf_dtor();
void rdmf_corr();
void rdmf_ideal();
void rdmf_mmtoras();
void rdmf_rastomm();

#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#include <math.h>
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include "kdm.h"

#include "kdm_ext.h"
#include "rdm_systyp.h"
#include "rdm.h"

#define MAX_SETS 10          /* Maximum no. of crystal sets */
#define MAX_IMAGES 2000      /* Maximum no. of images / set */
#define MAX_NAME_LEN 50      /* Maximum allowed length of keyword */
#define MAXVLENGTH 250       /* Maximum allowed length for a parameter
                                value string including leading blank */
#define MAX_LEN 250          /* Maximum length of an input line */

/* Nearest integer macro */

#define nint(x)  x>=0.0?(int)(x+0.5):(int)(x-0.5)

/* Define internal routine types */

int rdm_axord (char * str, int len, int * order, int * swaptyp);
int rdm_symfr2 (char * str, int i1, int * ns, float rot[][4][4],  
                char * errstr, int maxe);
int rdm_read_symm (int lspgrp, char * namspg, char * errstr, int maxe);
int rdm_pgdefn(int nsym, float rsmt[][4][4], char * nampg, int * nsymmp);
void rdm_opstr (char * linout, char * outstr, int maxo, void (*op_func)());
void rdm_pgnlau (char * nampg, int * nlaue, char * launam);
int rdm_inasu (int * ih, int nlaue);
int rdm_set_centric ();
void rdm_determ (float a[4][4], float * det);
void rdm_invsym (float a[4][4], float ainv[4][4]);
void debug_symm(int k);


/*---------------------------------------------------------------
|                  RDM      SYMMETRY STORAGE                    |
----------------------------------------------------------------*/

static char spgnam[11];          /* Space group name (may be null if symmetry 
                                    not i/p from SYMOP) */
static char pgnam[11];           /* Point group name */
static char systyp[13];          /* System type name */
static int nspgrp = 0;           /* Space group number (may be 0 if symmetry 
                                    not i/p from SYMOP) */
static int nsymp = 0;            /* No. of primitive symmetry operators */
static int nsym = 0;             /* No. of symmetry operators (total); 
                                    Note: 0 means symmetry undefined. */
static float rsym [192][4][4];   /* Symmetry matrices */
static int sym_completed = 0;    /* Symmetry input completed flag =1 yes
                                    (e.g. input from SYMOP file) =0 no
                                    SYMMETRY cards input but no final
                                    'SYMMETRY END' card yet given.
                                    Only when this has been set will PGDEFN
                                    etc be called so that rdm_asuput,
                                    rdm_asuget & rdm_centr may be called */


/* Symmetry changes mask */

static int sym_ch_mask = 0xff;
static int is_symm = 0;
int ch_symm = 0;

/* Derived parameters  (set via calls to rdm_set_symm) */
/* =================================================== */

static int isys;                    /* Flags for systematic absence checks
                                       rdm_sysabs, rdm_sysahk - depend on 
                                       lattice type */
static int ksys[3];                 /* " */
static float rsym_inv [192][4][4];  /* Inverted symmetry matrices */

int nlaue;                          /* Laue group number set via rdm_pgnlau
                                       call */
char launam[9];                     /* Laue group name set via rdm_pgnlau
                                       call */


/* Parameters from rdm_pgdefn (called automatically from rdm_set_symm) */

int klass;
int icentr;
int iscrew[3];

/* Parameters from rdm_pgnlau (called automatically from rdm_set_symm) */

int hrng0;
int krng0;
int lrng0;
int hrng1;
int krng1;
int lrng1;

/* Parameters from rdm_set_centric (called automatically from rdm_set_symm) */

int cent_setup = 0;
int ncent;
float cproj[20][3];

/*-------------------------------------------------------------------
|                                                                   |
|                            RDM  ROUTINES                          |
|                                                                   |
--------------------------------------------------------------------*/

/*-Section: Set up the RDM Dataset
A single routine sets up all the RDM parameters except for the symmetry
keyword which must be handled separately. The keyworded data is set up
using the general purpose Keyword Data Module (KDM) routines.
-end*/

/*-Routine: Set up the RDM dataset - rdm_setup
The initialisation routine rdm_setup (rdmf_setup) initialises the
Rotation Data Module (RDM) defining all the parameters
(except for the symmetry parameter) and allocating storage for them.
-end*/

/*
***************************
**     rdm_setup         **
***************************

Purpose: Initialise RDM dataset

Author:  John W. Campbell, November 1997

CD-Fortran:
         CALL RDMF_SETUP (KINDX, IERR)
CD-end:
*/
/*-Parameters:
KINDX       i  (W)  Returns the KDM index for the RDM dataset to be used 
                    in all calls to RDM and KDM routines when accessing
                    RDM parameter.
IERR        i  (W)  The error return flag from rdm_setup.
-end*/

/*-C:*/
int rdm_setup (int * kindx)
/*end*/

/*-Parameters:
int * kindx;        Returns the KDM index for the RDM dataset to be used 
                    in all calls to RDM and KDM routines when accessing
                    RDM parameter (W)

Return:  Error flag =0 OK
                    =1 Memory allocation error or too many KDM datasets 
                       defined
-end*/
{
   int err;
   int kdx;
   int i_deflt[10];
   float f_deflt[10];
   char *sys_str[] ={"Triclinic",
                     "Monoclinic",
                     "Orthorhombic",
                     "Tetragonal",
                     "Hexagonal",
                     "Rhombohedral",
                     "Cubic"};
   char *latt_str[] ={"P","A","B","C","I","F","R"};
   static char * det_str[] = {"unknown","mar"};
   static char * gonio_str[] = {"simple","3-circle"};
   static char * gonio_ori_str[] = {"right","left","up", "down"};
   static char * gonio_rot_str[] = {"phi","omega"};
   static char * imgtyp_str[] = {"unknown","ip","ccd"};
   static char * imgfrm_str[] = {"none","raw","byte","i2","mar"};
   static char * distor_str[] = {"standard","rtoff"};
   static float u_default[] ={1.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,1.0};
   static char * synch_str[] = {"Yes","No"};
   static char * titl_str[] = {" "," "};
   static float detr_default[] = {1.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,1.0};
   static float deta_default[] = {0.0,1.0,0.0,0.0,0.0,1.0};
   static float s0_default[] = {-1.0,0.0,0.0};
   static float scax_default[] = {0.0,0.0,1.0};
   char str[51];
   char *dumstrs[] = {" "," "};

   *kindx = 0;

   /* Initialise KDM keyword dataset */

   err = kdm_init (MAX_SETS, MAX_IMAGES, "NUMSETS", 0, " ", 0, "SET", 
                   0, "IMAGE", 0, &kdx);
   if (err!=0) goto laberr1;


   /*Define parameters */
    
   strcpy(str,"TITLE");
   err = kdm_define_str (kdx, str, 0, 4, 4, 1, 120," ", 0, titl_str,
                         2, 0, 0, 0, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"NUMSETS");
   i_deflt[0] = 1;
   err = kdm_define_int (kdx, str, 0, 4, 1, 1, 1, i_deflt, 0, 0, 7, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"SYSTEM");
   err = kdm_define_str (kdx, str, 0, 4, 1, 0, 20, "triclinic", 0, 
                         sys_str, 7, 0, 1, -3, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"LATTICE");
   err = kdm_define_str (kdx, str, 0, 4, 1, 0, 20, "p", 0, latt_str,
                         7, 0, 1, -1, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"DET_TYPE");
   err = kdm_define_str (kdx, str, 0, 5, 2, 1, 20, "unknown", 0, det_str,
                         2, 0, 1, 1, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"DISTANCE");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 1, 1, f_deflt, 0.0, 0.0, 2, 
                           0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"RMAX");
   f_deflt[0] = 150.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 1, 1, f_deflt, 0.0, 0.0, 
                           2, 0, 2, 1);
   if (err!=0) goto laberr2;
  
   strcpy(str,"DET_ROTATIONS");
   err = kdm_define_float (kdx, str, 0, 5, 2, 0, 9, detr_default, -1.0, 1.0,
                           5, 0, -4, 1);
   if (err!=0) goto laberr2;
   
   strcpy(str,"DET_AXES");
   err = kdm_define_float (kdx, str, 0, 5, 2, 0, 6, deta_default, -1.0, 1.0,
                           5, 0, -4, 1);
   if (err!=0) goto laberr2;
   
   strcpy(str,"TAU1");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 0, 1, f_deflt, 0.0, 0.0, 0, 
                           0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"TAU2");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 0, 1, f_deflt, 0.0, 0.0, 0, 
                           0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"TAU3");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 0, 1, f_deflt, 0.0, 0.0, 0, 
                           0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"IAX_H");
   i_deflt[0] = 3;
   err = kdm_define_int (kdx, str, 0, 5, 1, 0, 1, i_deflt, 1, 3, 5, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"IAX_V");
   i_deflt[0] = 2;
   err = kdm_define_int (kdx, str, 0, 5, 1, 0, 1, i_deflt, 1, 3, 5, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"BEAM_VECTOR");
   err = kdm_define_float (kdx, str, 0, 4, 2, 0, 3, s0_default, -1.0, 1.0, 
                           5, 0, -4, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"SCAN_AXIS");
   err = kdm_define_float (kdx, str, 0, 4, 2, 0, 3, scax_default, -1.0, 1.0,
                           5, 0, -4, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"DET_GAIN");
   f_deflt[0] = 1.0;
   err = kdm_define_float (kdx, str, 0, 5, 2, 1, 1, f_deflt, 0.0, 0.0, 
                           2, 0, 2, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"WAVELENGTH");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 1, 1, f_deflt, 0.0, 0.0, 
                           2, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"SYNCHROTRON");
   err = kdm_define_str (kdx, str, 0, 4, 2, 1, 6, "yes", 0, synch_str,
                         2, 0, 1, 1, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"DISPERSION");
   f_deflt[0] = 0.0015;
   err = kdm_define_float (kdx, str, 0, 4, 2, 0, 1, f_deflt, 0.0, 0.0, 
                           1, 0, -4, 4);
   if (err!=0) goto laberr2;

   strcpy(str,"DIVV");
   f_deflt[0] = 0.01;
   err = kdm_define_float (kdx, str, 0, 4, 2, 0, 1, f_deflt, 0.0, 0.0, 
                           1, 0, -4, 2);
   if (err!=0) goto laberr2;

   strcpy(str,"DIVH");
   f_deflt[0] = 0.10;
   err = kdm_define_float (kdx, str, 0, 4, 2, 0, 1, f_deflt, 0.0, 0.0, 
                           1, 0, -4, 2);
   if (err!=0) goto laberr2;

   strcpy(str,"DELCOR");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 0, 1, f_deflt, 0.0, 0.0, 
                           0, 0, -4, 2);
   if (err!=0) goto laberr2;

   strcpy(str,"POLARISATION");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 0, 1, f_deflt, -1.0, 1.0, 
                           5, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"GONIO_TYPE");
   err = kdm_define_str (kdx, str, 0, 7, 2, 1, 8, "simple", 0, gonio_str,
                         2, 0, 1, 1, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"GONIO_ORIENT");
   err = kdm_define_str (kdx, str, 0, 7, 2, 0, 8, "right", 0, gonio_ori_str,
                         4, 0, 1, 1, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"GONIO_ROT");
   err = kdm_define_str (kdx, str, 0, 7, 2, 0, 5, "phi", 0, gonio_rot_str,
                         2, 0, 1, 1, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"GONSTART_PHI");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 10, 2, 0, 1, f_deflt, 
                           -720.0, 720.0, 5, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"GONSTART_KAPPA");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 10, 2, 0, 1, f_deflt, 
                           -720.0, 720.0, 5, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"GONSTART_OMEGA");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 10, 2, 0, 1, f_deflt, 
                           -720.0, 720.0, 5, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"IMAGE_TYPE");
   err = kdm_define_str (kdx, str, 0, 7, 2, 0, 7, "ip", 0, imgtyp_str,
                         3, 0, 1, 1, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"IMAGE_FORMAT");
   err = kdm_define_str (kdx, str, 0, 7, 2, 1, 4, "none", 0, imgfrm_str,
                         5, 0, 1, 1, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"IMAGE_DATA");
   err = kdm_define_str (kdx, str, 0, 7, 2, 0, 7, "i2", 0, imgfrm_str,
                         7, 0, 1, 0, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"IMAGE_RECL");
   i_deflt[0] = 0;
   err = kdm_define_int (kdx, str, 0, 7, 2, 0, 1, i_deflt, 0, 0, 1, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"IMAGE_HEADER");
   i_deflt[0] = 0;
   err = kdm_define_int (kdx, str, 0, 7, 2, 0, 1, i_deflt, 0, 0, 0, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"AXORD");
   err = kdm_define_str (kdx, str, 0, 4, 2, 0, 12, "xy", 0, dumstrs,
                         0, 0, 0, 0, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"NULL_PIXEL");
   i_deflt[0] = 0;
   err = kdm_define_int (kdx, str, 0, 6, 2, 0, 1, i_deflt, 0, 0, 0, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"OVLD_PIXEL");
   i_deflt[0] = 0;
   err = kdm_define_int (kdx, str, 0, 6, 2, 0, 1, i_deflt, 0, 0, 1, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"NXRASTS");
   i_deflt[0] = 0;
   err = kdm_define_int (kdx, str, 0, 4, 2, 1, 1, i_deflt, 0, 0, 1, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"NYRASTS");
   i_deflt[0] = 0;
   err = kdm_define_int (kdx, str, 0, 4, 2, 1, 1, i_deflt, 0, 0, 1, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"X_CEN");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 1, 1, f_deflt, 0.0, 0.0, 
                           1, 0, 2, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"Y_CEN");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 1, 1, f_deflt, 0.0, 0.0, 
                           1, 0, 2, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"RASTER");
   f_deflt[0] = 10.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 1, 1, f_deflt, 0.0, 0.0, 
                           2, 0, 2, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"TWOTH_MIN");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 7, 2, 0, 1, f_deflt, 0.0, 0.0, 
                           1, 0, 2, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"RMIN");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 0, 1, f_deflt, 0.0, 0.0, 
                           1, 0, 2, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"R_XCEN");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 0, 1, f_deflt, 0.0, 0.0, 
                           1, 0, 2, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"R_YCEN");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 0, 1, f_deflt, 0.0, 0.0, 
                           1, 0, 2, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"X_MIN");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 0, 1, f_deflt, 0.0, 0.0, 
                           0, 0, 2, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"X_MAX");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 0, 1, f_deflt, 0.0, 0.0, 
                           0, 0, 2, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"Y_MIN");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 0, 1, f_deflt, 0.0, 0.0, 
                           0, 0, 2, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"Y_MAX");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 0, 1, f_deflt, 0.0, 0.0, 
                           0, 0, 2, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"X_C");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 3, 3, 0, 1, f_deflt, 0.0, 0.0, 
                           0, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"Y_C");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 3, 3, 0, 1, f_deflt, 0.0, 0.0, 
                           0, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"W_C");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 3, 3, 0, 1, f_deflt, 0.0, 0.0, 
                           0, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"DISTOR_TYPE");
   err = kdm_define_str (kdx, str, 0, 8, 2, 0, 8, "standard", 0, distor_str,
                         2, 0, 1, 1, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"TWIST");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 4, 3, 0, 1, f_deflt, 0.0, 0.0, 
                           0, 0, 2, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"TILT");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 4, 3, 0, 1, f_deflt, 0.0, 0.0, 
                           0, 0, 2, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"ROFF");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 4, 3, 0, 1, f_deflt, 0.0, 0.0, 
                           0, 0, 2, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"TOFF");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 4, 3, 0, 1, f_deflt, 0.0, 0.0, 
                           0, 0, 2, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"Y_SCALE");
   f_deflt[0] = 1.0;
   err = kdm_define_float (kdx, str, 0, 4, 3, 0, 1, f_deflt, 0.0, 0.0, 
                           0, 0, 2, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"IMAGE_DIR");
   err = kdm_define_str (kdx, str, 0, 7, 2, 1, 250, " ", 0, dumstrs,
                         0, 0, 0, 0, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"IMAGE_TEMPLATE");
   err = kdm_define_str (kdx, str, 0, 7, 2, 1, 250, "img###.img", 0, dumstrs,
                         0, 0, 0, 0, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"A");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 1, 2, 1, 1, f_deflt, 0.0, 0.0, 
                           2, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"B");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 1, 2, 1, 1, f_deflt, 0.0, 0.0, 
                           2, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"C");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 1, 2, 1, 1, f_deflt, 0.0, 0.0, 
                           2, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"ALPHA");
   f_deflt[0] = 90.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 1, 1, f_deflt, 0.0, 180.0, 
                           5, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"BETA");
   f_deflt[0] = 90.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 1, 1, f_deflt, 0.0, 180.0, 
                           5, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"GAMMA");
   f_deflt[0] = 90.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 1, 1, f_deflt, 0.0, 180.0, 
                           5, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"RESOLUTION");
   f_deflt[0] = 2.5;
   err = kdm_define_float (kdx, str, 0, 4, 2, 1, 1, f_deflt, 0.0, 0.0, 
                           2, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"UMATRIX");
   err = kdm_define_float (kdx, str, 0, 4, 2, 1, 9, u_default, 0.0, 0.0, 
                           0, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"PHI1");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 1, 1, f_deflt, -360.0, 360.0,
                           5, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"PHI2");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 1, 1, f_deflt, -360.0, 360.0,
                           5, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"PHI3");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 1, 1, f_deflt, -360.0, 360.0,
                           5, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"PHI_ORIENT");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 5, 2, 0, 1, f_deflt, -360.0, 360.0,
                           5, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"ROTSTART");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 5, 2, 1, 1, f_deflt, -360.0, 360.0,
                           5, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"ROTEND");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 5, 2, 1, 1, f_deflt, -360.0, 720.0,
                           5, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"ROTS2");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 5, 2, 0, 1, f_deflt, -360.0, 360.0,
                           5, 0, 3, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"ROTE2");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 5, 2, 0, 1, f_deflt, -360.0, 720.0,
                           5, 0, 3, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"ROTS3");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 5, 2, 0, 1, f_deflt, -360.0, 360.0,
                           5, 0, 3, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"ROTE3");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 5, 2, 0, 1, f_deflt, -360.0, 720.0,
                           5, 0, 3, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"ROTS4");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 5, 2, 0, 1, f_deflt, -360.0, 360.0,
                           5, 0, 3, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"ROTE4");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 5, 2, 0, 1, f_deflt, -360.0, 360.0,
                           5, 0, 3, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"ROTS5");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 5, 2, 0, 1, f_deflt, -360.0, 720.0,
                           5, 0, 3, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"ROTE5");
   f_deflt[0] = 0.0;
   err = kdm_define_float (kdx, str, 0, 5, 2, 0, 1, f_deflt, -360.0, 720.0,
                           5, 0, 3, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"OSCANGLE");
   f_deflt[0] = 1.0;
   err = kdm_define_float (kdx, str, 0, 4, 2, 1, 1, f_deflt, 0.01, 360.0, 
                           5, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"NWMAX");
   i_deflt[0] = 3;
   err = kdm_define_int (kdx, str, 0, 4, 2, 1, 1, i_deflt, 3, 1000, 5, 0);
   if (err!=0) goto laberr2;


   strcpy(str,"MOSAICITY");
   f_deflt[0] = 0.05;
   err = kdm_define_float (kdx, str, 0, 4, 2, 1, 1, f_deflt, 0.0, 0.0, 
                           1, 0, 4, 2);
   if (err!=0) goto laberr2;

   strcpy(str,"SPOT_DIAMETER");
   f_deflt[0] = 0.5;
   err = kdm_define_float (kdx, str, 0, 6, 2, 1, 1, f_deflt, 0.0, 0.0, 
                           2, 0, 3, 1);
   if (err!=0) goto laberr2;

   strcpy(str,"AUTIDX_IMAGE");
   i_deflt[0] = 1;
   err = kdm_define_int (kdx, str, 0, 4, 2, 1, 1, i_deflt, 0, 0, 1, 0);
   if (err!=0) goto laberr2;


   strcpy(str,"AUTIDX_IMAGE2");
   i_deflt[0] = 0;
   err = kdm_define_int (kdx, str, 0, 4, 2, 1, 1, i_deflt, 0, 0, 1, 0);
   if (err!=0) goto laberr2;

   strcpy(str,"AUTIDX_IMAGE3");
   i_deflt[0] = 0;
   err = kdm_define_int (kdx, str, 0, 4, 2, 1, 1, i_deflt, 0, 0, 1, 0);
   if (err!=0) goto laberr2;

   *kindx = kdx;
   return 0;

   /* Error Conditions - Programming Errors */

   laberr1:
   if (err==-1) return 1;
   printf("Error %d initialising RDM data set\n",err);
   exit(1);

   laberr2:
   printf("Error %d setting up keyword %s in RDM data set\n",err,str);
   exit(1);
}
/* Fortran binding: rdmf_setup */

void rdmf_setup (int * kindx, int *ierr)
{

   *ierr = rdm_setup (kindx);
   return;
}

/*-Section: Symmetry handling
This section conatins routines for handling the space group symmetry
for the Rotation Data Module. They could also be used for handling
symmetry in conjunction with some other set of KDM based parameters.
-end*/

/*-Routine: Parse symmetry data  - rdm_parsesym
This routine parses a given line of data, examining it for the presence
of the keyword 'SYMMETRY' (as the first token) Any such data found is 
automatically checked and, if valid, stored in the RDM.
If the line does not contain symmetry data from the requested set, 
then it is merely passed back to the calling program to allow further
interpretation by that program. The symmetry data line must contain
only symmetry data. A user supplied routine may make 
an additional check and disable its update if desired.
-end*/


/*
***************************
**     rdm_parsesym      **
***************************

Purpose: Parse line for symmetry data

Author:  John W. Campbell, December 1997

CD-Fortran:
         CALL RDMF_PARSESYM (KDMSTR(LINE), LENL, CHK_FUN, KDMSTR(BADTOK),
        +                MAXB, KDMSTR(ERRSTR), MAXE, IFLAG)
CD-end:
*/
/*-Parameters:
LINE        c  (R)  The line containing the keyworded symmetry string
                    ** Pass address index using the KDMSTR function **
LENL        i  (R)  The length of the fine
CHK_FUN     s  (R)  Subroutine which will be called as an additional
                    check that symmetry is to be set (if found).

                    call CHK_FUN (IOK)

                    Return IOK = 1 if symmetry to be set, otherwise return
                    IOK = 0.
BADTOK      c  (W)  Returns a bad token if error found
MAXB        c  (W)  Maximum length for the bad token string
ERRSTR      c  (W)  Error message if IFLAG < 0
                    ** Pass address index using the KDMSTR function **
MAXE        i  (R)  Maximum length for the error string
IFLAG       i  (W)  The return flag from rdm_parsesym.
-end*/

/*-C:*/
int rdm_parsesym (char * line, int lenl, void (*chk_func)(),
                  char * badtok, int maxb, char * errstr, int maxe)
/*end*/

/*-Parameters:
char * line;        Character string containing line to be parsed
                    (Assume comments & continuations already removed) (R)
int lenl;           Length of line; may be 0 if line is null terminated.
                    (R)
void (*chk_func)(); Function to make additional check on KDM parameters
                    to restrict updates (or return other information to 
                    calling program about the parameters found via common 
                    blocks)

                    The call will be as follows: chk_func(iok)    

                    int * iok  Return 1 if symmetry is to be updated,
                               otherwise return 0; 
                    (R)
char * badtok;      Returns the bad token string - must be at least maxb+1
                    characters long (W)
int maxb;           if >0 Maximum length of returned bad token string allowed
                          (excluding terminating null - badtok must be at
                          least maxb + 1 characters in length)
                    if <0 abs(maxb) characters will be returned padded
                          with blanks if needed (for 'fortran' use) 
                    
char * errstr;      Error string ( max 80 characters) - must be at least
                    maxe+1 characters long (W)
int maxe;           if >0 Maximum length of returned error string allowed
                          (excluding terminating null - errstr must be at
                          least maxe + 1 characters in length)
                    if <0 abs(maxe) characters will be returned padded
                          with blanks if needed (for 'fortran' use) 
                    (should allow 80 chars for error messages) (R)
int maxe;           Maximum length for error string (R)

Return    = 0,  Symmetry data data found & OK
          = 1, z Not symmetry data
          = 2,  No non-blank tokens of requested set
          < 0,  Symmetry data but  error(s)
          =-1,  Invalid or inappropriate set specification
          =-2,  Invalid or inappropriate subset specification
          =-3,  Syntax error in set/subset specification
          =-4,  Invalid value 
-end*/
{
   int iflag;
   int i;
   int err;
   int il;
   int ll;
   int p1, p2;
   int iset;
   int isubset;
   int iss;
   int ncat;
   int iok;
   char keyword[MAX_NAME_LEN+1];
   char tmpstr[MAX_LEN+1];
   char valstr[MAXVLENGTH+1];
   static char sep[] = {' ',9,',','=',0};

   err = 0;
   iflag = 2;
   badtok[0] = 0;
   errstr[0] = 0;
   il = -1;
   ll = strlen(line);

   /* Get next token */
   next_token:
   i = kdm_next_field (line, lenl, sep, &p1, &p2, &il);
   if (i==-1) return iflag;
   if (i==0) goto next_token;

   /* See if SYMMETRY  name */

   tmpstr[0] = 0;
   strncat (tmpstr, &line[p1], p2-p1+1);
   err = kdm_parseitem (tmpstr, keyword, &iset, &isubset, &iss);
   kdm_upc(keyword);
   keyword[4] = 0;
   if (strcmp(keyword, "SYMM")!=0)
   {
      err = 1;
   }
   else
   {
      if (chk_func!=0)
      {
         chk_func(&iok);
      }
      else
      {
         iok = 1;
      }
      if (iss==1) err = -1;
      if (iss==2) err = -2;
   }
   if (err<=0)
   {
      iflag = 0;
      if (err<0)
      {
         ncat = p2-p1+1;
         if (ncat>maxb) ncat = maxb;
         if (ncat>0) strncat(badtok, &line[p1], ncat);
         iflag = err;
         tmpstr[0] = 0;
         if (err==-1)
            strcpy (tmpstr, "Invalid or inappropriate set specification");
         if (err==-2)
            strcpy (tmpstr, "Invalid or inappropriate subset specification");
         if (err==-3)
            strcpy (tmpstr, "Syntax error in set/subset specification");
         ncat = strlen (tmpstr);
         if (ncat>maxe) ncat = maxe;
         if (ncat>0) strncat(errstr, tmpstr, ncat);
         return iflag;
      }
      i = kdm_next_field (line, lenl, sep, &p1, &p2, &il);
      if (i<=0)
      {
         if (iok) rdm_reset_symm(0);
         return iflag;
      }
      else
      {
         p2 = ll-1;
         il = ll-1;
         err = 0;
         if (iok)
         {
            valstr[0] = 0;
            strncat (valstr, &line[p1], p2-p1+1); 
            err = rdm_set_symm (valstr, 0, errstr, maxe);
         }
         if (err>0)
         {
            ncat = p2-p1+1;
            if (ncat>maxb) ncat = maxb;
            if (ncat>0) strncat(badtok, &line[p1], ncat);
            iflag = -4;
            return iflag;
         }
      }
   }
   else
   {
     iflag = 1;
     return iflag;
   }
   return iflag;
}
/* Fortran binding: rdmf_parsesym */

void rdmf_parsesym (int * line, int * lenl, void (*chk_func)(),
                   int * badtok,  int * maxb, int * errstr, int *maxe,
                   int *iflag)
{
   int mxb;
   int mxe;

   mxb = -(*maxb);
   if (mxb>0) mxb = -mxb;
   mxe = -(*maxe);
   if (mxe>0) mxe = -mxe;

   *iflag = rdm_parsesym (KDM_pointers[*line], *lenl, chk_func,
                KDM_pointers[*badtok], mxb, KDM_pointers[*errstr], mxe); 
   return;
}





/*-Routine: Set the symmetry - rdm_set_symm
This routine is used to take a string containing the symmetry data
(space group number, name or symmetry operator) and store the
symmetry data internally for access by other RDM symmetry routines.
-end*/

/*
***************************
**     rdm_set_symm      **
***************************

Purpose: Set RDM symmetry

Author:  John W. Campbell, December 1997

CD-Fortran:
         CALL RDMF_SET_SYMM (KDMSTR(VALSTR), LENV, KDMSTR(ERRSTR),
        +                    MAXE, IERR)
CD-end:
*/
/*-Parameters:
VALSTR      c  (R)  The string containing the symmetry data
                    ** Pass address index using the KDMSTR function **
LENV        i  (R)  Length of the value string
ERRSTR      c  (W)  Error message if IERR non zero
                    ** Pass address index using the KDMSTR function **
MAXE        i  (R)  Maximum length for the error string
IERR        i  (W)  The error return flag from rdm_set_symm.
-end*/


/*-C:*/
int rdm_set_symm (char * valstr, int lenv, char * errstr, int maxe)
/*end*/

/*-Parameters:
char * valstr;   The string containing the symmetry data (R)
int lenv;        The length of the value string (may be 0 id
                 a null terminated string) (R)
char *errstr;    Error return string (W)
int maxe;        if >0 Maximum length of returned error string allowed
                       (excluding terminating null - errstr must be at
                       least maxe + 1 characters in length)
                 if <0 abs(maxe) characters will be returned padded
                       with blanks if needed (for 'fortran' use) 
                 (should allow 80 chars for error messages) (R)

Return:  Error flag = 0  no error
                    = 1  symmetry error
-end*/
{
   int i, j, k;
   int ii, jj;
   int err;
   int il;
   int m1, m2;
   int m3, m4;
   int ival;
   int nsym_sav;
   int ilatt;
   float aa[4][4], aainv[4][4];
   char tmpstr[121];
   static int latt[7] ={1,2,3,4,6,5,7};

   kdm_copy_chars (" ", 0, errstr, maxe);

/* Look at next field on line: this can be
      (a) a space-group number
      (b) a space-group name, ie a string beginning P,I,R,F,A,B or C
      (c) a symmetry operation (anything else)
*/ 

   il = -1;
   i = kdm_next_field (valstr, lenv, " ,", &m1, &m2, &il); 
   if (i<=0)
   {
      kdm_copy_chars ("No symmetry data given", 0, errstr, maxe);
      return 1;
   } 
   i = kdm_next_field (valstr, lenv, " ,", &m3, &m4, &il); 

   /* If single token, see if number or name */
   if (i<0)
   {
      tmpstr[0] = 0;
      strncat (tmpstr, &valstr[m1], m2-m1+1);
      i = sscanf (tmpstr,"%d",&ival);
      if (i>0&&ival>0)
      {
         nspgrp = ival;
         spgnam[0] = 0;
      }
      else
      {
         kdm_upc(tmpstr);
         if (strcmp(tmpstr,"END")==0)
	 {
           sym_completed=1;
           goto sym_comp;
         }
         nspgrp = 0;
         spgnam[0] = 0;
         strncat (spgnam, &valstr[m1], m2-m1+1);
         kdm_upc (spgnam);         
      }
      err = rdm_read_symm (nspgrp, spgnam, errstr, maxe);
      if (err!=0)
      {
         nspgrp = 0;
         spgnam[0] = 0;
         pgnam[0] = 0;
         systyp[0] = 0;
         nspgrp = 0;
         nsymp = 0;
         nsym = 0;
         for (i=0;i<192;++i)
         {
            for (j=0;j<4;++j)
            {
               for (k=0;k<4;++k) rsym [i][j][k] = 0.0;
            }
         }
         is_symm = 0;
         ch_symm= sym_ch_mask;
         sym_completed = 0;
         return err;
      }
      sym_completed = 1;
   }
   /* More than one field - interpret symmetry operator */
   else
   {
      nsym_sav = nsym;
      nsym = nsym + 1;
      nsymp = nsym;
      strcpy (tmpstr, valstr);
      kdm_upc(tmpstr);
      err = rdm_symfr2 (tmpstr, 1, &nsym, rsym, errstr, maxe);
      if (err!=0)
      {
         nsym = nsym_sav;
         return err;
      }
      nspgrp = 0;
      spgnam[0] = 0;
      pgnam[0] = 0;
      systyp[0] = 0;
   }
   is_symm = 1;
   ch_symm = sym_ch_mask;
  
   /* Get derived parameters */
   sym_comp:
   if (nsym>0&&sym_completed)
   {
      rdm_pgdefn (nsym, rsym, pgnam, &nsymp);
      for (i=0;i<nsym;++i)
      {
         for (ii=0;ii<4;++ii)
         {
             for (jj=0;jj<4;++jj) aa[ii][jj] = rsym[i][ii][jj];
         }
         rdm_invsym (aa, aainv);
         for (ii=0;ii<4;++ii)
         {
             for (jj=0;jj<4;++jj) rsym_inv[i][ii][jj] = aainv[ii][jj];
         }
      }
      ilatt = latt[icentr-1];
      isys = 0;
      ksys[0] = 0;
      ksys[1] = 0;
      ksys[2] = 0;
      switch (ilatt)
      {
         case 1:
         isys = 0;
         break;

         case 2:
         isys = 2;
         ksys[1] = 1;
         ksys[2] = 1;
         break;

         case 3:
         isys = 2;
         ksys[0] = 1;
         ksys[2] = 1;
         break;

         case 4:
         isys = 2;
         ksys[0] = 1;
         ksys[1] = 1;
         break;

         case 5:
         isys = 2;
         ksys[0] = 1;
         ksys[1] = 1;
         ksys[2] = 1;
         break;

         case 6:
         isys = 4;
         break;

         case 7:
         isys = 3;
         ksys[0] = -1;
         ksys[1] = 1;
         ksys[2] = 1;
         break;
      }
      rdm_pgnlau (pgnam, &nlaue, launam);
      rdm_set_centric();
   }
   return 0;
}
/* Fortran binding: rdmf_set_symm */

void rdmf_set_symm (int * valstr, int * lenv, int * errstr, int *maxe,
                    int *ierr)
{
   int max_err;

   max_err = -(*maxe);
   if (max_err>0) max_err = -max_err;

   *ierr = rdm_set_symm (KDM_pointers[*valstr], *lenv, KDM_pointers[*errstr], 
           max_err);
   return;
}

/*-Routine: Reset the default symmetry - rdm_reset_symm
This routine is used to reset the default (undefined) symmetry
-end*/

/*
***************************
**     rdm_reset_symm    **
***************************

Purpose: Reset default RDM symmetry

Author:  John W. Campbell, December 1997

CD-Fortran:
         CALL RDMF_RESET_SYMM (IDUM)
CD-end:
*/
/*-Parameters:
IDUM        i  (R)  Dummy parameter
-end*/


/*-C:*/
void rdm_reset_symm (int idum)
/*end*/

/*-Parameters:
int idum;        Dummy parameter (R)
-end*/
{
   int i, j, k;
   
   spgnam[0] = 0;
   pgnam[0] = 0;
   systyp[0] = 0;
   nspgrp = 0;
   nsymp = 0;
   nsym = 0;
   for (i=0;i<192;++i)
   {
      for (j=0;j<4;++j)
      {
         for (k=0;k<4;++k) rsym[i][j][k] = 0.0;
      }
   }
   is_symm = 0;
   ch_symm = sym_ch_mask;
   return;
}
/* Fortran binding: rdmf_reset_symm */

void rdmf_reset_symm (int * idum)
{

   rdm_reset_symm (*idum);
   return;
}


/*-Routine: Reset symmetry changed flags - rdm_symm_chreset
This routine is used to reset (cancel) the symmetry data changed flags for
a requested channel or for all channels.
-end*/

/*
***************************
**     rdm_symm_chreset **
***************************

Purpose: Reset default RDM symmetry

Author:  John W. Campbell, December 1997

CD-Fortran:
         CALL RDMF_SYMM_CHRESET (ICHANN, IERR)
CD-end:
*/
/*-Parameters:
ICHAN       i  (R)  Channel number for reset (1-16), 0 = all channels
IERR        i  (W)  =0 OK, =1 Invalid channel parameter
-end*/


/*-C:*/
int rdm_symm_chreset (int ichan)
/*end*/

/*-Parameters:
int ichan;     Channel number for reset (1-16), 0 = all channels (R)

Return =0 OK
       =1 invalid cahnnel number
-end*/
{
   int imask;
   if (ichan<0||ichan>16) return 1;
   if (ichan==0)
   {
      ch_symm = 0;
   }
   else
   {
      imask = 1;
      if (ichan>1) imask <<= ichan-1;
      imask = (!imask)&0xff;
      ch_symm &= imask;
   }
   return 0;
}
/* Fortran binding: rdmf_symm_chreset */

void rdmf_symm_chreset (int * ichan, int * ierr)
{

   *ierr = rdm_symm_chreset (*ichan);
   return;
}


/*-Routine: Test symmetry changed flag - rdm_symm_changed
This routine is used to test the symmetry data changed flag set for
a requested channel.
-end*/

/*
***************************
**     rdm_symm_changed  **
***************************

Purpose: Check if symmetry has been changed

Author:  John W. Campbell, December 1997

CD-Fortran:
         CALL RDMF_SYMM_CHANGED (ICHAN, IFLAG)
CD-end:
*/
/*-Parameters:
ICHAN       i  (R)  Channel number (1-16)
IFLAG       i  (W)  Return flag =0  symmetry not changed
                                =1  symmetry changed
                                =-1 error in channel parameter
-end*/


/*-C:*/
int rdm_symm_changed (int ichan)
/*end*/

/*-Parameters:
int ichan;     Channel number (1-16) (R)

Return =0  not changed
       =1  changed
       -1  invalid parameter
-end*/
{
   int imask;
    
   if (ichan<1||ichan>16) return -1;
   imask = 1;
   if (ichan>1) imask <<= ichan-1;
   if (imask&ch_symm) return 1;
   return 0;
}
/* Fortran binding: rdmf_symm_changed */

void rdmf_symm_changed (int * ichan, int * iflag)
{

   *iflag = rdm_symm_changed (*ichan);
   return;
}
/*-Routine: Get full symmetry data - rdm_get_symm
This routine is used to get the full symmetry detail including such
items as the space group, point group and symmetry operator matrices
-end*/

/*
***********************
**     rdm_get_symm  **
***********************

Purpose: Get symmetry data

Author:  John W. Campbell, December 1997

CD-Fortran:
         CALL RDMF_GET_SYMM (NSPG, KDMSTR(SPGNAME), MAXS,
                             KDMSTR(PGNAME), MAXP, NSYMMP, NSYMM,
     +                       RSM, RSMI, IFLAG)
CD-end:
*/
/*-Parameters:
NSPG     i (W)   Space group number (will be 0 if symmetry operators 
                  were input explicitly.
SPGNAME  c (W)   Space group name (will be blank if symmetry operators 
                 were input explicitly) (10 characters)
                 ** Pass address index using the KDMSTR function **
MAXS     i (R)   Maximum length for returned space group name
PGNAME   c (W)   Point group name (10 characters)
                 ** Pass address index using the KDMSTR function **
MAXP     i (R)   Maximum length for returned point group name
NSYMMP   i (W)   No. of primitive symmetry operators
NSYMM    i (W)   Total number of symmetry operators (0 if symmetry not
                 defined)
RSM()    r (W)   The NSYMM symmetry operators -  dimensioned (4,4,192)
RSMI()   r (W)   The NSYMM inverted symmetry operators - dimensioned 
                 (4,4,192)
IFLAG    i (W)   Flag =  0 default valuen(undefined)
                      =  1 value set explicitly
                      =  2 value set globally
                      = -1 Symmetry not fully set up - symmetry cards with
                           operators given but no 'SYMMETRY END' given
-end*/


/*-C:*/
int rdm_get_symm (int * nspg, char * spgname, int maxs, char * pgname,
                  int maxp, int * nsymmp, int * nsymm, 
                  float rsm[192][4][4], float rsmi[192][4][4])
/*end*/

/*-Parameters:
int * nspg        Space group number (will be 0 if symmetry operators 
                  were input explicitly. (W)
char * spgname;   Space group name (will be blank if symmetry operators 
                  were input explicitly) (10 characters) (W)
int maxs;         Maximum length for returned space group name
                  if >0 Maximum length of returned string allowed
                        (excluding terminating null - spgname must be at
                        least maxs + 1 characters in length)
                  if <0 abs(maxs) characters will be returned padded
                        with blanks if needed (for 'fortran' use) (R)
char * pgname;    Point group name (10 characters) (W)
int maxp;         Maximum length for returned point group name
                  if >0 Maximum length of returned string allowed
                        (excluding terminating null - pgname must be at
                        least maxp + 1 characters in length)
                  if <0 abs(maxp) characters will be returned padded
                        with blanks if needed (for 'fortran' use) (R)
int * nsymmp;     No. of primitive symmetry operators (W)
int * nsymm;      Total number of symmetry operators (0 if symmetry not
                  defined) (W)
float rsm[192][4][4];  The nsymm symmetry operators (W)
float rsmi[192][4][4]; The nsymm inverted symmetry operators (W)

Return  = -0 default value (undefined)
        =  1 value set explicitly
        =  2 value set globally
        = -1 Symmetry not fully set up - symmetry cards with
             operators given but no 'SYMMETRY END' given
-end*/
{
   int n, i, j;

   *nspg = nspgrp;
   kdm_copy_chars (spgnam, 0, spgname, maxs);
   kdm_copy_chars (pgnam, 0, pgname, maxp);
   *nsymmp = nsymp;
   *nsymm = nsym;
   if (nsym>0)
   {
      for (n=0;n<192;++n)
      {
         for (i=0;i<4;++i)
         {
            for (j=0;j<4;++j)
            {
               rsm[n][i][j] = rsym[n][i][j];
               rsmi[n][i][j] = rsym_inv[n][i][j];
            }
         }
      }
   }

   if (!sym_completed) return -1;
   return is_symm;
}
/* Fortran binding: rdmf_get_symm */

void rdmf_get_symm (int * nspgrp, int * spgname, int * maxs,
                    int * pgname, int * maxp, int * nsymmp,
                    int * nsymm, float * rsm, float * rsmi, 
                    int * iflag)
{
   int n,i,j,k;
   int mxs;
   int mxp;
   float rs_tmp[192][4][4];
   float rsi_tmp[192][4][4];

   mxs = - *maxs;
   if (mxs>0) mxs = - mxs;
   mxp = - *maxp;
   if (mxp>0) mxp = - mxp;

   *iflag = rdm_get_symm (nspgrp, KDM_pointers[*spgname], mxs,
                          KDM_pointers[*pgname], mxp,
                          nsymmp, nsymm, rs_tmp, rsi_tmp);
   k = 0;
   for (n=0;n<192;++n)
   {
      for (i=0;i<4;++i)
      {
         for (j=0;j<4;++j)
         {
            rsm[k] = rs_tmp[n][j][i];
            rsmi[k] = rsi_tmp[n][j][i];
            k++;
         }
      }
   }
   return;
}
/*-Routine: Get basic space group details - rdm_get_spgrp
This routine renturns where possible the space group number and name and
the number of primitive and total number of symmetry operators.
-end*/

/*
************************
**     rdm_get_spgrp  **
************************

Purpose: Get symmetry space group data

Author:  John W. Campbell, December 1997

CD-Fortran:
         CALL RDMF_GET_SPGRP (NSPG, KDMSTR(SPGNAME), MAXS,
                              NSYMMP, NSYMM, IFLAG)
CD-end:
*/
/*-Parameters:
NSPG     i (W)   Space group number (will be 0 if symmetry operators 
                  were input explicitly.
SPGNAME  c (W)   Space group name (will be blank if symmetry operators 
                 were input explicitly) (10 characters)
                 ** Pass address index using the KDMSTR function **
MAXS     i (R)   Maximum length for returned space group name
NSYMMP   i (W)   No. of primitive symmetry operators
NSYMM    i (W)   Total number of symmetry operators (0 if symmetry not
                 defined)
IFLAG    i (W)   0 OK, -1 symmetry input not fully set up - symmetry 
                 cards with operators given but no 'SYMMETRY END' given
-end*/


/*-C:*/
int rdm_get_spgrp (int * nspg, char * spgname, int maxs,  
                    int * nsymmp, int * nsymm)
/*end*/

/*-Parameters:
int * nspg        Space group number (will be 0 if symmetry operators 
                  were input explicitly. (W)
char * spgname;   Space group name (will be blank if symmetry operators 
                  were input explicitly) (10 characters) (W)
int maxs;         Maximum length for returned space group name
                  if >0 Maximum length of returned string allowed
                        (excluding terminating null - spgname must be at
                        least maxs + 1 characters in length)
                  if <0 abs(maxs) characters will be returned padded
                        with blanks if needed (for 'fortran' use) (R)
int * nsymmp;     No. of primitive symmetry operators (W)
int * nsymm;      Total number of symmetry operators (0 if symmetry not
                  defined) (W)

Return  =  0 OK
        = -1 Symmetry not fully set up - symmetry cards with
             operators given but no 'SYMMETRY END' given
-end*/
{
   *nspg = nspgrp;
   kdm_copy_chars (spgnam, 0, spgname, maxs);
   *nsymmp = nsymp;
   *nsymm = nsym;
   if (!sym_completed) return -1;
   return is_symm;
}
/* Fortran binding: rdmf_get_spgrp */

void rdmf_get_spgrp (int * nspgrp, int * spgname, int * maxs,
                    int * nsymmp, int * nsymm, int * iflag)
{
   int mxs;

   mxs = - *maxs;
   if (mxs>0) mxs = - mxs;

   *iflag = rdm_get_spgrp (nspgrp, KDM_pointers[*spgname], mxs,
                          nsymmp, nsymm);
}
/*-Routine: Get number of symmetry operators - rdm_get_nsym
This routine returns the number of symmetry operators or 0 if no symmetry
has been defined. This provides the simplest way of determining whether
the symmetry has been defined.         
-end*/

/*
************************
**     rdm_get_nsym   **
************************

Purpose: Get symmetry space group data

Author:  John W. Campbell, December 1997

CD-Fortran:
         CALL RDMF_GET_NSYM (NSYMM, IFLAG)
CD-end:
*/
/*-Parameters:
NSYMM    i (W)   Total number of symmetry operators (0 if symmetry not
                 defined)
IFLAG    i (W)   Flag =  0 default (undefined) value
                      =  1 value set explicitly
                      = -1 Symmetry not fully set up - symmetry cards 
                           with operators given but no 'SYMMETRY END' given
-end*/


/*-C:*/
int rdm_get_nsym (int * nsymm)
/*end*/

/*-Parameters:
int * nsymm;      Total number of symmetry operators (0 if symmetry not
                  defined) (W)

Return  =  0 default (undefined) value
        =  1 value set explicitly
        = -1 Symmetry not fully set up - symmetry cards with
             operators given but no 'SYMMETRY END' given

-end*/
{
   *nsymm = nsym;
   if (!sym_completed) return -1;
   return is_symm;
}
/* Fortran binding: rdmf_get_nsym */

void rdmf_get_nsym (int * nsymm, int * iflag)
{
   *iflag = rdm_get_nsym (nsymm);
}
/*-Routine: Get the crystal system and lattice type  - rdm_get_sylat
This routine returns the crystal system type and lattice
type if the symmetry has been defined via the space group number or
name.        
-end*/

/*
*************************
**     rdm_get_sylat   **
*************************

Purpose: Get the crystal system and lattice type

Author:  John W. Campbell, December 1997

CD-Fortran:
         CALL RDMF_GET_SYLAT (KDMSTR(SYST), MAXS, KDMSTR(LATT), MAXL)
CD-end:
*/
/*-Parameters:
SYST     c (W)  Crystal system (blank if symmetry not defined or 
                defined using a set of operators and not via the 
                space group number or name). Allow minimum of 12 
                characters. Name wlll be returned in upper case. 
                e.g. TRICLINIC, MONOCLINIC, ORTHORHOMBIC, 
                     TETRAGONAL, TRIGONAL, HEXAGONAL, CUBIC
MAXS     i (R)  Maximum no. of characters to be returned in SYST
LATT     c (W)  Lattice type (blank if symmetry not defined or 
                defined using a set of operators and not via the 
                space group number or name). Allow minimum of 1 
                characters. Name wlll be returned in upper case.
                e.g. P, A, B, C, I, F, R 
MAXL     i (R)  Maximum no. of characters to be returned in LATT
-end*/


/*-C:*/
void rdm_get_sylat (char * syst, int maxs, char * latt, int maxl)
/*end*/

/*-Parameters:
char * syst;    Crystal system (blank if symmetry not defined or 
                defined using a set of operators and not via the 
                space group number or name). Allow minimum of 12 
                characters. Name wlll be returned in upper case. 
                e.g. TRICLINIC, MONOCLINIC, ORTHORHOMBIC, 
                     TETRAGONAL, TRIGONAL, HEXAGONAL, CUBIC (W)
int maxs;       Maximum length for returned system name
                if >0 Maximum length of returned string allowed
                      (excluding terminating null - syst must be at
                      least maxs + 1 characters in length)
                if <0 abs(maxs) characters will be returned padded
                      with blanks if needed (for 'fortran' use) (R)
char * latt;    Lattice type (blank if symmetry not defined or 
                defined using a set of operators and not via the 
                space group number or name). Allow minimum of 1 
                characters. Name wlll be returned in upper case.
                e.g. P, A, B, C, I, F, R (W)
int maxl;       Maximum length for returned lattice type
                if >0 Maximum length of returned string allowed
                      (excluding terminating null - latt must be at
                      least maxl + 1 characters in length)
                if <0 abs(maxl) characters will be returned padded
                      with blanks if needed (for 'fortran' use) (R)
-end*/
{
   char tmpstr[21];

   kdm_copy_chars (" ", 0, syst, maxs);
   kdm_copy_chars (" ", 0, latt, maxl);
   if (nspgrp==0) return;
   tmpstr[0] = 0;
   strcat (tmpstr, systyp);
   kdm_upc(tmpstr);
   kdm_copy_chars (tmpstr, 0, syst, maxs);
   tmpstr[0] = 0;
   strncat (tmpstr, spgnam, 1);
   kdm_upc(tmpstr);
   kdm_copy_chars (tmpstr, 0, latt, maxl);
}
/* Fortran binding: rdmf_get_sylat */

void rdmf_get_sylat (int * syst, int * maxs, int * latt, int * maxl)
{
   int mxs;
   int mxl;

   mxs = - *maxs;
   if (mxs>0) mxs = - mxs;
   mxl = - *maxl;
   if (mxl>0) mxl = - mxl;
   rdm_get_sylat (KDM_pointers[*syst], mxs, KDM_pointers[*latt], mxl);
}
/*-Routine: Get string from a symmetry operator - rdm_symm_str
This routine returns the a string from a symmetry operator in
an International Tables type form. The routine may be used when
listing the symmetry or outputting symmetry operators as part
or an RDM data set.         
-end*/

/*
************************
**     rdm_symm_str   **
************************

Purpose:  Get string from a symmetry operator

Author:  John W. Campbell, December 1997

CD-Fortran:
         CALL RDMF_SYMM_STR (ISYM, KDMSTR(STR), MAXS)
CD-end:
*/
/*-Parameters:
ISYM     i (R)   The symmetry number (1 to NSYM)
STR      c (W)   Returns the symmetry operator string
                 ** Pass address index using the KDMSTR function **
MAXS     c (R)   Maximum length allowed for returned string
-end*/


/*-C:*/
void rdm_symm_str (int isym, char * str, int maxs)
/*end*/

/*-Parameters:
int isym;         The symmetry number (1 to nsym) (R)
char * str;       Returns the symmetry operator string (W)
int maxs;         Maximum length for returned symmetry string
                  if >0 Maximum length of returned string allowed
                        (excluding terminating null - syst must be at
                        least maxs + 1 characters in length)
                  if <0 abs(maxs) characters will be returned padded
                        with blanks if needed (for 'fortran' use) (R)
-end*/
{
   int i, j;
   int ist;
   int itr;
   static char axis[] = {'X','Y','Z'};
   static char numb[] = {'1','2','3','4','5','6','7','8','9'};
   static int npntr1[] = {0,1,1,1,0,1,0,2,3,5};
   static int npntr2[] = {0,6,4,3,0,2,0,3,4,6};
   char tmpstr[121];

   kdm_copy_chars (" ", 0, str, maxs);
   if (isym<1||isym>nsym) return;
   tmpstr[0] = 0;
   for (j=0;j<3;++j)
   {
      ist = 0;
      for (i=0;i<4;++i)
      {
         if (rsym[isym-1][j][i]!=0.0)
         {
            if (rsym[isym-1][j][i]>0.0&&ist>0)
            {
               strcat(tmpstr,"+");
            }

            if (rsym[isym-1][j][i]<0.0)
            {
               strcat(tmpstr,"-");
               ist = 1;
            }
        
            if (i!=3)
            {
               strncat(tmpstr,&axis[i],1);
               ist = 1;
            }

            if (i==3&&rsym[isym-1][j][3]!=0.0)
            {
               itr = nint(fabsf(rsym[isym-1][j][3])*12.0);
               strncat(tmpstr, &numb[npntr1[itr-1]-1], 1);
               strcat(tmpstr,"/");
               strncat(tmpstr, &numb[npntr2[itr-1]-1], 1);
            }
         }
             
      }

      if (j!=2) strcat(tmpstr,", ");
   }   
   kdm_copy_chars (tmpstr, 0, str, maxs);
   return;
}
/* Fortran binding: rdmf_symm_str */

void rdmf_symm_str (int * isym, int * str, int * maxs)
{
   int mxs;

   mxs = - *maxs;
   if (mxs>0) mxs = -mxs;
   rdm_symm_str (*isym, KDM_pointers[*str], mxs);
}

/*-Routine:  List symmetry data - rdm_symm_list
This routine lists the symmetry data including the space group number,
name and point group (if known) and the symmetry operators. The user 
must supply a function to output a string to the required output stream
or file.
-end*/

/*
***************************
**     rdm_symm_list     **
***************************

Purpose:  List symmetry information

Author:  John W. Campbell, December 1997

CD-Fortran:
         CALL RDMF_SYMM_LIST (KDMSTR(OUTSTR), MAXO, OPFUN)
CD-end:
*/
/*-Parameters:
OUTSTR    c  (*) Buffer used in printing output (should be at least
                 120 characters in length)
                 ** Pass address index using the KDMSTR function **
MAXO      i  (R) Maximum length allowed for OUTSTR (should
                 be at least 120)
OP_FUNC   s  (R) Function to do the actual output. The data to be
                 output by this function will be stored in OUTSTR,

                 The call will be CALL OP_FUNC without any parameters
                 so that the string OUTSTR must be available to the
                 routine op_func via a common block possibly with
                 other parameters such as Fortran unit numbers etc.
-end*/


/*-C:*/
void rdm_symm_list (char * outstr, int maxo, void (*op_func)())
/*end*/

/*-Parameters:
char * outstr;       Returns the output string for printing (W)
int maxo;            if >0 Maximum length of returned output string allowed
                           (excluding terminating null - outstr must be at
                           least maxo + 1 characters in length); maxo
                           should be at least 2050.
                     if <0 abs(maxo) characters will be returned padded
                           with blanks if needed (for 'fortran' use) 
                           (abs(maxo) should be at least 2050) (R)
void (*op_func)();   Function to o/p the string; if NULL then o/p will be
                     done to the standard output. If defined then this
                     function will be called without parameters; the
                     string 'outstr' will contain the required o/p and
                     must be globally accessible to the output function
                     together with any other special parameters required.
                     (R)
-end*/
{
   int i;
   char linout[121];

   sprintf (linout, 
            "Space group number: %d, Space group name: %s, Point group: %s",
            nspgrp, spgnam, pgnam);
   rdm_opstr(linout, outstr, maxo, op_func);
   rdm_opstr(" ", outstr, maxo, op_func);
   sprintf (linout,"Number of symmetry operators: primitive=%d, total=%d",
            nsymp, nsym);
   rdm_opstr(linout, outstr, maxo, op_func);
   rdm_opstr(" ", outstr, maxo, op_func);

   if (nsym>0)
   {
      for (i=0;i<nsym;++i)
      {
         rdm_symm_str(i+1, linout, 120);
         rdm_opstr(linout, outstr, maxo, op_func);
      } 
      if (!sym_completed)
      {
        rdm_opstr("Symmetry input not yet completed",outstr, maxo, op_func);
      }
      rdm_opstr(" ", outstr, maxo, op_func);
   }
   
   return;
}
/* Fortran binding: rdmf_symm_list */

void rdmf_symm_list (int * outstr, int *maxo, void (*op_func)())
{
   int mxo;

   mxo = - *maxo;
   if (mxo>0) mxo = -mxo;
   rdm_symm_list (KDM_pointers[*outstr], mxo, op_func);
}

/*-Routine:  Write symmetry data keyword data - rdm_symm_write
This routine writes out the symmetry keyword data. The user 
must supply a function to output a string to the required output stream
or file.
-end*/

/*
***************************
**     rdm_symm_write    **
***************************

Purpose:  List symmetry information

Author:  John W. Campbell, December 1997

CD-Fortran:
         CALL RDMF_SYMM_WRITE (KDMSTR(OUTSTR), MAXO, OPFUN)
CD-end:
*/
/*-Parameters:
OUTSTR    c  (*) Buffer used in printing output (should be at least
                 120 characters in length)
                 ** Pass address index using the KDMSTR function **
MAXO      i  (R) Maximum length allowed for OUTSTR (should
                 be at least 120)
OP_FUNC   s  (R) Function to do the actual output. The data to be
                 output by this function will be stored in OUTSTR,

                 The call will be CALL OP_FUNC without any parameters
                 so that the string OUTSTR must be available to the
                 routine op_func via a common block possibly with
                 other parameters such as Fortran unit numbers etc.
-end*/


/*-C:*/
void rdm_symm_write (char * outstr, int maxo, void (*op_func)())
/*end*/

/*-Parameters:
char * outstr;       Returns the output string for printing (W)
int maxo;            if >0 Maximum length of returned output string allowed
                           (excluding terminating null - outstr must be at
                           least maxo + 1 characters in length); maxo
                           should be at least 2050.
                     if <0 abs(maxo) characters will be returned padded
                           with blanks if needed (for 'fortran' use) 
                           (abs(maxo) should be at least 2050) (R)
void (*op_func)();   Function to o/p the string; if NULL then o/p will be
                     done to the standard output. If defined then this
                     function will be called without parameters; the
                     string 'outstr' will contain the required o/p and
                     must be globally accessible to the output function
                     together with any other special parameters required.
                     (R)
-end*/
{
   int i;
   char linout[121];
   char tmpstr[121];

   if (nsym==0)
   {
      strcpy (linout, "SYMMETRY");
      rdm_opstr(linout, outstr, maxo, op_func);
      return;
   }

   if (nspgrp>0||strlen(spgnam)>0)
   {
      if (nspgrp>0)
      {
         sprintf (linout, "SYMMETRY %d", nspgrp);
      }
      else
      {
         sprintf (linout, "SYMMETRY %s", spgnam);
      }
      rdm_opstr(linout, outstr, maxo, op_func);
   }
   else
   {
      if (nsym>0)
      {

         for (i=0;i<nsym;++i)
         {
            rdm_symm_str(i+1, tmpstr, 120);
            sprintf (linout, "SYMMETRY %s", tmpstr); 
            rdm_opstr(linout, outstr, maxo, op_func);
         } 
         if (sym_completed) rdm_opstr("SYMMETRY END", outstr, maxo, op_func); 
      }
   }
   return;
}
/* Fortran binding: rdmf_symm_write */

void rdmf_symm_write (int * outstr, int *maxo, void (*op_func)())
{
   int mxo;

   mxo = - *maxo;
   if (mxo>0) mxo = -mxo;
   rdm_symm_write (KDM_pointers[*outstr], mxo, op_func);
}

/*-Routine: Check for systematic absence - rdm_sysabs
This routine checks whether a given reflection is a systematic absence
based on the symmetry information available. Its initial checks are
based on the lattice type but it will then use the full symmetry
information if this has been defined.
-end*/

/*
************************
**     rdm_sysabs     **
************************

Purpose:  Check for systematic absence

Author:  John W. Campbell, December 1997

CD-Fortran:
         CALL RDMF_SYSABS (IH, IK, IL, IABS)
CD-end:
*/
/*-Parameters:
IH       i (R)   'h' index
IK       i (R)   'k' index
IL       i (R)   'l' index
IABS     i (W)   Return 1 if absent, 0 if present
-end*/


/*-C:*/
int rdm_sysabs (int ih, int ik, int il)
/*end*/

/*-Parameters:
int ih;    'h' index (R)
int ik;    'k' index (R)
int ik;    'l' index (R)

Return  1 systematic absence
        0 spot present
-end*/
{
   int i, j;
   int absnt;
   int h[3];
   int jh, jk, jl;
   float phas;
   float err;

   absnt = 0;
   if (isys==0) goto lab100;
   h[0] = ih;
   h[1] = ik;
   h[2] = il;
   if (isys==4)
   {
 
      absnt = 1;
      for (i=0;i<2;++i)
      {
	  if ((h[i]+h[i+1])%2!=0) return absnt;
      }
      absnt = 0;
      goto lab100;
   }
   else

   /* This version for a,b,c,r,i & f lattice absences */

   {
      absnt = 0;
      j = 0;
      for (i=0;i<3;++i)
      {
         j = j + h[i]*ksys[i];
      }
      if (j%isys!=0)
      {
         absnt = 1;
         return absnt;
      }
   }

   /* Additional full check for further absences if full symmetry defined */

   lab100:
   if (nsym<2) return absnt;
   if (ih!=0)
   {
      if (ik!=0)
      {
         if (il!=0)
         {
            if (ih!=ik) return absnt;
            if (ih!=il) return absnt;
         }
      }
   }

   /* Test whether h' k' l' equals h k l */

   for (j=1;j<nsym;++j)
   {
      jh = ih*rsym_inv[j][0][0]+ik*rsym_inv[j][1][0]+il*rsym_inv[j][2][0];
      if (jh==ih)
      {
         jk = ih*rsym_inv[j][0][1]+ik*rsym_inv[j][1][1]+il*rsym_inv[j][2][1];
         if (jk==ik)
         {
            jl = ih*rsym_inv[j][0][2]+ik*rsym_inv[j][1][2]+il*rsym_inv[j][2][2];
            if (jl==il)
            {

            /*  Test whether this Symmetry equivalent has a different phase to
                IH, IK, IL - if so it is a systematic absence 
                (... believe me, EJD) */

               phas = ih*rsym[j][0][3] + ik*rsym[j][1][3] + il*rsym[j][2][3];
               err = fabsf(phas-nint(phas));
               if (err>0.05)
               {
                  absnt = 1;
                  return absnt;
               }
            }
         }
      }
   }
   return absnt;   
}
/* Fortran binding: rdmf_sysabs */

void rdmf_sysabs (int * ih, int * ik, int * il, int * iabs)
{
   *iabs = rdm_sysabs (*ih, *ik, *il);
}

/*-Routine: Check for systematic absence - rdm_sysahk
This is a variant on rdm_sysabs and checks to see if a given 'h', 'k'
reflection is absence regardless of 'l'. It only uses the lattice type
and not the full symmetry.
-end*/

/*
************************
**     rdm_sysahk     **
************************

Purpose:  Variant of rdm_sysabs checks to see if a given h,k is absent whatever
C          the l value (using only lattice type)

Author:  John W. Campbell, December 1997

CD-Fortran:
         CALL RDMF_SYSAHK (IH, IK, IL, IABS)
CD-end:
*/
/*-Parameters:
IH       i (R)   'h' index
IK       i (R)   'k' index
IL       i (R)   'l' index
IABS     i (W)   Return 1 if absent, 0 if present
-end*/


/*-C:*/
int rdm_sysahk (int ih, int ik, int il)
/*end*/

/*-Parameters:
int ih;    'h' index (R)
int ik;    'k' index (R)
int ik;    'l' index (R)

Return  1 systematic absence
        0 spot present
-end*/
{
   int absnt;

   absnt = 0;
   if (isys==0) return absnt;

   if ((isys==2||isys==4)&&ksys[2]==0)
   {
      if ((ih+ik)%2!=0) absnt = 1;
   }
   return absnt;
   
}
/* Fortran binding: rdmf_sysahk */

void rdmf_sysahk (int * ih, int * ik, int * il, int * iabs)
{
   *iabs = rdm_sysahk (*ih, *ik, *il);
}

/*-Routine: Put reflection into asymmetric unit - rdm_asuput
This uses internally stored and derived rdm symmetry data to put
a requested reflection into a unique asymmetric usint for the
space group.
-end*/

/*
************************
**     rdm_asuput     **
************************

Purpose: Put reflection into asymmetric unit

Author:  John W. Campbell, December 1997

CD-Fortran:
         CALL RDMF_ASUPUT (IHKL, JHKL, ISYM, IERR)
CD-end:
*/
/*-Parameters:
IHKL(3)  i (R)   Input reflection indices
JHKL(3)  i (W)   Output (unique) reflection indices
ISYM     i (W)   Symmetry number used odd for I+, even for I-
IERR     i (W)   =0 OK, =1 cannot put reflection into asymmetric unit,
                 =2 symmetry input not completed
-end*/


/*-C:*/
int rdm_asuput (int *ihkl, int *jhkl, int *isym)
/*end*/

/*-Parameters:
int ihkl[3];  Input reflection indices (R)
int jhkl[3];  Output (unique) reflection indices (W)
int * isym;   Symmetry number used odd for I+, even for I- (W)

Return =0 OK
       =1 Error - cannot put reflection into asymmetric unit
       =2 Symmetry input not completed
-end*/
{
   int l;
   int i;
   int isgn;

   jhkl[0] = 0;  
   jhkl[1] = 0;  
   jhkl[2] = 0;  
   *isym = 0;
   if (!sym_completed) return 2;


   for (l=0;l<nsymp;++l)
   {
      for (i=0;i<3;++i)
      {
         jhkl[i] = ihkl[0]*rsym[l][0][i] + ihkl[1]*rsym[l][1][i]
                 + ihkl[2]*rsym[l][2][i];  
      }
      isgn = rdm_inasu (jhkl, nlaue);
      if (isgn!=0) goto lab100;
   }
   return 1;

   lab100:
   for (i=0;i<3;++i)
   {
      jhkl[i] = isgn*jhkl[i];
   }
   if (isgn>0)
   {
      *isym = 2*l+1;
   }
   else
   {
      *isym = 2*l + 2;
   }
   return 0;
}
/* Fortran binding: rdmf_asuput */

void rdmf_asuput (int * ihkl, int * jhkl, int * isym, int * ierr)
{
   *ierr = rdm_asuput (ihkl, jhkl, isym);
}

/*-Routine: Get original indices - rdm_asuget
This uses internally stored and derived rdm symmetry data to restore
the original indices of a reflection from the unique indices
and symmetry number e.g. as determined by rdm_asuput.
-end*/

/*
************************
**     rdm_asuget     **
************************

Purpose: Get original indices

Author:  John W. Campbell, December 1997

CD-Fortran:
         CALL RDMF_ASUGET (IHKL, ISYM, JHKL, IERR)
CD-end:
*/
/*-Parameters:
IHKL(3)  i (R)   Input (unique)reflection indices
ISYM     i (R)   Symmetry number used odd for I+, even for I-
JHKL(3)  i (W)   Output (measured) reflection indices
IERR     i (W)   Error flag =0 OK, =1 out of range symmetry number,
                 =2 symmetry input not completed
-end*/


/*-C:*/
int rdm_asuget (int *ihkl, int isym,  int *jhkl)
/*end*/

/*-Parameters:
int ihkl[3];  Input (unique) reflection indices (R)
int isym;     Symmetry number used odd for I+, even for I- (R)
int jhkl[3];  Output (measured) reflection indices (W)

Return =0 OK
       =1 Out of range symmetry number
       =2 Symmetry input not completed
-end*/
{
   int l;
   int i;
   int isgn;


   jhkl[0] = 0;  
   jhkl[1] = 0;  
   jhkl[2] = 0;  
   if (!sym_completed) return 2;
   l = (isym+1)/2;
   if (l<1||l>nsymp) return 1;
   isgn = isym%2;
   if (isgn==0) isgn = -1;
   for (i=0;i<3;++i)
   {
       jhkl[i] = ihkl[0]*rsym_inv[l-1][0][i] + ihkl[1]*rsym_inv[l-1][1][i]
               + ihkl[2]*rsym_inv[l-1][2][i];
   }
   for (i=0;i<3;++i)
   {
      jhkl[i] = jhkl[i]*isgn;
   }
   return 0;
}
/* Fortran binding: rdmf_asuget */

void rdmf_asuget (int * ihkl, int *isym,  int * jhkl, int * ierr)
{
   *ierr = rdm_asuget (ihkl, * isym, jhkl);
}

/*-Routine: Test for centric reflection - rdm_centr
This uses internally stored and derived rdm symmetry data to
determine whether or not a requested reflection is centric.
-end*/

/*
************************
**     rdm_centr      **
************************

Purpose: Test for centric reflection

Author:  John W. Campbell, December 1997

CD-Fortran:
         CALL RDMF_CENTR (IHKL, IC, IERR)
CD-end:
*/
/*-Parameters:
IHKL(3)  i (R)   Input reflection indices
IC       i (W)   Centric flag =1 yes, =0 no
IERR     i (W)   Error flag =0 OK, =1 symmetry not properly set up,
                 =2 symmetry input not completed
-end*/


/*-C:*/
int rdm_centr (int *ihkl, int *ic)
/*end*/

/*-Parameters:
int ihkl[3];  Input  reflection indices (R)
int *ic;      Centric flag =1 yes, =0 no (W)
Return =0 OK  
       =1 symmetry not properly set up
       =2 Symmetry input not completed
       
-end*/
{
   int i;

   *ic = 0;
   if (!sym_completed) return 2;
   if (!cent_setup) return 1;
   if (ncent>0)
   {
      for (i=0;i<ncent;++i)
      {
         if ((cproj[i][0]*ihkl[0]+cproj[i][1]*ihkl[1]+
            cproj[i][2]*ihkl[2])==0.0) goto lab20;
      }
   }
   return 0;
   lab20:
   *ic = 1;
   return 0;
}
/* Fortran binding: rdmf_centr */

void rdmf_centr (int * ihkl, int * ic, int * ierr)
{
   *ierr = rdm_centr (ihkl, ic);
}


/*-Section: Get RDM data items
All the RDM data item values may be retrieved via the general KDM handling
routine kdm_getvalue (kdmf_getvalue). For convenience some 
additioinal routines have been written to retriev certain groups
of parameters through a single call.
-end*/

/*-Routine: Get RDM rotations parameters - rdm_getrotations
The routine rdm_getrotations (rdmf_getrotations) returns the values
of the rotation ranges and oscillation angle for a requested
crystal set. In addition the number of images is calculated
from these parameters and returned.
-end*/

/*
*****************************
**     rdm_getrotations    **
*****************************

Purpose: Get RDM rotation ranges parameters

Author:  John W. Campbell, November 1997

CD-Fortran:
         CALL RDMFGET_ROTATIONS (KINDX, ISET, ROTSTART, ROTEND,
                                 ROTS2, ROTE2, ROTS3, ROTE3,
                                 ROTS4, ROTE4, ROTS5, ROTE5,
                                 OSCANG, NIMG, NUMIMG, IERR)
CD-end:
*/
/*-Parameters:
KINDX       i  (R)  The KDM index for the RDM dataset.
ISET        i  (R)  The crystal set no. (>0)
ROTSTART    r  (W)  Returns the rotation start angle (range 1)
ROTEND      r  (W)  Returns the rotation end angle (range 1)
ROTS2       r  (W)  Returns the rotation start angle (range 2)
ROTE2       r  (W)  Returns the rotation end angle (range 2)
ROTS3       r  (W)  Returns the rotation start angle (range 3)
ROTE3       r  (W)  Returns the rotation end angle (range 3)
ROTS4       r  (W)  Returns the rotation start angle (range 4)
ROTE4       r  (W)  Returns the rotation end angle (range 4)
ROTS5       r  (W)  Returns the rotation start angle (range 5)
ROTE5       r  (W)  Returns the rotation end angle (range 5)
OSCANG      r  (W)  Returns the oscillation angle
NIMG(5)     i  (W)  Returns number of images in each range
NUMIMG      i  (W)  Returns the number of images
IERR        i  (W)  The error return flag from rdm_getrotations.
-end*/

/*-C:*/
int rdm_getrotations (int kindx, int iset, float * rotstart, float * rotend,
                      float * rots2, float * rote2,
                      float * rots3, float * rote3,
                      float * rots4, float * rote4,
                      float * rots5, float * rote5,
                      float * oscang, int nimg[], int * numimg)
/*end*/

/*-Parameters:
int kindx;        The KDM index for the RDM dataset (R)
int iset;         The crystal set no. (>0) (R)
float * rotstart; Returns the rotation start angle (range 1) (W)
float * rotend;   Returns the rotation end  angle (range 1) (W)
float * rots2;    Returns the rotation start angle (range 2) (W)
float * rote2;    Returns the rotation end  angle (range 2) (W)
float * rots3;    Returns the rotation start angle (range 3) (W)
float * rote3;    Returns the rotation end  angle (range 3) (W)
float * rots4;    Returns the rotation start angle (range 4) (W)
float * rote4;    Returns the rotation end  angle (range 4) (W)
float * rots5;    Returns the rotation start angle (range 5) (W)
float * rote5;    Returns the rotation end  angle (range 5) (W)
float * oscang;   Returns the oscillation angle (W)
int nimg[5];      Returns the no. of images in each of the 5 ranges (W)
int * numimg;     Returns the number of images (W)

Return:   Error flag =0 OK
                     =1 error
-end*/
{
   int err;
   int i;
   int num_inp;
   int istat;
   int ityp;
   int ival[10];
   int nd[10];
   float fval[10];
   float rots[5];
   float rote[5];
   float x;
   char valstr[81];
   char errstr[81];
   
   err = kdm_getvalue (kindx, "ROTSTART", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &ityp, 
                       errstr, 80);
   *rotstart = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "ROTEND", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &ityp, 
                       errstr, 80);
   *rotend = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "ROTS2", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &ityp, 
                       errstr, 80);
   *rots2 = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "ROTE2", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &ityp, 
                       errstr, 80);
   *rote2 = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "ROTS3", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &ityp, 
                       errstr, 80);
   *rots3 = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "ROTE3", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &ityp, 
                       errstr, 80);
   *rote3 = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "ROTS4", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &ityp, 
                       errstr, 80);
   *rots4 = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "ROTE4", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &ityp, 
                       errstr, 80);
   *rote4 = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "ROTS5", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &ityp, 
                       errstr, 80);
   *rots5 = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "ROTE5", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &ityp, 
                       errstr, 80);
   *rote5 = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "OSCANGLE", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &ityp, 
                       errstr, 80);
   *oscang = fval[0];
   if (err!=0) return 1;

   rots[0] = *rotstart;
   rote[0] = *rotend;
   rots[1] = *rots2;
   rote[1] = *rote2;
   rots[2] = *rots3;
   rote[2] = *rote3;
   rots[3] = *rots4;
   rote[3] = *rote4;
   rots[4] = *rots5;
   rote[4] = *rote5;

   *numimg = 0;
   for (i=0;i<5;++i)
   {
      nimg[i] = 0;
      if(rote[i]<=rots[i]) continue;
      x = (rote[i]-rots[i])/(*oscang)-0.001;
      nimg[i] = (int) (floorf(x)) + 1;
      *numimg += (int) nimg[i];
   }

   return 0;
}
/* Fortran binding: rdmf_getrotations */

void rdmf_getrotations (int * kindx, int * iset, float * rotstart,
                        float * rotend, float * rots2, float * rote2,
                        float * rots3, float * rote3,
                        float * rots4, float * rote4,
                        float * rots5, float * rote5,
                        float * oscang, int * nimg, 
                        int * numimg, int *ierr)
{

   *ierr = rdm_getrotations (*kindx, *iset, rotstart, rotend,
                             rots2, rote2, rots3, rote3,
                             rots4, rote4, rots5, rote5,
                             oscang, nimg, numimg);
   return;
}
/*-Routine: Get no. of images in a set - rdm_numimg
The routine rdm_numimg (rdmf_numimg) returns the number of images
in a set based on the currently defined rotation ranges.
-end*/
/*
*************************
**     rdm_numimg      **
*************************

Purpose: Get no of images in a set

Author:  John W. Campbell, February 1998

CD-Fortran:
         CALL RDMF_NUMIMG(KINDX, ISET, NUMIMG, IERR)
CD-end:
*/
/*-Parameters:
KINDX       i  (R)  The KDM index for the RDM dataset.
ISET        i  (R)  The crystal set no. (>0)
NUMIMG      i  (W)  Returns the number of images in the set
IERR        i  (W)  Error return flag from rdm_numimg
-end*/

/*-C:*/
int rdm_numimg (int kindx, int iset, int * numimg) 
/*end*/

/*-Parameters:
int kindx;        The KDM index for the RDM dataset (R)
int iset;         The crystal set no. (>0) (R)
int *numimg;      Returns the number of images in the set (W)

Return:  Error flag =0 OK
                    =1 error
-end*/
{
   int err;
   int nimg[5];
   float rotstart;
   float rotend;
   float rots2;
   float rote2;
   float rots3;
   float rote3;
   float rots4;
   float rote4;
   float rots5;
   float rote5;
   float oscang;

   err = rdm_getrotations (kindx, iset, &rotstart, &rotend,
                           &rots2, &rote2,
                           &rots3, &rote3,
                           &rots4, &rote4,
                           &rots5, &rote5,
                           &oscang, nimg, numimg);
   return err;
}
/* Fortran binding: rdmf_numimg */

void rdmf_numimg (int * kindx, int * iset, int *numimg, int * ierr)
{
   *ierr = rdm_numimg (*kindx, *iset, numimg);
}
/*-Routine: Get rotation start for an image - rdm_rot_start
The routine rdm_rot_start(rdmf_rot_start) returns the rotation start
angle for a requested image in a set.
-end*/

/*
*************************
**     rdm_rot_start   **
*************************

Purpose: Get rotation start angle for an image

Author:  John W. Campbell, February 1998

CD-Fortran:
         CALL RDMF_ROT_START(KINDX, ISET, IMGNUM, ROT_START, IERR)
CD-end:
*/
/*-Parameters:
KINDX       i  (R)  The KDM index for the RDM dataset.
ISET        i  (R)  The crystal set no. (>0)
IMGNUM      i  (R)  The image number within the set
ROT_START   r  (W)  Returns the rotation start angle
IERR        i  (W)  Error return flag from rdm_rot_start
-end*/

/*-C:*/
int rdm_rot_start (int kindx, int iset, int imgnum, float * rot_start) 
/*end*/

/*-Parameters:
int kindx;          The KDM index for the RDM dataset (R)
int iset;           The crystal set no. (>0) (R)
int imgnum;         The image number within the set (R)
float * rot_start;  Returns the rotation start angle (W)

Return:  Error flag =0 OK
                    =1 error in getting parameters
                    =2 image out of defined ranges
-end*/
{
   int err;
   int numimg;
   int nimg[5];
   int i;
   int nioff;
   int ir;
   int ii;
   int found;
   float rotstart;
   float rotend;
   float rots2;
   float rote2;
   float rots3;
   float rote3;
   float rots4;
   float rote4;
   float rots5;
   float rote5;
   float oscang;
   float x;

   *rot_start = 0.0;

   err = rdm_getrotations (kindx, iset, &rotstart, &rotend,
                           &rots2, &rote2,
                           &rots3, &rote3,
                           &rots4, &rote4,
                           &rots5, &rote5,
                           &oscang, nimg, &numimg);
   if (err!=0) return err;
   x = rotstart;
   found = 0;
   ii = 0;
   for (i=0;i<5;++i)
   {
     ii += nimg[i];
     if (imgnum<=ii)
     {
        found = 1;
        switch(i)
        {
           case 0:
           x = rotstart;
           break;

           case 1:
           x = rots2;
           break;

           case 2:
           x = rots3;
           break;

           case 3:
           x = rots4;
           break;

           case 4:
           x = rots5;
           break;
        }
        ir = i;
        break;
     }
   }
   nioff = imgnum + nimg[ir] - ii - 1;
   if (found)
   {
      *rot_start = x + nioff*oscang;
      return 0;
   }
   return 2;
}

/* Fortran binding: rdmf_rot_start */

void rdmf_rot_start (int * kindx, int * iset, int * imgnum, 
                     float * rot_start, int * ierr)
{
   *ierr = rdm_rot_start (*kindx, *iset, *imgnum, rot_start);
}
 

/*-Routine: Get RDM image parameters - rdm_getimgpars
The routine rdm_getimgpars (rdmf_getimgpars) returns the values
of the image format related parameter data items for example for use
prior to reading in an image file.
-end*/

/*
*****************************
**     rdm_getimgpars      **
*****************************

Purpose: Get RDM image format parameters

Author:  John W. Campbell, November 1997

CD-Fortran:
         CALL RDMF_GETIMGPARS(KINDX, ISET, IFTYPE, ITYP, LHEAD,
        +                     LREC, IORDER, ISWAP, IROTN,
        +                     NXRASTS, NYRASTS, IERR)
CD-end:
*/
/*-Parameters:
KINDX       i  (R)  The KDM index for the RDM dataset.
ISET        i  (R)  The crystal set no. (>0)
IFTYPE      i  (W)  Returns the file type (format) flag
                    =-1, none
                    = 0, raw
                    = 1, byte
                    = 2, i2
                    = 3, MAR
ITYP        i  (W)  Returns the image data type flag    
                    1 = unsigned byte data
                    2 = unsigned two-byte data (i2)
                    3 = signed integer data (i4)
                    4 = 'squashed i2' data (if stored I<0
                        ipix = -i*8
                    5 = 'squashed i2' data if stored I<0
                        ipix = (I+32768)*8
                    6 = 'squashed i2' data if stored I<0
                        ipix = (I+32768)*32
                    7 = 'squashed i2' data if stored I<0
                        ipix = -(I+1)*256 + 32768
LHEAD       i  (W)  Returns the header length >0 = no. of records
                    <0 -(no. of header bytes)
LREC        i  (W)  Returns the record length in bytes
IORDER      i  (W)  Returns the axis order flag 1-8 (0=undefined)
ISWAP       i  (W)  Returns the byte swap flag =1 yes, =0 no
IROTN       i  (W)  Returns the rotation axis orientation flag
                    =1 right, =2 left, =3 up, =4 down looking
                    along the beam.
NXRASTS     i  (W)  Returns the number of x rasters
NYRASTS     i  (W)  Returns the number of y rasters
IERR        i  (W)  Error flag return from rdm_getimgpars
-end*/

/*-C:*/
int rdm_getimgpars (int kindx, int iset, int * iftype, int * ityp, 
                    int * lhead, int * lrec, int * iorder, int * iswap, 
                    int * irotn, int * nxrasts, int * nyrasts)
/*end*/

/*-Parameters:
int kindx;        The KDM index for the RDM dataset (R)
int iset;         The crystal set no. (>0) (R)
int *iftype;      Returns the file type (format) flag
                  =-1, none
                  = 0, raw
                  = 1, byte
                  = 2, i2
                  = 3, MAR (W)
int * ityp;       Returns the image data type flag    
                  1 = unsigned byte data
                  2 = unsigned two-byte data (i2)
                  3 = signed integer data (i4)
                  4 = 'squashed i2' data (if stored I<0
                      ipix = -i*8
                  5 = 'squashed i2' data if stored I<0
                      ipix = (I+32768)*8
                  6 = 'squashed i2' data if stored I<0
                      ipix = (I+32768)*32
                  7 = 'squashed i2' data if stored I<0
                      ipix = -(I+1)*256 + 32768
int * lhead;      Returns the header length >0 = no. of records
                  <0 -(no. of header bytes)
int * lrec;       Returns the record length in bytes
int * iorder;     Returns the axis order flag 1-8 (0=undefined)
int * iswap;      Returns the byte swap flag =1 yes, =0 no
int * irotn;      Returns the rotation axis orientation flag
                  =1 right, =2 left, =3 up, =4 down looking
                  along the beam.
int * nxrasts;    Returns the number of x rasters
int * nyrasts;    Returns the number of y rasters

Return:  Error flag =0 OK
                    =1 error
-end*/
{
   int err;
   int i;
   int num_inp;
   int istat;
   int itype;
   int ival[10];
   int nd[10];
   float fval[10];
   char valstr[81];
   char errstr[81];
   static char * format_str[] = {"none",
                                 "raw",
                                 "byte",
                                 "i2",
                                 "mar",
                                 0};
   static char * imtyp_str[] = {"byte",
                                "i2",
                                "i4",
                                "squash",
                                "squash2",
                                "squash3",
                                "squash4",
                                0};
   static char * orient_str[] = {"right",
                                 "left",
                                 "up",
                                 "down",
                                0};

   err = kdm_getvalstr (kindx, "IMAGE_FORMAT", 0, iset, 1,  80, valstr, 
                        &istat, &itype, errstr, 80);  
   if (err!=0) return 1;
   *iftype = -1;
   for (i=0;i<5;++i)
   {
      if (strcasecmp(valstr,format_str[i])==0)
      {
         *iftype = i-1;
         break;
      }
   }

   err = kdm_getvalstr (kindx, "IMAGE_DATA", 0, iset, 1,  80, valstr, 
                        &istat, &itype, errstr, 80);  
   if (err!=0) return 1;
   *ityp = 2;
   for (i=0;i<7;++i)
   {
      if (strcasecmp(valstr,imtyp_str[i])==0)
      {
         *ityp = i+1;
         break;
      }
   }

   err = kdm_getvalue (kindx, "IMAGE_HEADER", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   *lhead = ival[0];
   if (err!=0) return 1;

   err = kdm_getvalue (kindx, "IMAGE_RECL", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   *lrec = ival[0];
   if (err!=0) return 1;

   err = kdm_getvalstr (kindx, "AXORD", 0, iset, 1,  80, valstr, 
                        &istat, &itype, errstr, 80);  
   if (err!=0) return 1;

   rdm_axord (valstr, 0, iorder, iswap);

   err = kdm_getvalstr (kindx, "GONIO_ORIENT", 0, iset, 1,  80, valstr, 
                        &istat, &itype, errstr, 80);  
   if (err!=0) return 1;
   *irotn = -1;
   for (i=0;i<4;++i)
   {
      if (strcasecmp(valstr,orient_str[i])==0)
      {
         *irotn = i+1;
         break;
      }
   }

   err = kdm_getvalue (kindx, "NXRASTS", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   *nxrasts = ival[0];
   if (err!=0) return 1;

   err = kdm_getvalue (kindx, "NYRASTS", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   *nyrasts = ival[0];
   if (err!=0) return 1;

   return 0;
}
/* Fortran binding: rdmf_getimgpars */

void rdmf_getimgpars (int * kindx, int * iset, int * iftype, int * ityp,
                        int * lhead, int * lrec, int * iorder, int * iswap,
                        int * irotn, int * nxrasts, int * nyrasts, int *ierr)
{

   *ierr = rdm_getimgpars (*kindx, *iset, iftype, ityp, lhead, lrec,
                           iorder, iswap, irotn, nxrasts, nyrasts);
   return;
}

/*-Routine: Get RDM detector parameters - rdm_getdetpars
The routine rdm_getdetpars (rdmf_getdetpars) returns the values
of the detector related parameter data items which are required to
set up the detector matrix using rdm_mosdet (rdmf_mosdet).
-end*/

/*
*****************************
**     rdm_getdetpars      **
*****************************

Purpose: Get RDM detector parameters

Author:  John W. Campbell, November 1997

CD-Fortran:
         CALL RDMF_GETDETPARS(KINDX, ISET, CTOD, TAU1, TAU2,
        +                     TAU3, DETR, DETA, S0, IERR)
CD-end:
*/
/*-Parameters:
KINDX       i  (R)  The KDM index for the RDM dataset.
ISET        i  (R)  The crystal set no. (>0)
CTOD        r  (W)  Returns the crystal to detector distance in mm
TAU1        r  (W)  Returns the TAU1 detector rotation angle in degrees
TAU2        r  (W)  Returns the TAU2 detector rotation angle in degrees
TAU3        r  (W)  Returns the TAU3 detector rotation angle in degrees
DETR[9]     r  (W)  Returns the detector rotation vectors 
                    (1'st vector in first three elements)
                    (2'nd vector in second three elements)
                    (3'rd vector in last three elements)
DETA[6]     r  (W)  Returns the detector axis vectors
                    (1'st vector in first three elements)
                    (2'nd vector in last three elements)
S0(3)       r  (W)  Returns the beam vector
IERR        i  (W)  Error flag return from rdm_getdetpars
-end*/

/*-C:*/
int rdm_getdetpars (int kindx, int iset, float * ctod, 
                    float * tau1, float * tau2, float * tau3,
                    float detr[], float deta[], float s0[3])
/*end*/

/*-Parameters:
int kindx;        The KDM index for the RDM dataset (R)
int iset;         The crystal set no. (>0) (R)
float * ctod;     Returns the crystal to detector distance in mm (W)
float * tau1;     Returns the TAU1 detector rotation angle in degrees (W)
float * tau2;     Returns the TAU2 detector rotation angle in degrees (W)
float * tau3;     Returns the TAU3 detector rotation angle in degrees (W)
float detr[9];    Returns the detector rotation vectors 
                  (1'st vector in first three elements)
                  (2'nd vector in second three elements)
                  (3'rd vector in last three elements) (W)
float deta[6];    Returns the detector axis vectors
                  (1'st vector in first three elements)
                  (2'nd vector in last three elements) (W)
float s0[3];      Returns the beam vector

Return:  Error flag =0 OK
                    =1 error
-end*/
{
   int err;
   int num_inp;
   int istat;
   int itype;
   int ival[10];
   int nd[10];
   float fval[10];
   char valstr[81];
   char errstr[81];
   
   err = kdm_getvalue (kindx, "DISTANCE", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   *ctod = fval[0];
   if (err!=0) return 1;
    
   err = kdm_getvalue (kindx, "TAU1", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   *tau1 = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "TAU2", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   *tau2 = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "TAU3", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   *tau3 = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "DET_ROTATIONS", 0, iset, 1, 
                       80, valstr, detr, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "DET_AXES", 0, iset, 1, 
                       80, valstr, deta, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   if (err!=0) return 1;

   err = kdm_getvalue (kindx, "BEAM_VECTOR", 0, iset, 1, 
                       80, valstr, s0, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   if (err!=0) return 1;

   return 0;
}
/* Fortran binding: rdmf_getdetpars */

void rdmf_getdetpars (int * kindx, int * iset, float * ctod, 
                      float * tau1, float * tau2, float * tau3,
                      float * detr, float * deta, float * s0, int *ierr)
{

   *ierr = rdm_getdetpars (*kindx, *iset, ctod, tau1, tau2, 
                          tau3, detr, deta, s0);
   return;
}

/*-Routine: Get RDM cell and wavelength - rdm_getcellwave
The routine rdm_getcellwave (rdmf_getcellwave) returns the 
crystal cell parameters and the wavelength.
-end*/

/*
*****************************
**     rdm_getcellwave     **
*****************************

Purpose: Get RDM cell and wavelength parameters

Author:  John W. Campbell, November 1997

CD-Fortran:
         CALL RDMF_GETCELLWAVE (KINDX, ISET, CELL, WAVE, IERR)
CD-end:
*/
/*-Parameters:
KINDX       i  (R)  The KDM index for the RDM dataset.
ISET        i  (R)  The crystal set no. (>0)
CELL(6)     r  (W)  Returns the cell parameters in Angstroms and degrees 
WAVE        r  (W)  Returns the wavelength in Angstroms
IERR        i  (W)  Error flag return from rdm_getcellwave
-end*/

/*-C:*/
int rdm_getcellwave (int kindx, int iset, float * cell, float * wave) 
/*end*/

/*-Parameters:
int kindx;        The KDM index for the RDM dataset (R)
int iset;         The crystal set no. (>0) (R)
float cell[6];    Returns the  cell parameters in Angstroms and degrees (W)
float * wave;     Returns the wavelength in Angstroms (W)

Return:  Error flag =0 OK
                    =1 error
-end*/
{
   int err;
   int num_inp;
   int istat;
   int itype;
   int ival[10];
   int nd[10];
   float fval[10];
   char valstr[81];
   char errstr[81];
   
   err = kdm_getvalue (kindx, "A", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   cell[0] = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "B", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   cell[1] = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "C", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   cell[2] = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "ALPHA", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   cell[3] = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "BETA", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   cell[4] = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "GAMMA", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   cell[5] = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "WAVELENGTH", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   *wave = fval[0];
   if (err!=0) return 1;

   return 0;
}
/* Fortran binding: rdmf_getcellwave */

void rdmf_getcellwave (int * kindx, int * iset, float * cell, float * wave,
                      int *ierr)
{

   *ierr = rdm_getcellwave (*kindx, *iset, cell, wave); 
   return;
}
/*-Routine: Get centre related parameters - rdm_getcentre
The routine rdm_getcentre (rdmf_getcentre) returns the image
centre and offsets parameters.
-end*/

/*
**************************
**     rdm_getcentre    **
**************************

Purpose: Get RDM centre and offsets parameters

Author:  John W. Campbell, November 1997

CD-Fortran:
         CALL RDMF_GETCENTRE (KINDX, ISET, IMGNUM, X_CEN, Y_CEN, X_C, Y_C,
        +                     W_C, COSWC, SINWC, XCOFF, YCOFF, IERR)
CD-end:
*/
/*-Parameters:
KINDX       i  (R)  The KDM index for the RDM dataset.
ISET        i  (R)  The crystal set no. (>0)
IMGNUM      i  (R)  Image number within set
X_CEN       r  (W)  Returns the 'x' centre position in rasters
X_CEN       r  (W)  Returns the 'y' centre position in rasters
X_C         r  (W)  Returns the x_c offset in mm
Y_C         r  (W)  Returns the y_c offset in mm
W_C         r  (W)  Returns the w_c offset in degrees
COSWC       r  (W)  Returns cos(w_c)
SINWC       r  (W)  Returns sin(w_c)
XCMID       r  (W)  Offset in 'x' in mm of corrected pattern centre
                    from image mid point
YCMID       r  (W)  Offset in 'y' in mm of corrected pattern centre
                    from image mid point
IERR        i  (W)  Error flag return from rdm_getcentre

                    If the centre was undefined, the image mid point 
                    is returned for X_CEN, Y_CEN
-end*/

/*-C:*/
int rdm_getcentre (int kindx, int iset, int imgnum, 
                  float * x_cen, float * y_cen,
                  float *x_c, float *y_c, float * w_c, float * coswc,
                  float * sinwc, float * xcmid, float * ycmid) 
/*end*/

/*-Parameters:
int kindx;        The KDM index for the RDM dataset (R)
int iset;         The crystal set no. (>0) (R)
int imgnum;       The image number within the set (R)
float * x_cen;    Returns the 'x' centre position in rasters (W)
float * y_cen;    Returns the 'y' centre position in rasters (W)
float * x_c;      Returns the x_c offset in mm (W)
float * y_c;      Returns the y_c offset in mm (W)
float * w_c;      Returns the w_c offset in degrees (W)
float * coswc;    Returns cos(w_c) (W)
float * sinwc;    Returns sin(w_c) (W)
float * xcmid;    Offset in 'x' in mm of corrected pattern centre
                  from image mid point (W)
float * ycmid;    Offset in 'y' in mm of corrected pattern centre
                  from image mid point (W)

                  If the centre was undefined, the image mid point 
                  is returned for x_cen, y_cen

Return:  Error flag =0 OK
                    =1 error
-end*/
{
   int err;
   int num_inp;
   int istat;
   int itype;
   int ival[10];
   int nd[10];
   float fval[10];
   char valstr[81];
   char errstr[81];
   int nxrasts;
   int nyrasts;
   float wr;
   float mm_rast_x;
   float mm_rast_y;
   
   err = kdm_getvalue (kindx, "X_CEN", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   *x_cen = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "Y_CEN", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   *y_cen = fval[0];
   if (err!=0) return 1;

   err = kdm_getvalue (kindx, "NXRASTS", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   nxrasts = ival[0];
   if (err!=0) return 1;
   err = kdm_getvalue (kindx, "NYRASTS", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   nyrasts = ival[0];
   if (err!=0) return 1;

   if (*x_cen==0.0&&*y_cen==0.0)
   {
      *x_cen = nxrasts/2.0;
      *y_cen = nyrasts/2.0;
   }
   
   err = kdm_getvalue (kindx, "X_C", 0, iset, imgnum, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   *x_c = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "Y_C", 0, iset, imgnum, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   *y_c = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "W_C", 0, iset, imgnum, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   *w_c = fval[0];
   if (err!=0) return 1;
   
   wr = 0.0174533*(*w_c);
   *coswc = cosf(wr);
   *sinwc = sinf(wr);

   rdm_getmmrast (kindx, iset, &mm_rast_x, &mm_rast_y);

   *xcmid = *x_cen + (*x_c)*mm_rast_x - nxrasts/2.0;
   *ycmid = *y_cen + (*y_c)*mm_rast_y - nyrasts/2.0;

   return 0;
}
/* Fortran binding: rdmf_getcentre */

void rdmf_getcentre (int * kindx, int * iset, int * imgnum, float * x_cen, 
                     float * y_cen, float * x_c, float * y_c, float * w_c, 
                     float * coswc, float * sinwc, float * xcmid, 
                     float * ycmid, int *ierr)
{

   *ierr = rdm_getcentre (*kindx, *iset, *imgnum, x_cen, y_cen, x_c, y_c, w_c,
                          coswc, sinwc, xcmid, ycmid); 
   return;
}

/*-Routine: Get distortion related parameters - rdm_getdistor
The routine rdm_getdistor (rdmf_getdistor) returns the distortion
related parameters.
-end*/

/*
**************************
**     rdm_getdistor    **
**************************

Purpose: Get RDM distortion parameters

Author:  John W. Campbell, November 1997

CD-Fortran:
         CALL RDMF_GETDISTOR (KINDX, ISET, IMGNUM, IDTYPE, TWIST, TILT, 
        +                     Y_SCALE, ROFF, TOFF, IERR)
CD-end:
*/
/*-Parameters:
KINDX       i  (R)  The KDM index for the RDM dataset.
ISET        i  (R)  The crystal set no. (>0)
IMGNUM      i  (R)  Image number within set
IDTYPE      i  (W)  Returns the distortion type =1 standard
                                                =2 rtoff
TWIST       r  (W)  Returns the distortion twist parameter
TILT        r  (W)  Returns the distortion tiltt parameter
Y_SCALE     r  (W)  Returns the distortion y-scale parameter
ROFF        r  (W)  Returns the radial offset distortion parameter
TOFF        r  (W)  Returns the tangential offset distortion parameter
IERR        i  (W)  Error flag return from rdm_getdistor
-end*/

/*-C:*/
int rdm_getdistor (int kindx, int iset, int imgnum, int * idtype, 
                   float * twist, float * tilt, float * y_scale, 
                   float * roff, float * toff) 
/*end*/

/*-Parameters:
int kindx;        The KDM index for the RDM dataset (R)
int iset;         The crystal set no. (>0) (R)
int imgnum;       The image number within the set (R)
int * idtype;     Returns the distortion type =1 standard
                                              =2 rtoff (W)
float * twist;    Returns the distortion twist parameter (W)
float * tilt;     Returns the distortion tilt parameter (W)
float * y_scale;  Returns the distortion y-scale parameter (W)
float * roff;     Returns the radial offset distortion parameter (W) 
float * toff;     Returns the tangential offset distortion parameter (W) 

Return:  Error flag =0 OK
                    =1 error
-end*/
{
   int i;
   int err;
   int num_inp;
   int istat;
   int itype;
   int ival[10];
   int nd[10];
   float fval[10];
   char valstr[81];
   char errstr[81];
   static char * dstyp_str[] = {"standard",
                                "rtoff",
                                0};

   err = kdm_getvalstr (kindx, "DISTOR_TYPE", 0, iset, 1,  80, valstr, 
                        &istat, &itype, errstr, 80);  
   if (err!=0) return 1;
   *idtype = 1;
   for (i=0;i<2;++i)
   {
      if (strcasecmp(valstr,dstyp_str[i])==0)
      {
         *idtype = i+1;
         break;
      }
   }
   
   err = kdm_getvalue (kindx, "TWIST", 0, iset, imgnum, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   *twist = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "TILT", 0, iset, imgnum, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   *tilt = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "Y_SCALE", 0, iset, imgnum, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   *y_scale = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "ROFF", 0, iset, imgnum, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   *roff = fval[0];
   if (err!=0) return 1;
   
   err = kdm_getvalue (kindx, "TOFF", 0, iset, imgnum, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   *toff = fval[0];
   if (err!=0) return 1;

   return 0;
}
/* Fortran binding: rdmf_getdistor */

void rdmf_getdistor (int * kindx, int * iset, int * imgnum,  int *idtype, 
                     float * twist, float * tilt, float * y_scale, 
                     float * roff, float * toff, int *ierr)
{

   *ierr = rdm_getdistor (*kindx, *iset, *imgnum, idtype, twist, tilt, 
                          y_scale, roff, toff);
   return;
}

/*-Routine: Get mm to raster conversion factors - rdm_getmmrast
The routine rdm_getmmrast (rdmf_getmmrast) returns the conversion
factors from mm to rasters in the detector xd and yd directions.
-end*/

/*
**************************
**     rdm_getmmrast    **
**************************

Purpose: Get mm to raster conversion factors

Author:  John W. Campbell, November 1997

CD-Fortran:
         CALL RDMF_GETMMRAST (KINDX, ISET, MM_RAST_X, MM_RAST_Y, IERR)
CD-end:
*/
/*-Parameters:
KINDX       i  (R)  The KDM index for the RDM dataset.
ISET        i  (R)  The crystal set no. (>0)
MM_RAST_X   r  (W)  Conversion factor for mm to x-rasters
MM_RAST_Y   r  (W)  Conversion factor for mm to y-rasters
IERR        i  (W)  Error flag return from rdm_getmmrast
-end*/

/*-C:*/
int rdm_getmmrast (int kindx, int iset, float * mm_rast_x, float * mm_rast_y) 
/*end*/

/*-Parameters:
int kindx;         The KDM index for the RDM dataset (R)
int iset;          The crystal set no. (>0) (R)
float * mm_rast_x; Conversion factor for mm to x-rasters (W)
float * mm_rast_y; Conversion factor for mm to y-rasters (W)
Return:  Error flag =0 OK
                    =1 error
-end*/
{
   int err;
   int num_inp;
   int istat;
   int itype;
   int ival[10];
   int nd[10];
   float fval[10];
   char valstr[81];
   char errstr[81];
   float ras;
   float y_scale;

   err = kdm_getvalue (kindx, "RASTER", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   ras = fval[0];
   if (err!=0) return 1;

   err = kdm_getvalue (kindx, "Y_SCALE", 0, iset, 1, 
                       80, valstr, fval, nd, ival, &num_inp, &istat, &itype, 
                       errstr, 80);
   y_scale = fval[0];
   if (err!=0) return 1;

   *mm_rast_x = 1000.0/ras;
   *mm_rast_y = (*mm_rast_x)*y_scale;

   return 0;
}
/* Fortran binding: rdmf_getmmrast */

void rdmf_getmmrast (int * kindx, int * iset, float * mm_rast_x, 
                     float * mm_rast_y, int *ierr)
{

   *ierr = rdm_getmmrast (*kindx, *iset, mm_rast_x, mm_rast_y);
   return;
}
/*-Section: Output RDM data set 
This section contains a routine to enable the output of RDM (including 
symmetry data) for preparing an updated RDM file or for output
of current parameter values to a log file.
-end*/

/*-Routine:  Write RDM keyword data - rdm_write
This routine rdm_write (rdmf_write) writes out the RDM keyword data 
including the symmetry data. The user must supply a function to output 
a string to the required output stream or file. A number of options
selecting the data to be written are available.
-end*/

/*
**********************
**     rdm_write    **
**********************

Purpose:  Write out RDM data (including symmetry)

Author:  John W. Campbell, January 1998

CD-Fortran:
         CALL RDMF_WRITE (KDX, IOPTYP, ICHANGED, ISYMM, KDMSTR(OUTSTR), 
                          MAXO, OP_FUNC, IERR)
CD-end:
*/
/*-Parameters:
KDX       i  (R) Index of RDM parameter data set
IOPTYP    i  (R) Type of output = 1 minimum, do not o/p parameters with 
                                    default values
                                = 2 standard output, do not o/p parameters
                                    with default values except for those
                                    which have been flagged to be printed
                                    as standard.
                                = 3 full (all parameters will be included)
                                    but with reduced output where feasible
                                    by recombining values if all the same 
                                    for a set etc.
                                = 4 all pamaeters and individual values 
                                    given explicitly for all set/subset 
                                    parameters.
ICHANGED  i  (R) = 0 Output all parameters
                 1-16 Write changed parameter values since last reset
                 of flag for channel ICHANGED
ISYMM     i  (R) Output symmetry at position 'ISYMM' in the output
                 RDM parameters file. (If 0, will be put at end)
OUTSTR    c  (*) Buffer used in printing output (should be at least
                 120 characters in length)
                 ** Pass address index using the KDMSTR function **
MAXO      i  (R) Maximum length allowed for OUTSTR (should
                 be at least 120)
OP_FUNC   s  (R) Function to do the actual output. The data to be
                 output by this function will be stored in OUTSTR,

                 The call will be CALL OP_FUNC without any parameters
                 so that the string OUTSTR must be available to the
                 routine op_func via a common block possibly with
                 other parameters such as Fortran unit numbers etc.
IERR      i  (W) =0 OK, =1 invalid RDM index or invalid parameter
-end*/


/*-C:*/
int rdm_write (int kdx, int ioptyp, int ichanged, int isymm, 
                char * outstr, int maxo, void (*op_func)())
/*end*/

/*-Parameters:
int kdx;             Index of RDM parameter data set (R)
int ioptyp;          Type of output = 1 minimum, do not o/p parameters with 
                                    default values
                                = 2 standard output, do not o/p parameters
                                    with default values except for those
                                    which have been flagged to be printed
                                    as standard.
                                = 3 full (all parameters will be included)
                                    but with reduced output where feasible
                                    by recombining values if all the same 
                                    for a set etc.
                                = 4 all pamaeters and individual values 
                                    given explicitly for all set/subset 
                                    parameters.(R)
int ichanged;        = 0 Output all parameters
                     1-16 Write changed parameter values since last reset
                     of flag for channel ichanged (R)
int isymm;           Output symmetry at position 'ISYMM' in the output
                     RDM parameters file. (If 0, will be put at end) (R)
char * outstr;       Returns the output string for printing (W)
int maxo;            if >0 Maximum length of returned output string allowed
                           (excluding terminating null - outstr must be at
                           least maxo + 1 characters in length); maxo
                           should be at least 2050.
                     if <0 abs(maxo) characters will be returned padded
                           with blanks if needed (for 'fortran' use) 
                           (abs(maxo) should be at least 2050) (R)
void (*op_func)();   Function to o/p the string; if NULL then o/p will be
                     done to the standard output. If defined then this
                     function will be called without parameters; the
                     string 'outstr' will contain the required o/p and
                     must be globally accessible to the output function
                     together with any other special parameters required.
                     (R)

Return 0 OK
       1 Invalid RDM index of function call parameter
-end*/
{
   int err;
   int i;
   int ip;
   int nump;
   int nsymm;
   int ns;
   int nss[MAX_SETS];
   int write_symm;
   int ireduce = 0;
   int ival[10];
   int nd[10];
   int ityp;
   int istat;
   int num_inp;
   float fval[10];
   char name[51];
   char errstr[3];
   char tmpstr[121];

   /* Check parameters & set up dependent parameters */

   if (ichanged<0||ichanged>16) return 1;
   if (ioptyp<1||ioptyp>4) return 1;

   ireduce = 0;
   if (ioptyp==3) ireduce = 1;
   if (ioptyp==2) ireduce = 2;
   if (ioptyp==1) ireduce = 3;


   /* Get no. of RDM parameters (excluding symmetry) */
   err = kdm_numpars (kdx, &nump);
   if (err!=0) return 1;
   if (nump<1) return 1;

   if (isymm==0) isymm = nump+1;


   /* See if symmetry s to be written */
   write_symm = 0;
   if (ichanged>0)
   {
      if (rdm_symm_changed (ichanged)==1) write_symm = 1;
   }
   else
   {
      if (ioptyp==1)
      {
         if (rdm_get_nsym (&nsymm)) write_symm = 1;
      }
      else
      {
         write_symm = 1;
      }
   }

   /* Get current number of sets and numbers of subsets (images) */

   kdm_getvalue (kdx, "NUMSETS", 0, 1, 1, 
                 2, tmpstr,  fval, nd, ival, &num_inp, &istat, &ityp, 
                 errstr, 2);
   ns = ival[0];
   if (ns<1) ns = 1;

   for (i=0;i<ns;++i)
   {
      rdm_numimg (kdx, i+1, &nss[i]);
      if (nss[i]<1) nss[i] = 1;
   }
   
   for (ip=1;ip<=nump;++ip)
   {
      if (ip==isymm&&write_symm) rdm_symm_write (outstr, maxo, op_func);
      err = kdm_parname (kdx, ip, name, 50);
      if (err!=0) continue;
      kdm_output (kdx, name, 0, ichanged, ireduce, ns, nss, 
                  outstr, maxo, op_func, errstr, 2);
   }
   if (isymm>nump&&write_symm) rdm_symm_write (outstr, maxo, op_func);

   return 0;
}
/* Fortran binding: rdmf_write */

void rdmf_write (int * kdx, int * ioptyp, int * ichanged, int * isymm,
                 int * outstr, int *maxo, void (*op_func)(), int * ierr)
{
   int mxo;

   mxo = - *maxo;
   if (mxo>0) mxo = -mxo;
   *ierr = rdm_write (*kdx, *ioptyp, *ichanged, *isymm, 
                   KDM_pointers[*outstr], mxo, op_func);
}


/*-Section: Crystallographic Calculations
These are a number of routines which carry out some low level
crystallographic calculations based on RDM derived parameters.
They include the setting up of various matrices and coordinate
conversion routines.
-end*/

/*-Routine: Calculate detector matrices - rdm_mosdet
The routine rdm_mosdet (rdmf_mosdet) determines the detector
orientation matrix and its inverse and also returns the vector normal to
the detector. The calculations are based on the MOSFLM axis convention.
This routine assumes that the camera constants x_c, y_c (and w_c) are
taken care of elsewhere.
-end*/

/*
*************************
**     rdm_mosdet      **
*************************

Purpose: Determine dectector orientation matrix

Author:  John W. Campbell, November 1997

CD-Fortran:
         CALL RDMF_MOSDET(CTOD, TAU1, TAU2, TAU3,  DETR, DETA, 
        +                 DD, DDINV, DN)
CD-end:
*/
/*-Parameters:
CTOD        r  (R)  The crystal to detector distance in mm
TAU1        r  (R)  The TAU1 detector rotation angle in degrees
TAU2        r  (R)  The TAU2 detector rotation angle in degrees
TAU3        r  (R)  The TAU3 detector rotation angle in degrees
DETR[9]     r  (R)  The detector rotation vectors 
                    (1'st vector in first three elements)
                    (2'nd vector in second three elements)
                    (3'rd vector in last three elements)
DETA[6]     r  (R)  The detector axis vectors
                    (1'st vector in first three elements)
                    (2'nd vector in last three elements)
DD(3,3)     r  (W)  Returns the detector orientation matrix
DDINV(3,3)  r  (W)  Returns the inverse of the detector orientation 
                    matrix
DN(3)       r  (W)  Returns the detector normal vector
-end*/

/*-C:*/
void rdm_mosdet (float ctod, float tau1, float tau2, float tau3,
                float detr[], float deta[], float dd[3][3], 
                float ddinv[3][3], float dn[])
/*end*/

/*-Parameters:
float ctod;        The crystal to detector distance in mm (R)
float tau1;        The TAU1 detector rotation angle in degrees (R)
float tau2;        The TAU2 detector rotation angle in degrees (R)
float tau3;        The TAU3 detector rotation angle in degrees (R)
float detr[9];     The detector rotation vectors 
                   (1'st vector in first three elements)
                   (2'nd vector in second three elements)
                   (3'rd vector in last three elements) (R)
float deta[6];     The detector axis vectors
                   (1'st vector in first three elements)
                   (2'nd vector in last three elements) (R)
float dd[3][3];    Returns the detector orientation matrix (W)
float ddinv[3][3]; Returns the inverse of the detector orientation 
                   matrix (W)
dloat dn[3];       Returns the detector normal vector (W)
-end*/
{
   int i;
   float t1[3][3];
   float t2[3][3];
   float t3[3][3];
   float t21[3][3];
   float dgds[3];
   float po[3];
   float tau321[3][3];
   float tau32[3][3];
   float dx[3];
   float dy[3];
   float d0[3];
   float det;
   float x_c;
   float y_c;

   static float dg2rd = 0.0174533;

   x_c = 0.0;
   y_c = 0.0;

   dgds[0] = ctod;
   dgds[1] = x_c;
   dgds[2] = y_c;
   po[0] = ctod;
   po[1] = 0.0;
   po[2] = 0.0;
   rdm_rtm (tau3*dg2rd, &detr[6], t3);
   rdm_rtm (tau2*dg2rd, &detr[3], t2);
   rdm_rtm (tau1*dg2rd, &detr[0], t1);
   rdm_matmul (t21, t2, t1);
   rdm_matmul (tau321, t3, t21);
   rdm_matmul (tau32, t3, t2);
   rdm_matvec (dx, tau321, &deta[0]); 
   rdm_matvec (dy, tau321, &deta[3]);
   rdm_matvec (dn, tau32,  po);
   rdm_matvec (d0, tau321, dgds);
   for (i=0;i<3;++i)
   {
      dd[i][0] = dx[i];
      dd[i][1] = dy[i];
      dd[i][2] = d0[i];
   }
   rdm_minv3(ddinv, dd, &det);

/*>>>*/
/*   printf ("dd %6.3f %6.3f %6.3f %6.3f %6.3f %6.3f %6.3f %6.3f %6.3f\n",
dd[0][0], dd[0][1],
dd[0][2], dd[1][0], dd[1][1], dd[1][2], dd[2][0], dd[2][1], dd[2][2]);
printf ("ddinv %6.3f %6.3f %6.3f %6.3f %6.3f %6.3f %6.3f %6.3f %6.3f\n",
ddinv[0][0], ddinv[0][1],
ddinv[0][2], ddinv[1][0], ddinv[1][1], ddinv[1][2], ddinv[2][0], 
ddinv[2][1], ddinv[2][2]); */
   return;
}
/* Fortran binding: rdmf_mosdet */

void rdmf_mosdet (float * ctod, float * tau1, float * tau2, float * tau3,
                  float * detr, float * deta, float * dd ,
                  float * ddinv, float * dn)
{
   int i,j,k;
   float dd_tmp[3][3];
   float ddinv_tmp[3][3];

   rdm_mosdet (*ctod, *tau1, *tau2, *tau3, detr, deta, dd_tmp, 
                       ddinv_tmp, dn);
   k = 0;
   for (i=0;i<3;++i)
   {
      for (j=0;j<3;++j)
      {
         dd[k] = dd_tmp[j][i];
         k++;
      }
   }
   k = 0;
   for (i=0;i<3;++i)
   {
      for (j=0;j<3;++j)
      {
         ddinv[k] = ddinv_tmp[j][i];
         k++;
      }
   }
   return;
}

/*-Routine: Calculate 'B'matrix - rdm_bmatrx
The routine rdm_bmatrx (rdmf_bmatrx) determines the crystal
orientation  'B' matrix.
-end*/

/*
*************************
**     rdm_bmatrx      **
*************************

Purpose: Determine orientation 'B' matrix

Author:  John W. Campbell, November 1997

CD-Fortran:
         CALL RDMF_BMATRX (CELL, WAVE, BMATRX)
CD-end:
*/
/*-Parameters:
CELL(6)     r  (R)  The cell parameters in Angstroms and degrees
WAVE        r  (R)  The wavelength in Angstroms
BMATRX(3,3) r  (W)  Returns the 'B' matrix
-end*/

/*-C:*/
void rdm_bmatrx (float * cell, float wave, float bmatrx[3][3]) 
/*end*/

/*-Parameters:
float cell[];       The cell parameters in Angstroms and degrees (R)
float wave;         The wavelength in Angstroms (R)
float bmatrx[3][3]; Returns the 'B' matrix (W)

Return:  Error flag =0 OK
-end*/
{
   float r[6];
   static float dtor = 0.0174533;

   rdm_reccel (cell, wave, r);
   bmatrx[0][0] = r[0];
   bmatrx[0][1] = cosf(r[5]*dtor)*r[1];
   bmatrx[0][2] = cosf(r[4]*dtor)*r[2];
   bmatrx[1][0] = 0.0;
   bmatrx[1][1] = sinf(r[5]*dtor)*r[1];
   bmatrx[1][2] = -sinf(r[4]*dtor)*r[2]*cosf(cell[3]*dtor);
   bmatrx[2][0] = 0.0;
   bmatrx[2][1] = 0.0;
   bmatrx[2][2] = wave/cell[2];
   return;
}
/* Fortran binding: rdmf_bmatrx */

void rdmf_bmatrx (float * cell, float * wave, float * bmatrx)
{
   int i,j,k;
   float btmp[3][3];

   rdm_bmatrx (cell, *wave, btmp);
   k = 0;
   for (i=0;i<3;++i)
   {
      for (j=0;j<3;++j)
      {
         bmatrx[k] = btmp[j][i];
         k++;
      }
   }
   return;
}
/*-Routine: Reciprocal cell - rdm_reccel
The routine rdm_reccel (rdmf_reccel) converts between a real and
reciprocal cell.
-end*/

/*
*************************
**     rdm_reccel      **
*************************

Purpose: Convetr between real & reciprocal cell

Author:  John W. Campbell, November 1997

CD-Fortran:
         CALL RDMF_RECCEL (CELL, WAVE, RCELL)
CD-end:
*/
/*-Parameters:
CELL(6)     r  (R)  The input cell parameters in Angstroms and degrees
WAVE        r  (R)  The wavelength in Angstroms
RCELL(6)    r  (W)  The output reciprocal cell parameters in Angstroms 
                    and degrees
-end*/

/*-C:*/
void rdm_reccel (float cx[], float wave, float rx[]) 
/*end*/

/*-Parameters:
float cx[6];        The input cell parameters in Angstroms and 
                    degrees (R)
float wave;         The wavelength in Angstroms (R)
float rx[6];        The output reciprocal cell parameters in 
                    Angstroms and degrees (W)
-end*/
{
   int i;
   float c[6];
   float sum;
   float v;
   float wv;
   static float dtor = 0.0174533;

   for(i=0;i<3;++i) c[i] = cx[i];
   sum = 0.0;
   for (i=3;i<6;++i)
   {
      c[i] = dtor*cx[i];
      sum += 0.5*c[i];
   }
   v = 1.0;
   for (i=3;i<6;++i)
   {
      v = v*sinf(sum-c[i]);
   }
   wv = wave/(2.*c[0]*c[1]*c[2]*sqrtf(sinf(sum)*v));
   rx[0] = c[1]*c[2]*sinf(c[3])*wv;
   rx[1] = c[0]*c[2]*sinf(c[4])*wv;
   rx[2] = c[0]*c[1]*sinf(c[5])*wv; 
   rx[3] = acosf((cosf(c[4])*cosf(c[5])-cosf(c[3]))/(sinf(c[4])*sinf(c[5])));
   rx[3] = rx[3]/dtor;
   rx[4] = acosf((cosf(c[5])*cosf(c[3])-cosf(c[4]))/(sinf(c[5])*sinf(c[3])));
   rx[4] = rx[4]/dtor;
   rx[5] = acosf((cosf(c[3])*cosf(c[4])-cosf(c[5]))/(sinf(c[3])*sinf(c[4])));
   rx[5] = rx[5]/dtor; 
   return;
}
/* Fortran binding: rdmf_reccel */

void rdmf_reccel (float * cell, float * wave, float * rcell)
{
   rdm_reccel (cell, *wave, rcell);
   return;
}
/*-Routine: Calculate matrix for a rotation - rdm_rtm
The routine rdm_rtm (rdmf_rtm) calculates a rotation matrix derived
from a requested rotation around a requested axis.
-end*/

/*
***************************
**     rdm_rtm           **
***************************

Purpose: Set up rotation matrix for a rotation around a requested axis

CD-Fortran:
         CALL RDMF_RTM(ANGLE, AXIS, MATS)
CD-end:
*/
/*-Parameters:
ANGLE       r  (R)  The rotation angle around the axis in radians
AXIS(3)     r  (R)  The axsis vector around which the rotation is to
                    be done
MATS(3,3)   r  (W)  Returns the rotation matrix
-end*/

/*-C:*/
void rdm_rtm (float angle, float * axis, float mats[3][3])
/*end*/

/*-Parameters:
float angle;      The rotation angle around the axis in radians (R)
float axis[3];    The axsis vector around which the rotation is to
                  be done (R)
float mats[3][3]; Returns the rotation matrix (W)
-end*/
{
   float csr;
   float vcr;
   float snr;

   csr = cosf(angle);
   vcr = 1.0 - csr;
   snr = sinf(angle);

   /* Evaluate the rotation matrix explicitly */

   mats[0][0] = vcr*axis[0]*axis[0] + csr;
   mats[1][0] = vcr*axis[1]*axis[0] + snr*axis[2];
   mats[2][0] = vcr*axis[2]*axis[0] - snr*axis[1];
   mats[0][1] = vcr*axis[0]*axis[1] - snr*axis[2];
   mats[1][1] = vcr*axis[1]*axis[1] + csr;
   mats[2][1] = vcr*axis[2]*axis[1] + snr*axis[0];
   mats[0][2] = vcr*axis[0]*axis[2] + snr*axis[1];
   mats[1][2] = vcr*axis[1]*axis[2] - snr*axis[0];
   mats[2][2] = vcr*axis[2]*axis[2] + csr;
   return;
}
/* Fortran binding: rdmf_rtm */

void rdmf_rtm (float * angle, float * axis, float * mats)
{
   int i,j,k;
   float mtmp[3][3];

   rdm_rtm (*angle, axis, mtmp);
   k = 0;
   for (i=0;i<3;++i)
   {
      for (j=0;j<3;++j)
      {
         mats[k] = mtmp[j][i];
         k++;
      }
   }

}

/*-Routine: Reciprocal lattice to detector coordinates - rdm_rtod
The routine rdm_rtod (rdmf_rtod) calculates ideal dectector coordinates
in mm from the pattern centre from a set of reciprocal lattice
coordinates for a reflection.
-end*/

/*
***************************
**     rdm_rtod           **
***************************

Purpose: Reciprocal lattice to ideal detector coordinates

CD-Fortran:
         CALL RDMF_RTOD(XR, CTOD, S0, DDINV, DN, XF, YF)
CD-end:
*/
/*-Parameters:
XR(3)       r  (R)  The reciprocal lattice coordinates
CTOD        r  (R)  The crystal to detector distance in mm
S0(3)       r  (R)  The beam vector
DDINV(3,3)  r  (R)  The inverse of the detector orientation matrix
DN(3)       r  (R)  The detector normal vector
XF          r  (W)  Returns the ideal detector 'x' coordinate in mm.
XF          r  (W)  Returns the ideal detector 'y' coordinate in mm.
-end*/

/*-C:*/
void rdm_rtod (float xr[3], float ctod, float s0[3], float ddinv[3][3],
               float dn[3], float * xf, float *yf)
/*end*/

/*-Parameters:
float xr[3];       The reciprocal lattice coordinates (R)
float ctod;        The crystal to detector distance in mm (R)
float s0[3];       The beam vector (r)
float ddinv[3][3]; The inverse of the detector orientation matrix (R)
float dn[3];       The detector normal vector (R)
float *xf;         Returns the ideal detector 'x' coordinate in mm. (W)
float *yf;         Returns the ideal detector 'y' coordinate in mm. (W)
-end*/
{
   float scv1;
   float scv2;
   float rmfac;
   float vstor1[3];
   float ts[3];
   float vdcc[3];

   scv1 = rdm_dot(xr, dn);
   scv2 = rdm_dot(s0, dn);
   rmfac = ctod*ctod / (scv1 - scv2);
   rdm_vdif(vstor1, xr, s0);
   rdm_scalev(ts, rmfac, vstor1);
   rdm_matvec(vdcc, ddinv, ts);
   *xf = vdcc[0];
   *yf = vdcc[1];
   return;
}
/* Fortran binding: rdmf_rtod */

void rdmf_rtod (float * xr, float * ctod, float * s0, float * ddinv,
                float * dn, float * xf, float * yf)
{
   int i,j,k;
   float ddinv_tmp[3][3];

   k = 0;
   for (i=0;i<3;++i)
   {
      for (j=0;j<3;++j)
      {
         ddinv_tmp[j][i] = ddinv[k];
         k++;
      }
   }
   rdm_rtod (xr, *ctod, s0, ddinv_tmp, dn, xf, yf);
   return;
}

/*-Routine: Detector to reciprocal lattice coordinates - rdm_dtor
The routine rdm_dtor (rdmf_dtor) calculates reciprocal lattice
coordinates for a reflection from the ideal dectector coordinates
in mm from the pattern centre.
-end*/

/*
***************************
**     rdm_dtor           **
***************************

Purpose: Ideal detector to reciprocal lattice tocoordinates

CD-Fortran:
         CALL RDMF_DTOR(XF, YF, S0, DD, XR)
CD-end:
*/
/*-Parameters:
XF          r  (R)  The ideal detector 'x' coordinate in mm.
YF          r  (R)  The ideal detector 'y' coordinate in mm.
S0(3)       r  (R)  The beam vector
DD(3,3)     r  (R)  The detector orientation matrix
XR(3)       r  (W)  Returns the reciprocal lattice coordinates
-end*/

/*-C:*/
void rdm_dtor (float xf, float yf, float s0[3], float dd[3][3], float xr[3])
/*end*/

/*-Parameters:
float xf;          The ideal detector 'x' coordinate in mm. (R)
float yf;          The ideal detector 'y' coordinate in mm. (R)
float s0[3];       The beam vector (R)
float dd[3][3];    The detector orientation matrix (R)
float xr[3];       Returns The reciprocal lattice coordinates (W)
-end*/
{
   float s[3];
   float vdco[3];

   vdco[0] = xf;
   vdco[1] = yf;
   vdco[2] = 1.0;
   rdm_matvec (s, dd, vdco);
   rdm_vnorm (s, s);
   rdm_vadd (xr, s0, s);
   return;
}
/* Fortran binding: rdmf_dtor */

void rdmf_dtor (float * xf, float * yf, float * s0, float * dd, float * xr)
{
   int i,j,k;
   float dd_tmp[3][3];

   k = 0;
   for (i=0;i<3;++i)
   {
      for (j=0;j<3;++j)
      {
         dd_tmp[j][i] = dd[k];
         k++;
      }
   }
   rdm_dtor (*xf, *yf, s0, dd_tmp, xr);
   return;
}

/*-Routine: Apply distortion corrections - rdm_corr
The routine rdm_corr (rdmf_corr) applies distortion corrections to
a set of ideal detector coordinates in mm relative to the
pattern centre to provide a set of distortion corrected coordinates xfd, yfd 
in mm parallel to the detector axes (again relative to the pattern centre).
-end*/

/*
***************************
**     rdm_corr          **
***************************

Purpose: Ideal detector coordinates to distortion corrected detector
         coordinates

CD-Fortran:
         CALL RDMF_CORR(XF, YF, CTOD, COSWC, SINWC, IDTYPE, TWIST, TILT,
                       XCMID, YCMID, ROFF, TOFF, XFD, YFD)
CD-end:
*/
/*-Parameters:
XF          r  (R)  The ideal detector 'x' coordinate in mm.
YF          r  (R)  The ideal detector 'y' coordinate in mm.
CTOD        r  (R)  Crystal to detector distance in mm
COSWC       r  (R)  cos(w_c)
SINWC       r  (R)  sin(w_c)
IDTYPE      r  (R)  Distortion type =1 standard, =2 rtoff
TWIST       r  (R)  The distortion twist correction
TILT        r  (R)  The distortion tilt correction
XCMID       r  (R)  Offset in 'x' in mm of corrected pattern centre
                    from image mid point
YCMID       r  (R)  Offset in 'y' in mm of corrected pattern centre
                    from image mid point
ROFF        r  (R)  The distortion radial offset correction
TOFF        r  (R)  The distortion tangential offset correction
XFD         r  (W)  Returns the distortion corrected 'x' coordinate
                    in mm.
YFD         r  (W)  Returns the distortion corrected 'y' coordinate
                    in mm.
-end*/

/*-C:*/
void rdm_corr (float xf, float yf, float ctod, float coswc, float sinwc, 
               int idtype, float twist, float tilt, float xcmid, float ycmid,
               float roff, float toff, float * xfd, float * yfd)
/*end*/

/*-Parameters:
float xf;          The ideal detector 'x' coordinate in mm. (R)
float yf;          The ideal detector 'y' coordinate in mm. (R)
float ctod;        Crystal to detector distance in mm (R)
float coswc;       cos(w_c) (R)
float sinwc;       sin(w_c) (R)
int idtype;        Distortion type =1 standard, =2 rtoff (R)
float twist;       The distortion twist correction (R)
float tilt;        The distortion tilt correction (R)
float xcmid;       Offset in 'x' in mm of corrected pattern centre
                   from image mid point (R)
float ycmid;       Offset in 'y' in mm of corrected pattern centre
                   from image mid point (R)
float roff;        The distortion radial offset correction (R)
float toff;        The distortion tangential offset correction (R)
float * xfd;       Returns the distortion corrected 'x' coordinate
                   in mm. (W)
float * yfd;       Returns the distortion corrected 'y' coordinate
                   in mm. (W)
-end*/
{
   float torad = 0.0001745329;
   float fac;
   float kfac;
   float xq, yq;
   float xp, yp;
   float cpsi, spsi;
   float rq;

   fac = torad/ctod;

   xq = xf*coswc - yf*sinwc;
   yq = xf*sinwc + yf*coswc;

   kfac = 1.0 + fac*(tilt*xq + twist*yq);

   *xfd = kfac*xq;
   *yfd = kfac*yq;

   if (idtype==2)
   {
      xp = xcmid + *xfd;
      yp = ycmid + *yfd;
      rq = sqrt(xp*xp+yp*yp);
      if (rq<0.0001)
      {
          cpsi = 1.0;
          spsi = 0.0;
      }
      else
      {
          cpsi = xp/rq;
          spsi = yp/rq;
      }
      *xfd = kfac*xq + 0.01*(-toff*spsi+roff*cpsi);
      *yfd = kfac*yq + 0.01*(toff*cpsi+roff*spsi); 
   }
   return;
}
/* Fortran binding: rdmf_corr */

void rdmf_corr (float * xf, float * yf, float * ctod, 
                float * coswc, float *sinwc,
                int *idtype, float * twist, float * tilt, 
                float * xcmid, float * ycmid, float * roff, float * toff,
                float * xfd, float * yfd)
{
   rdm_corr (*xf, *yf, *ctod, *coswc, *sinwc, *idtype, *twist, *tilt,
            *xcmid, *ycmid, *roff, *toff, xfd, yfd);
   return;
}


/*-Routine: Get ideal coordinates - rdm_ideal
The routine rdm_ideal (rdmf_ideal) is essentially the reverse of
rdm_corr and uses the distortion parameters to convert measured coordinates
in mm from the pattern centre parallel to the detector axes to ideal
coordinates in mm. This reverse calculation requires an iterative
procedure.
-end*/

/*
***************************
**     rdm_ideal         **
***************************

Purpose: Measured detector coordinates to ideal detector coordinates

CD-Fortran:
         CALL RDMF_IDEAL(XFD, YFD, CTOD, COSWC, SINWC, IDTYPE, TWIST, TILT,
                       XCMID, YCMID, ROFF, TOFF, XF, YF)
CD-end:
*/
/*-Parameters:
XFD         r  (R)  The measured detector 'x' coordinate in mm.
YFD         r  (R)  The measured detector 'y' coordinate in mm.
CTOD        r  (R)  Crystal to detector distance in mm
COSWC       r  (R)  cos(w_c)
SINWC       r  (R)  sin(w_c)
IDTYPE      r  (R)  Distortion type =1 standard, =2 rtoff
TWIST       r  (R)  The distortion twist correction
TILT        r  (R)  The distortion tilt correction
XCMID       r  (R)  Offset in 'x' in mm of corrected pattern centre
                    from image mid point
YCMID       r  (R)  Offset in 'y' in mm of corrected pattern centre
                    from image mid point
ROFF        r  (R)  The distortion radial offset correction
TOFF        r  (R)  The distortion tangential offset correction
XF          r  (W)  Returns the 'ideal' detector 'x' coordinate
                    in mm.
YF          r  (W)  Returns the  'ideal' detector 'y' coordinate
                    in mm.
-end*/

/*-C:*/
void rdm_ideal (float xfd, float yfd, float ctod, float coswc, float sinwc, 
               int idtype, float twist, float tilt, float xcmid, float ycmid,
               float roff, float toff, float * xf, float * yf)
/*end*/

/*-Parameters:
float xfd;         The measured detector 'x' coordinate in mm. (R)
float yfd;         The measured detector 'y' coordinate in mm. (R)
float ctod;        Crystal to detector distance in mm (R)
float coswc;       cos(w_c) (R)
float sinwc;       sin(w_c) (R)
int idtype;        Distortion type =1 standard, =2 rtoff (R)
float twist;       The distortion twist correction (R)
float tilt;        The distortion tilt correction (R)
float xcmid;       Offset in 'x' in mm of corrected pattern centre
                   from image mid point (R)
float ycmid;       Offset in 'y' in mm of corrected pattern centre
                   from image mid point (R)
float roff;        The distortion radial offset correction (R)
float toff;        The distortion tangential offset correction (R)
float * xf;        Returns the 'ideal' detector 'x' coordinate
                   in mm. (W)
float * yf;        Returns the 'ideal' detector 'y' coordinate
                   in mm. (W)
-end*/
{
   int i;
   float torad = 0.0001745329;
   float fac;
   float kfac;
   float xq, yq;
   float xp, yp;
   float cpsi, spsi;
   float rq;

   fac = torad/ctod;


   if (idtype==2)
   {
      kfac = 1.0;
      xp = xcmid + xfd;
      yp = ycmid + yfd;
      for (i=0;i<3;++i)
      {
         rq = sqrt(xp*xp+yp*yp);
         if (rq<0.0001)
         {
             cpsi = 1.0;
             spsi = 0.0;
         }
         else
         {
             cpsi = xp/rq;
             spsi = yp/rq;
         }
         xq = (xfd - 0.01*(-toff*spsi+roff*cpsi))/kfac;
         yq = (yfd - 0.01*(toff*cpsi+roff*spsi))/kfac;
         kfac = 1.0 + fac*(tilt*xq+twist*yq);
         xp = xcmid + kfac*xq;
         yp = ycmid + kfac*yq;
      }
   }
   else
   {
      kfac = 1.0;
      for (i=0;i<3;++i)
      {
         xq = xfd/kfac;
         yq = yfd/kfac;
         kfac = 1.0 + fac*(tilt*xq+twist*yq);
      }
   }


   *xf = xq*coswc + yq*sinwc;
   *yf = -xq*sinwc + yq*coswc;
   return;
}
/* Fortran binding: rdmf_ideal */

void rdmf_ideal (float * xfd, float * yfd, float * ctod, 
                float * coswc, float *sinwc,
                int *idtype, float * twist, float * tilt, 
                float * xcmid, float * ycmid, float * roff, float * toff,
                float * xf, float * yf)
{
   rdm_ideal (*xfd, *yfd, *ctod, *coswc, *sinwc, *idtype, *twist, *tilt,
            *xcmid, *ycmid, *roff, *toff, xf, yf);
   return;
}

/*-Routine: Convert mm to raster coordinates - rdm_mmtorast
The routine rdm_mmtorast (rdmf_mmtorast) converts distortion corrected
coordinates in mm to raster coordinates.
-end*/

/*
***************************
**     rdm_mmtoras       **
***************************

Purpose: Convert coordinates from mm to rasters

CD-Fortran:
         CALL RDMF_MMTORAS (XFD, YFD, X_CEN, Y_CEN, X_C, Y_C,
                            MM_RAST_X, MM_RAST_Y, XD, YD)
CD-end:
*/
/*-Parameters:
XFD         r  (R)  The corrected detector 'x' coordinate in mm.
YFD         r  (R)  The corrected detector 'y' coordinate in mm.
X_CEN       r  (R)  The 'x' centre position in rasters
Y_CEN       r  (R)  The 'y' centre position in rasters
X_C         r  (R)  The x_c centre offset in mm
Y_C         r  (R)  The y_c centre offset in mm
MM_RAST_X   r  (R)  Conversion factor for mm to x-rasters
MM_RAST_Y   r  (R)  Conversion factor for mm to y-rasters
XD          r  (W)  Returns the 'x' coordinate in rasters
YD          r  (W)  Returns the 'y' coordinate in rasters
-end*/

/*-C:*/
void rdm_mmtoras  (float xfd, float yfd, float x_cen, float y_cen,
                   float x_c, float y_c, float mm_rast_x, float mm_rast_y,
                   float * xd, float * yd)
/*end*/

/*-Parameters:
float xfd;       The corrected detector 'x' coordinate in mm. (R)    
float yfd;       The corrected detector 'y' coordinate in mm. (R) 
float x_cen;     The 'x' centre position in rasters (R)
float y_cen;     The 'y' centre position in rasters (R)
float x_c;       The x_c centre offset in mm (R)
float y_c;       The y_c centre offset in mm (R)
float mm_rast_x; Conversion factor for mm to x-rasters (R)
float mm_rast_y; Conversion factor for mm to y-rasters (R)
float *xd;       Returns the 'x' coordinate in rasters (W)
float *yd;       Returns the 'y' coordinate in rasters (W)
-end*/
{
   *xd = x_cen + (x_c + xfd)*mm_rast_x;
   *yd = y_cen + (y_c + yfd)*mm_rast_y;
   return;
}
/* Fortran binding: rdmf_mmtoras  */

void rdmf_mmtoras  (float * xfd, float * yfd, float * x_cen, float * y_cen,
                   float * x_c, float * y_c, 
                   float * mm_rast_x, float * mm_rast_y,
                   float * xd, float * yd)
{

   rdm_mmtoras  (*xfd, *yfd, *x_cen, *y_cen, *x_c, *y_c,
                 *mm_rast_x, *mm_rast_y, xd, yd);
   return;
}

/*-Routine: Convert raster to mm coordinates - rdm_rastomm
The routine rdm_rastomm (rdmf_rastomm) converts raster coordinates
in to measured coordinates in mm.
-end*/

/*
***************************
**     rdm_rastomm       **
***************************

Purpose: Convert coordinates from rasters to mm

CD-Fortran:
         CALL RDMF_RASTOMM(XD, YD, X_CEN, Y_CEN, X_C, Y_C,
                            MM_RAST_X, MM_RAST_Y, XFD, YFD)
CD-end:
*/
/*-Parameters:
XD          r  (R)  The 'x' coordinate in rasters
YD          r  (R)  The 'y' coordinate in rasters
X_CEN       r  (R)  The 'x' centre position in rasters
Y_CEN       r  (R)  The 'y' centre position in rasters
X_C         r  (R)  The x_c centre offset in mm
Y_C         r  (R)  The y_c centre offset in mm
MM_RAST_X   r  (R)  Conversion factor for mm to x-rasters
MM_RAST_Y   r  (R)  Conversion factor for mm to y-rasters
XFD         r  (W)  The measured detector 'x' coordinate in mm.
YFD         r  (W)  The measured detector 'y' coordinate in mm.
-end*/

/*-C:*/
void rdm_rastomm (float xd, float yd, float x_cen, float y_cen,
                   float x_c, float y_c, float mm_rast_x, float mm_rast_y,
                   float * xfd, float * yfd)
/*end*/

/*-Parameters:
float *xd;       The 'x' coordinate in rasters (R)
float *yd;       The 'y' coordinate in rasters (R)
float x_cen;     The 'x' centre position in rasters (R)
float y_cen;     The 'y' centre position in rasters (R)
float x_c;       The x_c centre offset in mm (R)
float y_c;       The y_c centre offset in mm (R)
float mm_rast_x; Conversion factor for mm to x-rasters (R)
float mm_rast_y; Conversion factor for mm to y-rasters (R)
float xfd;       The measured detector 'x' coordinate in mm. (W)    
float yfd;       The measured detector 'y' coordinate in mm. (W) 
-end*/
{
   *xfd = (xd - x_cen)/mm_rast_x - x_c;
   *yfd = (yd - y_cen)/mm_rast_y - y_c;
   return;
}
/* Fortran binding: rdmf_rastomm */

void rdmf_rastomm (float * xd, float * yd, float * x_cen, float * y_cen,
                   float * x_c, float * y_c, 
                   float * mm_rast_x, float * mm_rast_y,
                   float * xfd, float * yfd)
{

   rdm_rastomm (*xd, *yd, *x_cen, *y_cen, *x_c, *y_c,
                 *mm_rast_x, *mm_rast_y, xfd, yfd);
   return;
}

/*-Section: 'C' Routines for Matrix Handling
These are a series of routine for handling 3x3 matrices and 3 element
vectors.
-end*/

/*-Routine: Angle between two vectors - rdm_vangle
The routine rdm_vangl calculates the angle between two vectors.
-end*/

/*
***************************
**     rdm_vangl         **
***************************

Purpose: Angle between vectors
*/
/*-C:*/
float rdm_vangl (float v1[3], float v2[3])
/*end*/
/*-Parameters:
float v1[3];   The first vector (R)
float v2[3];   The seconf vector (r)

Return   The angle between the two vectors in degrees
-end*/
{
   float v1m;
   float v2m;
   float cs;
   float ang;

   v1m = sqrtf (v1[0]*v1[0]+v1[1]*v1[1]+v1[2]*v1[2]);
   v2m = sqrtf (v2[0]*v2[0]+v2[1]*v2[1]+v2[2]*v2[2]);
   cs = rdm_dot (v1,v2) / (v1m*v2m);
   if (cs>1.0) cs = 1.0;
   if (cs<-1.0) cs = -1.0;
   ang = acosf(cs);
   return 57.29578*ang;
}

/*-Routine: Dot product of two vectors - rdm_dot
The routine rdm_dot calculates the dot product of two vectors.
-end*/

/*
*************************
**     rdm_dot         **
*************************

Purpose: Dot product of two vectors
*/
/*-C:*/
float rdm_dot (float a[3], float b[3])
/*end*/
/*-Parameters:
float a[3];   The first vector (R)
float b[3];   The second vector (R)

Return   The dot product of the two vectors
-end*/
{
   return a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
}

/*-Routine: Cross product of two vectors - rdm_cross
The routine rdm_cross calculates the cross product of two vectors.
-end*/

/*
***************************
**     rdm_cross         **
***************************

Purpose: Cross product of two vectors
*/

/*-C:*/
void rdm_cross (float a[3], float b[3], float c[3])
/*end*/
/*-Parameters:
float a[3];   Returns the cross product of the two vectors (W)
float b[3];   The first vector (R)
float c[3];   The second vector (R)
-end*/
{
   /* compute vector product a = b x c */
   a[0] = b[1]*c[2] - c[1]*b[2];
   a[1] = b[2]*c[0] - c[2]*b[0];
   a[2] = b[0]*c[1] - c[0]*b[1];
   return;
}

/*-Routine: Invert a 3x3 matrix - rdm_minv3
The routine rdm_minv3 inverts a 3x3 matrix.
-end*/

  
/*
*************************
**     rdm_minv3       **
*************************

Purpose: Invert a 3x3 matrix
*/
/*-C:*/
void rdm_minv3 (float a[3][3], float b[3][3], float *d)
/*end*/
/*-Parameters:
float a[3][3];   Returns the inverted matrix (W)
float b[3][3];   The matrix to be inverted (R)
float *d;        Returns the determinant of the matrix (W)
-end*/
{
   int i;
   int j;
   float c[3][3];

   /* Invert a general 3x3 matrix a = (b)-1 and return determinant in d */
   rdm_cross (&c[0][0], &b[1][0], &b[2][0]);
   rdm_cross (&c[1][0], &b[2][0], &b[0][0]);
   rdm_cross (&c[2][0], &b[0][0], &b[1][0]);
   *d = rdm_dot (&b[0][0], &c[0][0]);
   if (fabsf(*d)>1.0e-30)
   { 
      for (i=0;i<3;++i)
      {
         for (j=0;j<3;++j)
	 {
            a[i][j] = c[j][i]/(*d);
	 }
      }
   }
   else
   {
      *d = 0.0;
   }
   return;
}

/*-Routine: Multiply two 3x3 matrices - rdm_matmul
The routine rdm_matmul calculates the product of two 3x3 matrices
-end*/
/*
*************************
**     rdm_matmul       **
*************************

Purpose: Multiply two 3x3 matrices
*/
/*-C:*/
void rdm_matmul (float a[3][3], float b[3][3], float c[3][3])
/*-Parameters:
float a[3][3];   Returns the product of the two matrices (W)
float b[3][3];   The first matrix (R)
float c[3][3];   The second matrix (R)
-end*/
{
   int i, j, k;

   for (i=0;i<3;++i)
   {
      for (j=0;j<3;++j)
      {
         a[i][j] = 0.0;
         for (k=0;k<3;++k) a[i][j] += b[i][k]*c[k][j];
      }
   }
   return;
}
/*-Routine: Multiply 3x3 matrix and vector - rdm_matvec
The routine rdm_matvec calculates the product of a 3x3 matrix and a vector
-end*/
  
/*
*************************
**     rdm_matvec      **
*************************

Purpose: Multiply 3x3 matrix and vector
*/
/*-C:*/
void rdm_matvec (float v[3], float a[3][3], float b[3])
/*end*/
/*-Parameters:
float v[3];     Returns the product of the matrix and vector (W)
float a[3][3];  The 3x3 matrix (R)
float b[3];     The vector (R)
-end*/
{
   float s;
   int i, j;

   for (i=0;i<3;++i)
   {
      s = 0.0;
      for (j=0;j<3;++j)
      {
         s += a[i][j]*b[j];
      }
      v[i] = s;
   }
   return;
}
/*-Routine: Scale a vector - rdm_scalev
The routine rdm_scalev scales a vector.
-end*/

  
/*
*************************
**     rdm_scalev      **
*************************

Purpose: Scale a vector
*/
/*-C:*/
void rdm_scalev (float a[3], float x, float b[3])
/*-Parameters:
float a[3];   Returns the scaled vector vector (W)
float x;      The scaling factor(R)
float b[3];   The vector  to be scaled(R)
-end*/
{
   int i;

   for (i=0;i<3;++i)
   {
      a[i] = b[i]*x;
   }
   return;
}

/*-Routine: Sum of two vectors - rdm_vadd
The routine rdm_vadd calculates the sum vector from two vectors.
-end*/
  
/*
*************************
**     rdm_vadd        **
*************************

Purpose: Add  two vctors
*/
/*-C:*/
void rdm_vadd (float a[3], float b[3], float c[3])
/*end*/
/*-Parameters:
float a[3];   Returns the vector sum (W)
float b[3];   The first vector (R)
float c[3];   The second vector (R)
-end*/
{
   int i;

   for (i=0;i<3;++i)
   {
      a[i] = b[i] + c[i];
   }
   return;
}

/*-Routine: Difference of two vectors - rdm_vdif
The routine rdm_vdif calculates the difference vector from two vectors.
-end*/
  
/*
*************************
**     rdm_vdif        **
*************************

Purpose: Difference of two vctors
*/
/*-C:*/
void rdm_vdif (float a[3], float b[3], float c[3])
/*end*/
/*-Parameters:
float a[3];   Returns the difference vector from two vectors (W)
float b[3];   The first vector (R)
float c[3];   The second vector (R)
-end*/
{
   int i;

   for (i=0;i<3;++i)
   {
      a[i] = b[i] - c[i];
   }
   return;
}

/*-Routine: Normalise a vector - rdm_vnorm
The routine rdm_vnorm normalises a vector.
-end*/
  
/*
*************************
**     rdm_vnorm       **
*************************

Purpose: Normalise a vector
*/
/*-C:*/
void rdm_vnorm (float a[3], float b[3])
/*end*/
/*-Parameters:
float a[3];   Returns the normalised vector (W)
float b[3];   The vector to normalise(R)
-end*/
{
   int i;
   float r;

   r = sqrt (b[0]*b[0]+b[1]*b[1]+b[2]*b[2]);
   if (r!=0.0)
   {
      for (i=0;i<3;++i)
      {
         a[i] = b[i]/r;
      }
   }
   return;
}


/*==================================================*/
/*         'Private' SERVICE ROUTINES               */
/*==================================================*/

/*
***************************
**     rdm_axord         **
***************************

Purpose: Interpret image axes order and byte swapping option string

Author:  John W. Campbell, November 1991

*/
int rdm_axord (char * str, int len, int * order, int * swaptyp)

 
/*
char *str;      String containing the axes order and byte swap flags (R)
                    
int len;        Length of the string; if 0 then assume a null
                terminated string was passed (R) 
int *order;     Returns the axis order given as a number from 1 to 8
                    (If none specified (or syntax error) return 0)
                     1   +xf slow   +yf fast
                     2   +xf slow   -yf fast
                     3   -xf slow   +yf fast
                     4   -xf slow   -yf fast
                     5   +yf slow   +xf fast
                     6   +yf slow   -xf fast
                     7   -yf slow   +xf fast
                     8   -yf slow   -xf fast
                     (W)
int *swaptyp;   Returns byte swap type   =0 do not swap bytes
                                         =1 swap bytes
                    (W)
Return:  Error flag =0 no error; 
                    >0 syntax error   =1 Only one axis defined.
                                      =2 Both axes the same.
                                      =3 More than two axes defined.
                                      =4 Byte swapping and no byte
                                         swapping defined.
*/
{
   int saxis, ssign;
   int faxis, fsign;
   int n;
   int l;
   char * p;
   char tmpstr[81];
   int err;

   /* Initialisations */

   if (len<=0) len=strlen(str);
   saxis = 0;
   ssign = 1;
   faxis = 0;
   fsign = 1;
   err = 0;
   *swaptyp = 0;
   *order = 0;
   if (len==0) return 0;
   tmpstr[0] = 0;
   l = len;
   if (l>80) l = 80;
   strncat (tmpstr,str,l);
   p = strtok (tmpstr," ");
   if (p!=0)
   {
      if (strcasecmp(p,"unset")==0) return 0;
   }
   n = 1;

   /* Decode the string */

   for (l=0;l<len;++l)
   {
      switch (str[l])
      {
         case '-':
            if (n==1) { ssign = -1; }else{ fsign = -1; }
            break;
         case '+':
            if (n==1) { ssign = 1; }else{ fsign = 1; }
            break;
         case 'x':
         case 'X':
            if (n==1) 
            {
               saxis = 1;
               n = 2;
            }
            else
            {
               if (n==3) { err=3; goto synterr; }
               faxis = 1;
               n = 3;
            }
            break;
         case 'y':
         case 'Y':
            if (n==1) 
            {
               saxis = 2;
               n = 2;
            }
            else
            {
               if (n==3) { err=3; goto synterr; }
               faxis = 2;
               n = 3;
            }
            break;
         case 's':
         case 'S':
            if (*swaptyp>0) { err=4; goto synterr; }
            *swaptyp = 2;
            break;
         case 'n':
         case 'N':
            if (*swaptyp>0) goto synterr;
            *swaptyp = 1;
            break;
      }
   }
   if (*swaptyp==0) *swaptyp = 1;
   *swaptyp -= 1;
   if (saxis==0&&faxis==0) return 0;
   if (saxis==0||faxis==0) { err=1; goto synterr; }
   if (saxis==faxis) { err=2; goto synterr; }
   *order = 4*saxis - 3;
   if (ssign<0) *order += 2;
   if (fsign<0) *order += 1;
   return 0;
   synterr:
   *order = 0;
   *swaptyp = 0;
   return err;
}
/*
**********************
**    rdm_symfr2    **
**********************

PURPOSE: Interpret symmetry operations

RETURN:  =0 Ok, =1 error

AUTHOR:  Based on CCP4 routine SYMFR2

**** RDM SERVICE ROUTINE ONLY ****

*/

int rdm_symfr2 (char * str, int i1, int * ns, float rot[][4][4],  
                char * errstr, int maxe)
/*
char *str;         String containing symmetry operators (R)
int i1;            No. of first column in 'str' containing the symmetry
                   operators data (R)
int * ns;          On input the number of 1'st symmetry operator to
                   store, on output the new number of symmetry operators
                   stored (R/W)
float rot[][4][4]; The symmetry matrix arrays (R/W)
char *errstr;      Error return string (W)
int maxe;          if >0 Maximum length of returned error string allowed
                         (excluding terminating null - errstr must be at
                         least maxe + 1 characters in length)
                   if <0 abs(maxe) characters will be returned padded
                         with blanks if needed (for 'fortran' use) 
                   (should allow 80 chars for error messages) (R)
*/
{
   char ich;
   int i, j, k;
   int icomst, ifound, len;
   int ip, isl, nop, np, nsym;
   float a, real, recip, s, t;
   static char inum[10] = {'0','1','2','3','4','5','6','7','8','9'};

   kdm_copy_chars (" ", 0, errstr, maxe);

   len = strlen (str);

   i = i1 - 2;
   *ns = *ns - 1;

   /* Interpret symmetry */

   lab30:
   *ns = *ns + 1;
   real = 0.0;
   recip = 0.0;
   nop = 1;
   for (j=0;j<4;++j)
   {
      for(k=0;k<4;++k) rot[*ns-1][j][k] = 0.0;
   }   
   rot[*ns-1][3][3] = 1.0;

   lab60:

   s = 1.0;

   /* Set j=4 for translation vector */

   j = 4;
   t = 0.0;
   ip = 0;
   np = 0;
   isl = 0;
   ifound = 0;
   icomst = 0;

   lab70:

   i = i + 1;
   if (i< len)
   {
     ich = str[i];
     if (ich==' ') goto lab70;
     if (ich!=','&&ich!='*')
     {
       ifound = 1;
       if (ich=='X'||ich=='x'||ich=='H'||ich=='h')
       {
          if (ich=='X'||ich=='x') real += 1.0;
          if (ich=='H'||ich=='h') recip += 1.0;
          j = 1;
          if (t==0.0) t = s;
          goto lab70;
       }
       else if (ich=='Y'||ich=='y'||ich=='K'||ich=='k')
       {
          if (ich=='Y'||ich=='y') real += 1.0;
          if (ich=='K'||ich=='k') recip += 1.0;
          j = 2;
          if (t==0.0) t = s;
          goto lab70;
       }
       else if (ich=='Z'||ich=='z'||ich=='L'||ich=='l')
       {
          if (ich=='Z'||ich=='z') real += 1.0;
          if (ich=='L'||ich=='l') recip += 1.0;
          j = 3;
          if (t==0.0) t = s;
          goto lab70;
       }
       else if (ich=='+')
       {
          s = 1.0;
          if (t==0.0&&j==4)
          {
             goto lab70;
          }
          else
          {
             goto lab100;
          }
       }
       else if (ich=='-')
       {
          s = -1.0;
          if (t==0.0&&j==4)
          {
             goto lab70;
          }
          else
          {
             goto lab100;
          }
       }
       else if (ich=='/')
       {
          isl = 1;
          goto lab70;
       }
       else if (ich=='.')
       {
          ip = 1;
          goto lab70;
       }
       else
       {
          for (k=0;k<10;++k)
          {
             if (ich==inum[k]) goto lab90;
          }
          kdm_copy_chars ("Symmetry error - invalid character", 0, 
                    errstr, maxe);
          return 1;
       
          lab90:
          a = (float) k;
          if (isl==1)
          {
             t = t/a;
          }
          else if (ip==1)
          {
             np +=1;
             t = s*a/(powf(10.0,(float)np)) + t;
          }
          else
          {
             t = 10.0*t + a*s;
          }
          goto lab70;
       }
    }
  }
  if (t==0.0&&j==4)
  {
     goto lab110;
  }
  else
  {
    icomst = 1;
  }
  
  lab100:
  rot[*ns-1][nop-1][j-1] = t;   
  j = 4;
  t = 0.0;
  ip = 0;
  np = 0;
  isl = 0;
  if (icomst==0) goto lab70;
  if (ifound==0&&i<len)
  {
     kdm_copy_chars ("Symmetry error - blank operator ", 0, 
                    errstr, maxe);
     return 1;
  }
  
  if (i<len)
  {
     nop ++;
     if (nop<=3)
     {
        goto lab60;
     }
     else
     {
        goto lab30;
     }
  }
  else
  {
     goto lab120;
  }

  lab110:
  kdm_copy_chars ("Symmetry error - no operator ", 0, 
                    errstr, maxe);
  return 1;

  lab120:
  if (nop!=1||ifound!=0)
  {
     if (nop==3&&ifound==1)
     {
        goto lab130;
     }
     else
     {
        kdm_copy_chars ("Symmetry error - incomplete operator ", 0, 
                    errstr, maxe);
        return 1;
     }
  }
  *ns = *ns - 1;

  lab130:
  nsym = *ns;
  if (real<3.0&&recip<3.0)
  {
     kdm_copy_chars ("Symmetry error - incomplete definition", 0, 
                    errstr, maxe);
     return 1;
  }
  if (recip>=3.0) nsym = - nsym;
  return 0;
}

/*
*************************
**    rdm_read_symm    **
**************************

PURPOSE: Read symmetry data from SYMOP#include "kdm.h"


RETURN:  =0 Ok, =1 error

**** RDM SERVICE ROUTINE ONLY ****

   Sets up globals  nspgrp
                    nsym
                    nsymp
                    spgnam
                    pgnam
                    systyp
                    rsym
*/

int rdm_read_symm (int lspgrp, char * namspg, char * errstr, int maxe)
/*
int lspgrp;        Space group no. to be found (if > 0) (R)
char * namspg;     Name of space group to be found (R)
                   Note: Specify one of the above two (if both given,
                         the space group number will be used.
char *errstr;      Error return string (W)
int maxe;          if >0 Maximum length of returned error string allowed
                         (excluding terminating null - errstr must be at
                         least maxe + 1 characters in length)
                   if <0 abs(maxe) characters will be returned padded
                         with blanks if needed (for 'fortran' use) 
                   (should allow 80 chars for error messages) (R)
*/
{
   int i;
   int l;
   int il;
   int m1, m2;
   int isg;
   int err;
   int nlin;
   int nlins;
   char *env;
   char line[121];
   char tmpstr[121];
   FILE *fp_symm;

   kdm_copy_chars (" ", 0, errstr, maxe);

   /* open symmenty file */

   env = getenv("SYMOP");
   if (env==0)
   {
      kdm_copy_chars ("Error opening symmetry operators file (SYMOP ", 
                      0, errstr, maxe);
      return 1;
   }
   fp_symm = fopen (env,"r");
   if (fp_symm==0)
   {
      kdm_copy_chars ("Error opening symmetry operators file (SYMOP ", 
                      0, errstr, maxe);
      return 1;
   }
   fseek (fp_symm, 0L, SEEK_SET);

   nsym = 0;

   /* Find correct space-group in file.
      Each space-group has header line of space-group number,
      number of line of symmetry operations for non-primitive
      and primitive cells. */

      while (fgets (line, 121, fp_symm))
      {
         l = strlen(line);
         if (l>0)
         {
            kdm_upc (line);
            if (line[l-1]=='\n') line[l-1] = 0;
         }
         if (line[0]=='!') continue;
         if (kdm_blank_string (line, 0)) continue;

/* Fields are space group number,
              number of lines,
              number of lines in primitive cell symmetry,
              spacegroup name 
              pointgroup name
              crystal system type */
      il = -1;
      i = kdm_next_field (line, 0, " ,", &m1, &m2, &il);
      if (i<=0) goto err2;
      tmpstr[0] = 0;
      strncat (tmpstr, &line[m1], m2-m1+1);
      i = sscanf (tmpstr,"%d",&isg);
      if (i<=0) goto err2;
      i = kdm_next_field (line, 0, " ,", &m1, &m2, &il);
      if (i<=0) goto err2;
      tmpstr[0] = 0;
      strncat (tmpstr, &line[m1], m2-m1+1);
      i = sscanf (tmpstr,"%d",&nlin);
      if (i<=0) goto err2;
      i = kdm_next_field (line, 0, " ,", &m1, &m2, &il);
      if (i<=0) goto err2;
      tmpstr[0] = 0;
      strncat (tmpstr, &line[m1], m2-m1+1);
      i = sscanf (tmpstr,"%d",&nlins);
      if (i<=0) goto err2;
      i = kdm_next_field (line, 0, " ,", &m1, &m2, &il);
      if (i<=0) goto err2;
 
      /* See if this space group */

      if (lspgrp>0)
      {
         if (lspgrp==isg) goto found;
      }
      else
      {
         tmpstr[0] = 0;
         strncat (tmpstr, &line[m1], m2-m1+1);
         if (strcasecmp(namspg,tmpstr)==0) goto found;
      }

      /* Not this one, skip 'nlin' lines */

      for (i=0;i<nlin;++i) fgets (line, 121, fp_symm);
   }
   kdm_copy_chars ("No symmetry in SYMOP file for requested space group", 
                      0, errstr, maxe);
   return 1;

   /* Space group found - read symmetry operators */

   found:
   nspgrp = isg;
   spgnam[0] = 0;
   l = m2-m1+1;
   if (l>10)l =10;
   strncat (spgnam, &line[m1], l);
   pgnam[0] = 0;   
   i = kdm_next_field (line, 0, " ,", &m1, &m2, &il);
   if (i>0)
   {
      l = m2-m1+1;
      if (l>10)l =10;
      strncat (pgnam, &line[m1], l);
   }
   i = kdm_next_field (line, 0, " ,", &m1, &m2, &il);
   if (i>0)
   {
      l = m2-m1+1;
      if (l>12)l =12;
      strncat (systyp, &line[m1], l);
   }
   for (i=0;i<nlins;++i)
   {
      fgets (line, 121, fp_symm);
      nsym ++;
      l = strlen(line);            
      if (line[l-1]=='\n') line[l-1] = 0;
      err = rdm_symfr2 (line, 1, &nsym, rsym, errstr, maxe);
      if (err!=0) return err;
   }
   nsymp = nsym;
   if (nlin>nlins)
   {
      for (i=nlins+1;i<=nlin;++i)
      {
         fgets (line, 121, fp_symm);
         nsym ++;
         l = strlen(line);            
         if (line[l-1]=='\n') line[l-1] = 0;
         err = rdm_symfr2 (line, 1, &nsym, rsym, errstr, maxe);
         if (err!=0) return err;
      }
   }
   fclose (fp_symm);
   return 0;


   err2:
   kdm_copy_chars ("Error in format of SYMOP file", 0, errstr, maxe);
   return 1;
}

/*
*************************
**    rdm_pgdefn       **
**************************

PURPOSE: Find point group from symmetry operators

RETURN:  =0 Ok, =1 error

**** RDM SERVICE ROUTINE ONLY ****

   Sets up globals  icentr, iscrew, klass

*/

int rdm_pgdefn(int nsym, float rsmt[][4][4], char * nampg, int * nsymmp)
/*
int nsym;            No. of symmetry oprators (R)
float rsmt[][4][4];  The symmetry operators (may be reordered to put
                     the primitive operators first (R/W)
char nampg[11];      Returns the point group name (W)
int * nsympp;        Returns the number of primitive operators (W)
*/
{
   int i, j, n;
   int ii, jj;
   int ism1, ism2;
   int isym;
   int jsm;
   int nsp; 
   int in[3];
   int nrep, nrepet;
   int juniqu;
   int iraxis;
   int ihr, ikr, ilr;
   int ih, ik, il;
   int irot, jrot;
   int iscr;
   int jmin;
   int irmin;
   int nrepp[192], norig[192];
   int nrot[192], nrots[192], jrots[192];
   float rjunk[192][4][4];
   float det, dx, dxy, dxyz, dxz, dy, dyz, dz;
   float aa[4][4];
   for (n=0;n<nsym;++n)
   {
      nrepp[n] = 0;
      for (i=0;i<4;++i)
      {
         for (j=0;j<4;++j) rjunk[n][i][j] = rsmt[n][i][j];
      }
   }
   nsp = 0;

   /* Which sym ops belong to primitive set? Check 1 - nsym-1 first. */

   if (nsym>1)
   {

      for (ism1=1;ism1<=nsym-1;++ism1)
      {
         /* We have already dealt with this Symm op. */
         if (rsmt[ism1-1][3][3]!=0)
         {
            nrep = 0;
            nsp += 1;
            norig[ism1-1] = nsp;

            /* Search all remaining ones for duplicates */
            for (ism2=ism1+1;ism2<=nsym;++ism2)
            {
               for (i=0;i<3;++i)
               {
                  for (j=0;j<3;++j)
                  {
                     if (rsmt[ism2-1][i][j]!=rsmt[ism1-1][i][j]) goto lab60;
                  }
               }
         
            /* This is a duplicate - count how often there is repetition and
               record which sym op it is a duplicate of. 
               Modify rsmt(4,4,ism2) to act as a flag. */

               nrep = nrep + 1;
               nrepp[ism2-1] = nrep;
               norig[ism2-1] = norig[ism1-1];
               rsmt[ism2-1][3][3] = 0.0;
               lab60:
               continue;
            }
         }
      }
   }

   /* Check last symmetry operator now. */

   if (rsmt[nsym-1][3][3]!=0.0)
   {
      nrep = 0;
      nsp += 1;
      norig[nsym-1] = nsp;
   }

   /* Now rewrite symmetry with all primitives first and define centring. */
   /* If nsym=nsp   you have a primitive spacegroup. */

   nrepet = nsym/nsp;
   icentr = 0;
   if (nrepet!=1)
   {
   /* If nrepet == 4 you must have FACE centring. (icentr = 4) */
      if (nrepet==4) icentr = 4;

   /* If nrepet == 3 you must have Rhombehedral centring. (icentr = 6) */

      if (nrepet==3) icentr = 6;

   /* If nrepet ==2 you may have A B C or I centring.

      Check translation components for appropriate
      symmetry axes after reordering. */

      isym =  0;
      for (n=0;n<nsym;++n)
      {
         isym = nrepp[n]*nsp + norig[n];

         for (i=0;i<4;++i)
         {
            for (j=0;j<4;++j) rsmt[isym-1][i][j] = rjunk[n][i][j];
         }
         rsmt[n][3][3] = 1.0;
      }
   }

   /* More centring. */

   /*If nrepet==2 you may have A B C or I centring. */

   if (nrepet==2)
   {
      dx = fabsf(rsmt[0][0][3]-rsmt[nsp][0][3]);
      dy = fabsf(rsmt[0][1][3]-rsmt[nsp][1][3]);
      dz = fabsf(rsmt[0][2][3]-rsmt[nsp][2][3]);

      /* If dx dy and dz are all = 0.5 you must have body centring. 
         (icentr = 5) */

     dxyz = dx*dy*dz;
     if (dxyz>0.1)
     {
        icentr = 5;
     }
     else
     {
        dxy = dx*dy;
        dxz = dx*dz;
        dyz = dy*dz;

        /* If dy and dz = 0.5 and dx=0 you must have A centring (icentr = 1) */

        if (dyz>0.2) icentr = 1;

        /* If dx and dz = 0.5 and dy=0 you must have B centring (icentr = 2) */

        if (dxz>0.2) icentr = 2;

        /* If dx and dy = 0.5 and dz=0 you musthave C centring (icentr = 3) */

        if (dxy>0.2) icentr = 3;
      }
   }

   /* Find whether 2fold 3 fold etc and define ISCREW(...)
      use general reflection  - taken as (1,4,8) */

   in[0] = 1;
   in[1] = 4;
   in[2] = 8;

   /* Generate primitive symm equivs and look at rotations. 

      Test whether h' k' l' equals  h  k  l

      Choose rotation axis - either a(3) b(2) c(1)  or
                             lines 1 1 0(4)  or  1 1 1(5)
   

      Define ISCREW[i] = 0,2,3,4,6
   */

   iscrew[0] = 0;
   iscrew[1] = 0;
   iscrew[2] = 0;
   jsm = 0;

   for (j=0;j<nsp;++j)
   {
      /* Is it a rotation at all? */

      for (ii=0;ii<4;++ii)
      {
         for (jj=0;jj<4;++jj) aa[ii][jj] = rsmt[j][ii][jj];
      }
      rdm_determ (aa, &det);
      
      if (fabsf(det-1.0)<0.00001)
      {
         nrot[j] = 0;
         iraxis = 0;

         /* Unique rotation axes must be a b c or the lines 1 1 0 or 1 1 1 */

         /* a */

         if (rsmt[j][0][0]==1.0) iraxis = 3;

         /*  b */

         if (rsmt[j][1][1]==1.0) iraxis = 2;

         /* c */

         if (rsmt[j][2][2]==1.0) iraxis = 1;

         /* 1 1 0 */

         if (rsmt[j][1][0]==1.0&&rsmt[j][0][1]==1.0&&
             rsmt[j][2][2]==-1.0) iraxis = 4;

         /* 1 1 1 */

         if (rsmt[j][1][0]==1.0&&rsmt[j][2][1]==1.0&&
             rsmt[j][0][2]==1.0) iraxis = 5;

         /* 1 1 1 */

         if (rsmt[j][2][0]==1.0&&rsmt[j][0][1]==1.0&&
             rsmt[j][1][2]==1.0) iraxis = 5;

         ihr = in[0];
         ikr = in[1];
         ilr = in[2];

         /* Ignore anything else eg -1 1 0 - they are consequences of others */

         if (iraxis!=0) 
         {

            /* What sort of rotation ? 2fold? 3fold? 4fold? 6fold? */
 
            jsm = jsm + 1;
            nrot[jsm-1] = 0;

            for(irot=0;irot<6;++irot)
            {
               ih = rsmt[j][0][0]*ihr + rsmt[j][1][0]*ikr + rsmt[j][2][0]*ilr;
               ik = rsmt[j][0][1]*ihr + rsmt[j][1][1]*ikr + rsmt[j][2][1]*ilr;
               il = rsmt[j][0][2]*ihr + rsmt[j][1][2]*ikr + rsmt[j][2][2]*ilr;

               /*  Back to h k l - how many rotations to get here? */

               if (ih==in[0]&&ik==in[1]&&il==in[2])
               {
                  goto lab120;
               }
               else
               {

               /* Apply symmetry again */

                  ihr = ih;
                  ikr = ik;
                  ilr = il;
               }
            }

            goto lab130;

            lab120:
            nrot[jsm-1] = 10*iraxis + irot+1;

            /* Check screwiness for a b and c axes -
               keep the highest order of it.
               (a 6 fold axis will also produce a 2 fold and a 3 fold...) */

            if (iraxis<=3)
            {
               iscr = 0;
               jrot = 4 - iraxis;

               /* Test translation component */

               if (rsmt[j][3][jrot-1]!=0.0) iscr = nint(1.0/
                   fabsf(rsmt[j][3][jrot-1]));

               /* If that one was 0 try this "translation component" */

               if (iscr==0&&rsmt[j][jrot-1][3]!=0.0) 
                         iscr = nint(1.0/fabsf(rsmt[j][jrot-1][3]));
               if (iscr>iscrew[jrot-1]) iscrew[jrot-1] = iscr;
            }
         }
      }
      lab130:
      continue;
   }

   /* Sort rotation info - nrot(...) gt 50 means iraxis = 5 means cubic etc. */

   for (i=0;i<jsm;++i)
   {
      irmin = 1000000;
      jmin = 0;
      for (j=0;j<jsm;++j)
      {
         if (irmin>nrot[j])
         {
            jmin = j;
            irmin = nrot[j];
         }
      }
      nrots[i] = irmin;
      jrots[i] = jmin+1;
      nrot[jmin] = 1000001;
   }

   /* Get rid of  duplications */

   juniqu = 1;

   for (i=2;i<=jsm;++i)
   {
      if (nrots[i-1]!=nrots[juniqu-1])
      {
         juniqu = juniqu + 1;
         nrots[juniqu-1] = nrots[i-1];
         jrots[juniqu-1] = jrots[i-1];
      }
   }

   /* Choose point group

      Rarest first

      Cubic
      14 pg23   m3        KLASS 13
      15 pg432  m3m       KLASS 14
   */
   if (nrots[juniqu-1]>50)
   {
      if (nrots[1]==12)
      {
         strcpy(nampg,"PG23");
         klass = 13;
         if (nrots[2]==14)
         {
            strcpy(nampg,"PG432");
            klass = 14;
         }
      }

   /*  Space groups with 2 fold axes along 1 1 0
       8 pg422 4/mmm       KLASS 5
       11 pg321  3/m        KLASS 9
       13 pg622 6/mmm       KLASS 12
   */
   }
   else if (nrots[juniqu-1]>40)
   {
      if (nrots[1]==12&&nrots[2]==14)
      {
         strcpy(nampg,"PG422");
         klass = 5;
      }
      if (nrots[1]==13)
      {
         strcpy(nampg,"PG321");
         klass = 9;
      }
      if (nrots[1]==12&&nrots[2]==13)
      {
         strcpy(nampg,"PG622");
         klass = 12;
      }

   /*  6 pg222  mmm     KLASS = 3
       11 pg312  3/m     KLASS = 8
       ?? pg2a   2       KLASS = 16    monoclinic A unique.
   */
   }
   else if (nrots[juniqu-1]>30)
   {
      if (nrots[1]==13)
      {
         strcpy(nampg,"PG312");
         klass = 8;
      }
      else if (nrots[1]==12&&nrots[2]==22)
      {
         strcpy(nampg,"PG222");
         klass = 3;
      }
      else if (nrots[juniqu-1]==32)
      {
         strcpy(nampg,"PG2A");
         klass = 16;
      }

   /*  4 pg2    2/m */
   }
   else if (nrots[juniqu-1]==22)
   {
      strcpy(nampg,"PG2");
      klass = 2;

   /*  3 pg1     1bar          KLASS 1
       5 pg2c    2/m  C unique KLASS 15
       7 pg4    4/m            KLASS 4
       9 pg3     3bar          KLASS 6
      12 pg6    6/m            KLASS 11
   */
   }
   else if (nrots[juniqu-1]<20)
   {
      if (nrots[juniqu-1]==11) strcpy(nampg,"PG1");
      if (nrots[juniqu-1]==11) klass = 1;
      if (nrots[juniqu-1]==12) strcpy(nampg,"PG2C");
      if (nrots[juniqu-1]==12) klass = 15;
      if (nrots[juniqu-1]==13) strcpy(nampg,"PG3");
      if (nrots[juniqu-1]==13) klass = 6;
      if (nrots[juniqu-1]==14) strcpy(nampg,"PG4");
      if (nrots[juniqu-1]==14) klass = 4;
      if (nrots[juniqu-1]==16) strcpy(nampg,"PG6");
      if (nrots[juniqu-1]==16) klass = 11;
   }
   *nsymmp = nsp;
   return 0;
}
/*
**********************
**    rdm_opstr     **
**********************

PURPOSE: Output string via user supplied function

**** RDM SERVICE ROUTINE ONLY ****

*/
void rdm_opstr (char * linout, char * outstr, int maxo, void (*op_func)())
{
   if (strlen(linout)!=0)
   {
      if (op_func)
      {
         kdm_copy_chars (linout, 0, outstr, maxo);
         op_func();
      }
      else
      {
          printf ("%s\n",linout);
      }
   }
   return;
}
/* inasu.f -- translated by f2c (version 19961001).
   You must link the resulting object file with the libraries:
	-lf2c -lm   (in that order)
*/
/*
*************************
**    rdm_pgnlau        **
**************************

PURPOSE: Choose Laue group from Point Group name

AUTHOR: Translated from CCP4 routine PGNLAU

**** RDM SERVICE ROUTINE ONLY ****
*/

void rdm_pgnlau (char * nampg, int * nlaue, char * launam)
/*
char * nampg;    The point group name (R)
int * nalue;     Returns the Laue grou number (0 if not found) (W)
char * launam;   Returns the laue group name if found
*/
{
   char locnam[21];

   hrng0 = -9999;
   krng0 = -9999;
   lrng0 = -9999;
   hrng1 =  9999;
   krng1 =  9999;
   lrng1 =  9999;
   *nlaue = 0;

   strcpy (locnam, nampg);
   kdm_upc (locnam);
   if (strncmp(locnam,"PG",2)==0)
   {
      locnam[0] = 0;
      strcat (locnam, &nampg[2]);
      kdm_upc (locnam);
   }
   /* Rarest first
    
      Cubic
      14 pg23   pgm3bar - Laue m3
      15 pg432  pg4bar3m pgm3barm - Laue m3m
      
      14 pg23    m3         hkl:h>=0, k>=0, l>=0 with l>=h,  k>=h
      pgm3bar
   */

   if (strcmp(locnam,"23")==0||strcmp(locnam,"M3BAR")==0)
   {
      *nlaue = 14;
      strcpy (launam,"M3BAR");
      hrng0 = 0;
      krng0 = 0;
      lrng0 = 0;

   /* 15 pg432   m3m        hkl:h>=0, k>=0, l>=0  with  k>=l
      pg4bar3m pgm3barm
   */

   }
   else if (strcmp(locnam,"432")==0||strcmp(locnam,"4BAR3M")==0||
         strcmp(locnam,"M3BARM")==0)
   {
      *nlaue = 15;
      hrng0 = 0;
      krng0 = 0;
      lrng0 = 0;
      strcpy(launam, "m3barm");
       
   /*   8 pg422 pg4mm pg4bar2m pg4barm2 - Laue 4/mmm
        11 pg321 pg32 pg3m1 pg3barm1 pg3m - Laue  3/m
        hkl:h>=0, k>=0 with k<=h for all l.
        if h = k  l>=0
        Space group numbers :   150-152-154
        13 pg622 pg6mm pg6bar2m pg6barm2 pg6/mmm - Laue 6/mmm
       
        8 pg422   4/mmm       hkl:h>=0, k>=0, l>=0            89..
        pg4mm pg4bar2m pg4barm2 pg4/mmm
   */

   }
   else if (strcmp(locnam,"422")==0||strcmp(locnam,"4/MMM")==0||
         strcmp(locnam,"4MM")==0||strcmp(locnam,"4BAR2M")==0||
         strcmp(locnam,"4BARM2")==0)
   {
      hrng0 = 0;
      krng0 = 0;
      lrng0 = 0;
      *nlaue = 8;
      strcpy(launam, "4/mmm");

   /*  11 pg321   3bar1m     hkl:h>=0, k>=0 with k<=h for all l.
       pg31m pg3bar1m      if h = k  l>=0
       Space group numbers :   150-152-154
   */

   }
   else if (strcmp(locnam,"321")==0||strcmp(locnam,"32")==0||
         strcmp(locnam,"3M1")==0||strcmp(locnam,"3BARM1")==0||
         strcmp(locnam,"3BARM")==0||strcmp(locnam,"3M")==0)
   {
      *nlaue = 11;
      hrng0 = 0;
      krng0 = 0;
      strcpy (launam, "3barm");

  /*  13 pg622   6/mmm       hkl:h>=0, k>=0, l>=0 with h>=k 177..
      pg6mm pg6barm2 pg6bar2m  pg 6/mmm
  */

   }
   else if (strcmp(locnam,"622")==0||strcmp(locnam,"6MM")==0||
         strcmp(locnam,"6BAR2M")==0||strcmp(locnam,"6BARM2")==0||
         strcmp(locnam,"6/MMM")==0)
   {
      *nlaue = 13;
      hrng0 = 0;
      krng0 = 0;
      lrng0 = 0;
      strcpy (launam, "6/mmm");
       
   /* 10 pg312  pg31m pg3bar1m - Laue 3bar1m 
      hkl:h>=0, k>=0 with k<=h for all l.
      if k = 0  l>=0
      Space group numbers :   149-151-153
       
      10 pg312   3/m        hkl:h>=0, k>=0 with k<=h for all l.
      pg32 pg3m pg3m1 pg3barm1 if k = 0  l>=0
      Space group numbers :   149-151-153 157 159 162 163
   */

   }
   else if (strcmp(locnam,"312")==0||strcmp(locnam,"31M")==0||
         strcmp(locnam,"3BAR1M")==0)
   {
      *nlaue = 10;
      hrng0 = 0;
      krng0 = 0;
      strcpy (launam, "3bar1m");

   /* 6 pg222   mmm        hkl:h>=0, k>=0, l>=0            16 ...
      pgmm2 pgmmm
   */

   }
   else if (strcmp(locnam,"222")==0||strcmp(locnam,"MMM")==0||
         strcmp(locnam,"MM2")==0||strcmp(locnam,"2MM")==0||
         strcmp(locnam,"M2M")==0)
   {
      *nlaue = 6;
      hrng0 = 0;
      krng0 = 0;
      lrng0 = 0;
      strcpy (launam, "mmm");

   /* 5 pg2 pgm pg2/m - Laue   2/m    C axis unique

      5 pg2     2/m        hkl:k>=0, l>=0  h0l:h>=0       1003,1004
      pgm pg2/m
   */

   }
   else if (strcmp(locnam,"2C")==0)
   {
      *nlaue = 5;
      hrng0 = 0;
      krng0 = 0;
      strcpy (launam, "2/m");
       
   /* 4 pg2 pgm pg2/m - Laue   2/m
      
      4 pg2     2/m        hkl:k>=0, l>=0  hk0:h>=0       3/b,4/b....
      pgm pg2/m
   */

   }
   else if (strcmp(locnam,"2")==0||strcmp(locnam,"M")==0||
         strcmp(locnam,"2/M")==0)
   {
      *nlaue = 4;
      krng0 = 0;
      lrng0 = 0;
      strcpy (launam, "2/m");
       
   /* 3 pg1     1bar
      7 pg4    4/m
      9  pg3     3bar
      12 pg6    6/m
      
      3 pg1     1bar      hkl:l>=0  hk0:h>=0  0k0:k>=0   1,2
      pg1bar
   */

   }
   else if (strcmp(locnam,"1")==0||strcmp(locnam,"1BAR")==0)
   {
      *nlaue = 3;
      lrng0 = 0;
      strcpy (launam, "-1");

   }
   else if (strcmp(locnam,"3")==0||strcmp(locnam,"3BAR")==0)
   {
      *nlaue = 9;
      hrng0 = 0;
      krng0 = 0;
      strcpy (launam, "-3");

   /* 7 pg4     4/m        hkl:h>=0, l>=0 with k>=0 if  h=0  and
      pg4bar pg4/m                            k>0 if h>0
   */

   }
   else if (strcmp(locnam,"4")==0||strcmp(locnam,"4/M")==0||
         strcmp(locnam,"4BAR")==0)
   {
      *nlaue = 7;
      hrng0 = 0;
      krng0 = 0;
      lrng0 = 0;
      strcpy (launam, "4/m");

   /* 12 pg6     6/m        hkl:h>=0, k>=0, l>=0 with k>=0 if  h=0
      pg6bar  6/m        and k> 0 if h>0
   */

   }
   else if (strcmp(locnam,"6")==0||strcmp(locnam,"6/M")==0||
         strcmp(locnam,"6BAR")==0)
   {
      *nlaue = 12;
      strcpy (launam, "6/m");
      hrng0 = 0;
      krng0 = 0;
      lrng0 = 0;
   }
   return;
}

/*
*************************
**    rdm_inasu        **
**************************

PURPOSE: See if reflection in asymmetric unit

RETURN:   +1  if  h k l chosen 
          -1  if -h-k-l chosen 
           0  if reflection is out-of-bounds 

AUTHOR: Translated from CCP4 routine INASU  (mostly using f2c)

**** RDM SERVICE ROUTINE ONLY ****
*/

int rdm_inasu (int * ih, int nlaue)
/*
int ih[3];   The reflection indices (R)
int nalue;   The Laue group number
*/
{
   int ret_val, i__1;
   int isgn, j, k, l;

    j = ih[0];
    k = ih[1];
    l = ih[2];
    isgn = 1;

L1:
    switch (nlaue) {
	case 1:  goto L100;
	case 2:  goto L150;
	case 3:  goto L200;
	case 4:  goto L250;
	case 5:  goto L300;
	case 6:  goto L350;
	case 7:  goto L400;
	case 8:  goto L450;
	case 9:  goto L550;
	case 10:  goto L570;
	case 11:  goto L600;
	case 12:  goto L650;
	case 13:  goto L700;
	case 14:  goto L750;
	case 15:  goto L800;
    }
L100:
/* ----1 BAR (ALTERNATE 1) */
/*     HKL --    H.GE.0 */
/*     0KL --    K.GE.0 */
/*     00L --    L.GE.0 */
    if (j < 0) {
	goto L7;
    } else if (j == 0) {
	goto L105;
    } else {
	goto L5;
    }
L105:
    if (k < 0) {
	goto L7;
    } else if (k == 0) {
	goto L110;
    } else {
	goto L5;
    }
L110:
    if (l >= 0) {
	goto L5;
    } else {
	goto L7;
    }

L150:
/* ----1 BAR (ALTERNATE I) */
/*     HKL --    K.GE.0 */
/*     H0L --    L.GE.0 */
/*     H00 --    H.GE.0 */
    if (k < 0) {
	goto L7;
    } else if (k == 0) {
	goto L155;
    } else {
	goto L5;
    }
L155:
    if (l < 0) {
	goto L7;
    } else if (l == 0) {
	goto L160;
    } else {
	goto L5;
    }
L160:
    if (j >= 0) {
	goto L5;
    } else {
	goto L7;
    }

/*  Corresponds to Data reduction unique set for  pg1 */
/*   3 pg1     1bar      hkl:l>=0  hk0:h>=0  0k0:k>=0   1,2 */
L200:
/* ----1 BAR (ALTERNATE 3) */
/*     HKL --    L.GE.0 */
/*     HK0 --    H.GE.0 */
/*     0K0 --    K.GE.0 */
    if (l < 0) {
	goto L7;
    } else if (l == 0) {
	goto L205;
    } else {
	goto L5;
    }
L205:
    if (j < 0) {
	goto L7;
    } else if (j == 0) {
	goto L210;
    } else {
	goto L5;
    }
L210:
    if (k >= 0) {
	goto L5;
    } else {
	goto L7;
    }

/*  Corresponds to Data reduction unique set for  pg2 */
/*   4 pg2    2/m        hkl:k>=0, l>=0  hk0:h>=0       3/b,4/b.... */
L250:
/* ----2/M (ALTERNATE 1) */
/*     HKL --    K.GE.0 AND L.GE.0 */
/*     HK0 --    H.GE.0 */
    if (k >= 0) {
	goto L255;
    } else {
	goto L7;
    }
L255:
    if (l < 0) {
	goto L7;
    } else if (l == 0) {
	goto L260;
    } else {
	goto L5;
    }
L260:
    if (j >= 0) {
	goto L5;
    } else {
	goto L7;
    }

/*  Corresponds to Data reduction unique set for  pg2 */
/*   5 pg2    2/m        hkl:k>=0, l>=0  h0l:h>=0       1003,1004 */
L300:
/* ----2/M (ALTERNATE 2) */
/*     HKL --    L.GE.0 AND H.GE.0 */
/*     H0L --    H.GE.0 */
    if (l >= 0) {
	goto L305;
    } else {
	goto L7;
    }
L305:
    if (k < 0) {
	goto L7;
    } else if (k == 0) {
	goto L310;
    } else {
	goto L5;
    }
L310:
    if (j >= 0) {
	goto L5;
    } else {
	goto L7;
    }

/*  Corresponds to Data reduction unique set for  pg222 */
/*   6 pg222  mmm        hkl:h>=0, k>=0, l>=0            16 ... */
L350:
/* ----MMM */
/*     HKL --    H.GE.0, K.GE.0, AND L.GE.0 */
    if (j >= 0) {
	goto L355;
    } else {
	goto L7;
    }
L355:
    if (k >= 0) {
	goto L361;
    } else {
	goto L7;
    }
/* ****NO 3-SICKLIES PLEASE */
L361:
    if (l >= 0) {
	goto L5;
    } else {
	goto L7;
    }

/*  Corresponds to Data reduction unique set for  pg4 */
/*   7 pg4    4/m        hkl:h>=0, l>=0 with k>=0 if  h=0  and */
/*                                           k> 0 if  h>0. */
L400:
/* ----4/M */
/*     HKL --    H.GE.0, L.GE.0, WITH K.GE.0 IF H.EQ.0 OR */
/*               K.GE.1 IF H.GT.0 */
    if (l >= 0) {
	goto L405;
    } else {
	goto L7;
    }
L405:
    if (j < 0) {
	goto L7;
    } else if (j == 0) {
	goto L410;
    } else {
	goto L415;
    }
L410:
    if (k >= 0) {
	goto L5;
    } else {
	goto L7;
    }
L415:
    if (k <= 0) {
	goto L7;
    } else {
	goto L5;
    }

/*  Corresponds to Data reduction unique set for  pg422 */
/*   8 pg422 4/mmm       hkl:h>=0, k>=0, l>=0 and h>=k   89.. */
L450:
/* ----4/MMM */
/*     HKL --    H.GE.0, K.GE.0, AND L.GE.0 WITH H.GE.K */
    if (j >= 0) {
	goto L455;
    } else {
	goto L7;
    }
L455:
    if (k >= 0) {
	goto L460;
    } else {
	goto L7;
    }
L460:
    if (l >= 0) {
	goto L465;
    } else {
	goto L7;
    }
L465:
    if (j - k >= 0) {
	goto L5;
    } else {
	goto L7;
    }

/* 500  CONTINUE */
/* ----3 BAR */
/*      HKL ---   H.LE.0, L.GE.0, WITH K.GE.0 IF H.LE.0 OR */
/*               K.GE.1 IF H.LT.0 */
/*      HK0  --  K.GT.-H */
/*  INCLUDE 0 0 L,  HK0 K.LT.0 H.GE.-K     HKL H.GE.0 K.LT.0 */
/*  Corresponds to Data reduction unique set for  pg3 */
/*  9 pg3     3bar      hkl:h>=0, k>0  00l:l>0         143.. */
L550:
/*   ALTERNATIVE FOR R3 */
/*  H.GE.0 K.GT.0    ALL L */
/*  H=0 K=0 L.GT.0 */
    if (j < 0) {
	goto L7;
    } else if (j == 0) {
	goto L555;
    } else {
	goto L556;
    }
L555:
    if (k < 0) {
	goto L7;
    } else if (k == 0) {
	goto L557;
    } else {
	goto L5;
    }
L556:
    if (k <= 0) {
	goto L7;
    } else {
	goto L5;
    }
L557:
    if (l <= 0) {
	goto L7;
    } else {
	goto L5;
    }

/*  Corresponds to Data reduction unique set for  pg312 */
/*  10 pg312        hkl:h>=0, k>=0 with k<=h for all l. */
/*                           if k = 0  l>=0 */
L570:
/* ----3 BAR M */
/*  H.GE.0 K.GE.0 K.LE.H   ALL L */
/*  H=K  L.GE.0 */
    if (j >= 0) {
	goto L575;
    } else {
	goto L7;
    }
L575:
    if (k < 0) {
	goto L7;
    } else if (k == 0) {
	goto L577;
    } else {
	goto L576;
    }
L576:
    if (j - k >= 0) {
	goto L5;
    } else {
	goto L7;
    }
L577:
    if (l >= 0) {
	goto L5;
    } else {
	goto L7;
    }

/*  Corresponds to Data reduction unique set for  pg321 */
/*  11 pg321        hkl:h>=0, k>=0 with k<=h for all l. */
/*                           if h = k  l>=0 */
L600:
/* ----3 BAR M */
/*  H.GE.0 K.GE.0 K.LE.H   ALL L */
/*  H=K  L.GE.0 */
    if (j >= 0) {
	goto L605;
    } else {
	goto L7;
    }
L605:
    if (k >= 0) {
	goto L606;
    } else {
	goto L7;
    }
L606:
    if ((i__1 = j - k) < 0) {
	goto L7;
    } else if (i__1 == 0) {
	goto L607;
    } else {
	goto L5;
    }
L607:
    if (l >= 0) {
	goto L5;
    } else {
	goto L7;
    }

/*  Corresponds to Data reduction unique set for  pg6 */
/*  12 pg6    6/m        hkl:h>=0, k>=0, l>=0 with k>=0 if  h=0 */
/*                                            and  k> 0 if  h>0. */
L650:
/* ----6/M */
/*     HKL --    H.GE.0, L.GE.0, WITH K.GE.0 IF H.EQ.0 OR */
/*               K.GE.1 IF H.GT.0 */
    if (l >= 0) {
	goto L655;
    } else {
	goto L7;
    }
L655:
    if (j < 0) {
	goto L7;
    } else if (j == 0) {
	goto L660;
    } else {
	goto L665;
    }
L660:
    if (k >= 0) {
	goto L5;
    } else {
	goto L7;
    }
L665:
    if (k <= 0) {
	goto L7;
    } else {
	goto L5;
    }

/*  Corresponds to Data reduction unique set for  pg622 */
/*  13 pg622 6/mmm       hkl:h>=0, k>=0, l>=0 with h>=k 177.. */
L700:
/* ----6/MMM */
/*     HKL --    H.GE.0, K.GE.0, AND L.GE.0 WITH H.GE.K */
    if (j >= 0) {
	goto L705;
    } else {
	goto L7;
    }
L705:
    if (k >= 0) {
	goto L710;
    } else {
	goto L7;
    }
L710:
    if (j - k >= 0) {
	goto L715;
    } else {
	goto L7;
    }
L715:
    if (l >= 0) {
	goto L5;
    } else {
	goto L7;
    }

/*  Corresponds to Data reduction unique set for  pg23 */
/*  14 pg23   m3         hkl:h>=0, k>=0, l>=0 with l>=h, */
/*                                                 k>=h if l=h */
/*                                                 k> h if l>h. */
L750:
/* ----M3 */
/*     HKL --    H.GE.0, K.GE.0, AND L.GE.0 WITH L.GE.H AND WITH */
/*               K.GE.H IF L.EQ.H OR K.GT.H IF L.GT.H */
    if (j >= 0) {
	goto L755;
    } else {
	goto L7;
    }
L755:
    if (k >= 0) {
	goto L760;
    } else {
	goto L7;
    }
L760:
    if (l >= 0) {
	goto L765;
    } else {
	goto L7;
    }
L765:
    if ((i__1 = l - j) < 0) {
	goto L7;
    } else if (i__1 == 0) {
	goto L770;
    } else {
	goto L775;
    }
L770:
    if (k - j >= 0) {
	goto L5;
    } else {
	goto L7;
    }
L775:
    if (k - j <= 0) {
	goto L7;
    } else {
	goto L5;
    }

/*  Corresponds to Data reduction unique set for  pg432 */
/*  15 pg432  m3m        hkl:h>=0, k>=0, l>=0 with k>=l, and l>=h. */
L800:
/* ----M3M */
/*     HKL --    H.GE.0, K.GE.0, AND L.GE.0 WITH */
/*              K.GE.L AND L.GE.H */
    if (j >= 0) {
	goto L805;
    } else {
	goto L7;
    }
L805:
    if (k >= 0) {
	goto L810;
    } else {
	goto L7;
    }
L810:
    if (l >= 0) {
	goto L815;
    } else {
	goto L7;
    }
L815:
    if (k - l >= 0) {
	goto L820;
    } else {
	goto L7;
    }
L820:
    if (l - j >= 0) {
	goto L5;
    } else {
	goto L7;
    }

/* ----REFLECTION IS IN BOUNDS */
L5:
    ret_val = isgn;
    return ret_val;

/* ----REFLECTION IS OUT OF BOUNDS */
L7:
    if (isgn == 1) {
/*     TRY -H -K -L */
	isgn = -1;
	j = -j;
	k = -k;
	l = -l;
	goto L1;
    }
/* Index failed, exit */
    ret_val = 0;
    return ret_val;
}
/*
*************************
**    rdm_set_centric  **
**************************

PURPOSE: Set up for centric reflections calculation

RETURN:   +1  if  h k l chosen 
          -1  if -h-k-l chosen 
           0  if reflection is out-of-bounds 

AUTHOR: Translated from CCP4 routine CENTRIC

**** RDM SERVICE ROUTINE ONLY ****
*/

int rdm_set_centric ()

{
   static int ihkl[][3] = {{0,1,2}, {1,0,2}, {1,2,0}, {1,1,10}, {1,10,1},
                          {10,1,1}, {1,-1,10}, {1,10,-1}, {10,1,-1}, {-1,2,10},
                          {2,-1,10}, {1,4,8}};
   static float cprj[][3] = {{1.0,0.0,0.0}, {0.0,1.0,0.0}, {0.0,0.0,1.0}, 
                             {1.0,-1.0,0.0}, {1.0,0.0,-1.0}, {0.0,1.0,-1.0}, 
                             {1.0,1.0,0.0}, {1.0,0.0,1.0}, {0.0,1.0,1.0}, 
                             {2.0,1.0,0.0}, {1.0,2.0,0.0}, {0.0,0.0,0.0}}; 
   int nc;
   int in[3];
   int ih, ik, il;
   int j;

   if (nsym==0)
   {
      cent_setup = 0;
      return 1;
   }
   cent_setup = 1;
   ncent = 0;

   for(nc=0;nc<10;++nc)
   {
      in[0] = ihkl[nc][0];
      in[1] = ihkl[nc][1];
      in[2] = ihkl[nc][2];

      /*  Generate symm equivs */

      /* test whether h' k' l' equals -h -k -l */

      for (j=0;j<nsym;++j)
      {
          ih = in[0]*rsym[j][0][0] +in[1]*rsym[j][1][0] +in[2]*rsym[j][2][0];
          if (ih==-in[0])
          {
            ik = in[0]*rsym[j][0][1] + in[1]*rsym[j][1][1] +
                 in[2]*rsym[j][2][1];
            if (ik==-in[1])
            {
              il = in[0]*rsym[j][0][2] + in[1]*rsym[j][1][2] +
                   in[2]*rsym[j][2][2];
              if (il==-in[2]) goto lab50;
            }
          }
      }
      continue;

      lab50:
      cproj[ncent][0] = cprj[nc][0];
      cproj[ncent][1] = cprj[nc][1];
      cproj[ncent][2] = cprj[nc][2];
      ncent = ncent + 1;
   }
   return 0;
}
/*
*********************
**    rdm_determ   **
*********************

PURPOSE: Calculate determinant of a 4x4 matrix

AUTHOR: Translated from CCP4 routine DETERM

**** RDM SERVICE ROUTINE ONLY ****
*/

void rdm_determ (float a[4][4], float * det)

{
   int ii, jj;
   int i, j;
   int i1, j1;
   float x[4][4];
   float c[4][4];
   float d;
   float am;

   for (ii=0;ii<4;++ii)
   {
      for (jj=0;jj<4;++jj)
      {
         i = -1;
         for (i1=0;i1<4;++i1)
         {
            if (i1!=ii)
            {
               i = i + 1;
               j = -1;
               for (j1=0;j1<4;++j1)
               {
                 if (j1!=jj)
                 {
                    j = j + 1;
                    x[i][j] = a[i1][j1];
                 }
               }
            }
         }

         am = x[0][0]*x[1][1]*x[2][2] - x[0][0]*x[1][2]*x[2][1] +
              x[0][1]*x[1][2]*x[2][0] - x[0][1]*x[1][0]*x[2][2] +
              x[0][2]*x[1][0]*x[2][1] - x[0][2]*x[1][1]*x[2][0];
         c[ii][jj] = powf (-1.0,(float)(ii+1+jj+1))*am;
      }
   }

   /*  Calculate determinant */

   d = 0.0;
   for (i=0;i<4;++i)
   {
      d = a[i][0]*c[i][0] + d;
   }
   *det = d;
   return;
}
/*
*********************
**    rdm_invsym   **
*********************

PURPOSE: Invert 4x4 symmetry matrix

AUTHOR: Translated from CCP4 routine INVSYM

**** RDM SERVICE ROUTINE ONLY ****
*/

void rdm_invsym (float a[4][4], float ainv[4][4])

{
   int ii, jj;
   int i, j;
   int i1, j1;
   float x[4][4];
   float c[4][4];
   float d;
   float am;

   for (ii=0;ii<4;++ii)
   {
      for (jj=0;jj<4;++jj)
      {
         i = -1;
         for (i1=0;i1<4;++i1)
         {
            if (i1!=ii)
            {
               i = i + 1;
               j = -1;
               for (j1=0;j1<4;++j1)
               {
                 if (j1!=jj)
                 {
                    j = j + 1;
                    x[i][j] = a[i1][j1];
                 }
               }
            }
         }

         am = x[0][0]*x[1][1]*x[2][2] - x[0][0]*x[1][2]*x[2][1] +
              x[0][1]*x[1][2]*x[2][0] - x[0][1]*x[1][0]*x[2][2] +
              x[0][2]*x[1][0]*x[2][1] - x[0][2]*x[1][1]*x[2][0];
         c[ii][jj] = powf (-1.0,(float) (ii+1+jj+1))*am;
      }
   }

   /*  Calculate determinant */

   d = 0.0;
   for (i=0;i<4;++i)
   {
      d = a[i][0]*c[i][0] + d;
   }
   

   /* Get inverse matrix */

   for (i=0;i<4;++i)
   {
      for(j=0;j<4;++j) ainv[i][j] = c[j][i]/d;
   }
   return;
}

void debug_symm(int k)
{
   int isym;
   int i;

   printf("Symmetry debug %d\n",k);
   for (isym=0;isym<nsym;++isym)
   {
      printf ("Symop %d\n",isym+1);
      for (i=0;i<4;++i)
      {
         printf ("%f %f %f %f\n", rsym[isym][i][0],
              rsym[isym][i][1],rsym[isym][i][2],rsym[isym][i][3]);
      }
   }     
   for (isym=0;isym<nsym;++isym)
   {
      printf ("Symop inv %d\n",isym+1);
      for (i=0;i<4;++i)
      {
         printf ("%f %f %f %f\n", rsym_inv[isym][i][0],
              rsym_inv[isym][i][1],rsym_inv[isym][i][2],rsym_inv[isym][i][3]);
      }
   }     
}
